/* JPI TopSpeed C demo: Bezier curves.             HUR 1/11/89 */
/*                                                             */
/* COPYRIGHT (C) 1989 Jensen and Partners. All Rights Reserved */

#define _JPI_WIN_
#include <stdlib.h>
#include <stdio.h>
#include <graph.h>
#include <conio.h>
#include <choose.h>

extern struct videoconfig cv;

typedef double coord;

typedef struct { coord X,Y; } point;


static void Bezier ( unsigned m, point Ctrl[],
              unsigned n, point Curve[] )
{
  unsigned i,j;
  coord M,N,I,T,P,X,Y;

  M = (coord)--m;
  N = (coord)--n;
  Curve[0] = Ctrl[0];
  for ( j = 1; j < n; ++j )
  {
    X = Y = 0.0;
    T = (coord)j / N;
    for ( P = 1.0, i = 1; i <= m; ++i )
      P *= 1.0 - T;
    for ( i = 0; i <= m; ++i )
    {
      X += P * Ctrl[i].X;
      Y += P * Ctrl[i].Y;
      I = (coord)i;
      P *= T / (1.0 - T) * (M - I) / (I + 1.0); /* Not too efficient... */
    }
    Curve[j].X = X;
    Curve[j].Y = Y;
  }
  Curve[n] = Ctrl[m];
}


static void Place ( point *p )
{
  p->X = (coord) random ( cv.numxpixels );
  p->Y = (coord) random ( cv.numypixels );
}


static void Draw ( unsigned c, unsigned n, point *Curve )
{
  _setcolor ( c );
  _moveto ( (short) Curve->X, (short) Curve->Y );
  while ( --n )
  {
    ++Curve;
    _lineto ( (short) Curve->X, (short) Curve->Y );
  }
}


#define N  20

void bezier_demo ()
{
  point Ctrl[4], Curve[N];
  unsigned p[4],i,c,z;

  clrscr();
  cputs ( "    Curves are determined by four anchor points.\n" );
  cputs ( "    Designs are created by keeping two points fixed\n" );
  cputs ( "    and letting the remaining two points float.\n\n" );
  cputs ( "    Press a key to start...<ESC> to stop\n" );
  getch();

  Choose ();

  randomize ();
  p[0] = 0; p[1] = 1; p[2] = 2; p[3] = 3;
  for (;;)
  {
    switch ( random(20) )
    { case 0: p[0] = 0; p[1] = 1; p[2] = 2; p[3] = 3; break;
      case 1: p[0] = 0; p[1] = 1; p[2] = 3; p[3] = 2; break;
      case 2: p[0] = 0; p[1] = 3; p[2] = 2; p[3] = 1; break;
      case 3: p[0] = 1; p[1] = 0; p[2] = 3; p[3] = 2; break;
      default: ;
    }
    Place ( &Ctrl[p[0]] );
    Place ( &Ctrl[p[3]] );
    z = 0;
    for ( i = 0; i < 1+random(4); ++i )
      z |= 1 << random ( cv.numcolors-1 );
    for ( i = 0; i < 5+random(20); ++i )
    {
      Place ( &Ctrl[p[1]] );
      Place ( &Ctrl[p[2]] );
      Bezier ( 4, Ctrl, N, Curve );
      do
        c = random ( cv.numcolors-1 );
      while ( (z & (1<<c)) == 0);
      Draw ( c+1, N, Curve );
      if (kbhit()) break;
    }
    if (kbhit() && getch() == 27) break;
    delay ( 1000 );
    _clearscreen ( _GCLEARSCREEN );
  }
  _setvideomode ( _DEFAULTMODE );
}
