/* The following program plays cribbage with computer acting as opponent.
   Originally written for the Zortech compiler and released under the
   title SamCribbage.

   Ported to JPI Topspeed C++ Jan 1991

*/

/* Will eventually need to be accompanied by a documentation file on rules
   of cribbage.
   Note specifically one type GO to pass
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <disp.hpp>
#ifdef __OS2__
#include "os2glue.hpp"
#else
#include <bios.h>
#include <dos.h>
#endif

#include "winkey.hpp"
class guessed_hand {
  private:
    float cards[52];
    int number;
    int possibles;
  public:
    guessed_hand(int n);
    void has_not(int c);
    float value_chance(int v);
    float suit_chance(int s);
    float card_chance(int c) { return cards[c]; }
};

class known_hand {
  private:
    int *cards;
    int number;
    int store_number;
    Window *my_win;
  public:
    known_hand(int n);
    void operator=(known_hand& h);
    known_hand(known_hand& h);
    ~known_hand() { delete cards; }
    int remove_card(int c);
    void assign_Window(Window& m) {my_win=&m;}
    void return_hand() { number=0; }
    void add_card(int c);
    void print(char const * const c,int hidden=0);
    int value_number(int v);
    int suit_number(int s);
    int card_number(int c);
    int pick_value(int v);
    int pick_suit(int s);
    int pick_within(int v);
    int size() { return number; }
    int start() { return cards[number-1]; }
    int operator[](int c) { return cards[c]; }
    int total();
    friend void print_box(known_hand& box);
};

typedef enum {ace,two,three,four,five,six,seven,eight,nine,ten,
              jack,queen,king} card_value;
typedef enum {clubs,diamonds,hearts,spades} suit_value;
#define suit(a) ((a)>>8)
#define value(a) ((signed char)(a))
#define make_card(st,va) ((st)<<8|(va))
#define numeric_value(a) (value(a)>=ten ? 10:value(a)+1)

class pack {
  private:
    int number;
    int cards[52];
  public:
    pack();
    void random_add_card(int c);
//  void back_add_card(int c) { pack[number++]=c; }
//  void shuffle_pack();
    int extract_card();
};

#define debugging 0
#define blue_on_white 0x71
#define white_on_blue 0x17
#define black_on_brown 0x60
#define white_on_brown 0x67
static int noise=1;
static waiting=1;

enum turn {player,computer};
static enum lev {naive,simple} level=simple;
static enum guess {none,single,forced} auto_play=single;
static enum turn scoring=player;

pack deck;
int start;
static whole_screen prog_screen(white_on_blue,0,0,80,25);
static Window title_win(blue_on_white,0);
static Window command_win(white_on_blue,24);
static Window comment_win(white_on_blue,23);
static Window player_win(white_on_blue,16,0,40,7);
static Window computer_win(white_on_blue,1,0,40,7);
static Window out_win(white_on_blue,8,0,70,7);
static Window start_win(white_on_blue,8,70,5,6);
static Window score_win(black_on_brown,1,40,40,7);


#if debugging
void card_error(const char const * c)
{ puts(c);
  exit(0);
}
#endif

void sound_beep(unsigned x)
{
#ifndef __OS2__
  sound(2000-x);
  delay(250);
  nosound();
#endif
}

void display_menu(char *m)
{ command_win.clear();
  disp_puts(m);
}

void command_process()
{ display_menu("Beeps Guessing Level Waiting Scoring Remember Quit");
  int b=bioskey(0)>>8;
  switch (b)
  { case 0x30:
      noise=noise^1;
      break;
    case 0x22:
      display_menu("None Single_key Forced_moves");
      b=bioskey(0)>>8;
      auto_play=b==0x31?none:b==0x1f?single:b==0x21?forced:auto_play;
      break;
    case 0x26:
      display_menu("Naive Simple");
      b=bioskey(0)>>8;
      level=b==0x31?naive:b==0x1f?simple:level;
      break;
    case 0x11:
      display_menu("Wait Nowait");
      b=bioskey(0)>>8;
      waiting=b==0x31?0:b==0x11?1:waiting;
      break;
    case 0x1f:
      display_menu("Computer Player");
      b=bioskey(0)>>8;
      scoring=b==0x2e?computer:b==0x19?player:scoring;
      break;
    case 0x13:
    { FILE *savefile=fopen("samcrib.cfg","wb");
      if (savefile)
      { fputc(noise,savefile);
	fputc(scoring,savefile);
        fputc(auto_play,savefile);
	fputc(level,savefile);
	fputc(waiting,savefile);
        fclose(savefile);
      }
      else
      { comment_win.clear();
        disp_puts("Unable to create file");
      }
      break;
    }
    case 0x10:
      display_menu("Do you really want to quit?");
      if (bioskey(0)>>8==0x15) exit(0);
  }
  command_win.clear();
}

char b_key(char *message)
{ int b,done;
  do
  { done=0;
    command_win.clear();
    disp_puts(message);
    b=bioskey(0);
    if (b>>8==60) {
      command_process();
      done=1;
    }
  } while (done);
  char c=b;
  if (c>='a' && c<='z') c-=0x20;
  return c;
}

void pause()
{ comment_win.clear();
  b_key("Waiting for a keypress");
  command_win.clear();
}
/*
guessed_hand::guessed_hand(int n)
{ for (int i=0;i<52;i++)
    cards[i]=float(n)/52.0;
  number=n;
  possibles=52;
}

void guessed_hand::has_not(int c)
{ cards[c]=0;
  for (int i=0;i<52;i++)
    if (cards[i]<1)
      cards[i]*=(float)possibles/(float)(possibles-1);
  possibles--;
}

float guessed_hand::value_chance(int v)
{ float t=0;
  for (int i=v;i<52;i+=13) t+=(1-t)*cards[v];
  return t;
}

float guessed_hand::suit_chance(int s)
{ float t=0;
  for (int i=s*13;i<(s+1)*13;i++)
    t+=(1-t)*cards[s];
  return t;
}
*/
known_hand::known_hand(int n)
{ number=0;
  store_number=n;
  cards=new int[n];
}

void known_hand::operator=(known_hand& h)
{ delete cards;
  cards=new int[h.store_number];
  store_number=h.store_number;
  number=h.number;
  my_win=h.my_win;
  memcpy(&cards[0],&h.cards[0],number*sizeof(int));
}

known_hand::known_hand(known_hand& h)
{ cards=new int[h.store_number];
  store_number=h.store_number;
  number=h.number;
  my_win=h.my_win;
  memcpy(&cards[0],&h.cards[0],number*sizeof(int));
}

int known_hand::remove_card(int c)
{
  #if debugging
  if (!number) card_error("Removed too many");
  #endif
  for (int i=0; i<number; i++)
    if (cards[i]==c)
    { for (int j=i; j<number-1; j++)
        cards[j]=cards[j+1];
      cards[--number]=0;
      return 1;
    }
  return 0;
}

void known_hand::add_card(int c)
{
  #if debugging
  if (number==store_number) card_error("Added too many");
  #endif
  cards[number++]=c;
}

int known_hand::value_number(int v)
{ int t=0;
  if (v<ace || v>king) return 0;
  for (int i=0 ; i<number ; i++)
    if (value(cards[i])==v) t++;
  return t;
}

int known_hand::suit_number(int s)
{ int t=0;
  for (int i=0 ; i<number ; i++)
    if (suit(cards[i])==s) t++;
  return t;
}

int known_hand::card_number(int c)
{ for (int i=0 ; i<number ; i++)
    if (cards[i]==c) return 1;
  return 0;
}

int known_hand::pick_value(int v)
{ if (value_number(v)!=1) return -1;
  for (int i=0; i<number; i++)
    if (value(cards[i])==v) return cards[i];
}

int known_hand::pick_suit(int s)
{ if (suit_number(s)!=1) return -1;
  for (int i=0; i<number; i++)
    if (suit(cards[i])==s) return cards[i];
}

int known_hand::pick_within(int v) //-2==>too many -1==>GO
{ int tot=0;
  int rv=-1;
  for (int i=0;i<number;i++)
    if (numeric_value(cards[i])<=v)
    { if (tot++) return -2;
      rv=cards[i];
    }
  return rv;
}

int known_hand::total()
{ int t=0;
  for (int i=0;i<number;i++) t+=numeric_value(cards[i]);
  return t;
}

pack::pack()
{ srand(unsigned(time(NULL)));
  number=0;
  for (int i=ace;i<=king;i++)
    for (int j=clubs;j<=spades;j++)
      random_add_card(make_card(j,i));
}

void pack::random_add_card(int c)
{ int pos=rand() % (number+1);
#if debugging
  for (int i=0;i<number;i++)
    if (cards[i]==c)
      puts("Cards duplicated");
#endif
  if (pos==number)
    cards[number++]=c;
  else
  { memmove(&cards[pos+1],&cards[pos],(number++-pos)*sizeof(int));
    cards[pos]=c;
  }
#if debugging
  if (number>52) card_error("Too much added to pack");
#endif
}

int pack::extract_card()
{ int t=cards[0];
  memmove(&cards[0],&cards[1],(--number)*sizeof(int));
  return t;
}

/* Quicky compute of value of 3 cards */
int three_score(int a,int b,int c)
{ int pairs= a==b ? b==c ? 6:2
                  : ( b==c ? 2 : ( a==c ? 2:0) );
  int fifteens = (a+b+c==15)+(a+b==15)+(a+c==15)+(b+c==15);
  if (a<b) { int t=b; b=a; a=t; }
  if (b<c) { int t=c; c=b; b=t; }
  if (a<b) { int t=b; b=a; a=t; }
  return pairs+fifteens*2+(a-c==2?3:0);
}

int lay_score(known_hand& lay)
{ int rmin=value(lay.start());
  int rmax=rmin;
  int ord=lay.size();
  int t=0;
  int tot=lay.total();
  if (tot==15 || tot==31) t+=2;
  if (tot>31) return -1;
  int ptot=0;
  for (int i=ord-2;i>=0 && value(lay[i])==rmin;i--) ptot+=ord-i-1;
  if (ptot) return t+ptot*2;
  ptot=0;
  unsigned int record=1<<rmin;
  for (i=ord-2;i>=0 && rmax-rmin<ord; i--)
  { int l=value(lay[i]);
    if (record & 1<<l) break;
    record|=1<<l;
    if (l<rmin) rmin=l;
    if (l>rmax) rmax=l;
    if (rmax-rmin==ord-i-1) ptot=ord-i;
  }
  if (ptot<3)
    return t;
  return t+ptot;
}

int play_card(known_hand& lay,known_hand& me)
{ static unsigned char point_values[]={40,15,14,13,12,31,16,11,10,9,19,23,22,21,
  20,2,6,7,8,27,28,30,29,26,25,24,18,17,5,4,3,1,40,40,40,40,40,40,40,40,40,40};
  int ord=me.size();
  int best=-1;
  int best_score=-1;
  int best_equal=0;
  for (int i=0;i<ord;i++)
  { lay.add_card(me[i]);
    int res=lay_score(lay);
    int c_tot=lay.total();
    lay.remove_card(me[i]);
    if (res>best_score ||
        level>naive && res==best_score && point_values[c_tot]<best_equal)
    { best_score=res;
      best=i;
      best_equal=point_values[c_tot];
    }
  }
  return best;
}

int pairs_make(known_hand& h,int amount)
{ int t=0;
  for (int i=0; i<4; i++)
    for (int j=i+1; j<5; j++)
      if (numeric_value(h[i])+numeric_value(h[j])==amount) t++;
  return t;
}

int hand_score(known_hand& h,int full=1)
{ int t=0;
  /* First deal with flushes */
  if (full)
  { for (int i=0;i<4;i++)
    { int num_in_suit=h.suit_number(i);
      if (num_in_suit==5)
        t+=5;
      if (num_in_suit==4 && suit(h.start())!=i)
        t+=4;
      if (num_in_suit>=2) break;
    }
    /* Jack of start suit */
    if (!(value(h.start())==jack))
      if (h.card_number(make_card(suit(h.start()),jack))) t+=1;
  }
  /* Pairs */
  int cum_tot=0;
  for (int i=ace; i<=king; i++)
  { int num_of_value=h.value_number(i);
    t+=num_of_value<2?0:num_of_value==2?2:num_of_value==3?6:12;
    cum_tot+=num_of_value;
    if (cum_tot>=4) break;
  }
  /* Runs */
  for (i=ace; i<queen; i++)
    if (h.value_number(i) && h.value_number(i+1) && h.value_number(i+2))
    { if (h.value_number(i+3))
        if (h.value_number(i+4))
	  t+=5;
        else
          t+=4*h.value_number(i)*h.value_number(i+1)*h.value_number(i+2)*
	       h.value_number(i+3);
      else
	t+=3*h.value_number(i)*h.value_number(i+1)*h.value_number(i+2);
      break;
    }
  /* 15ns */
  int tot=h.total();
  if (tot==15) return t+2;
  if (tot<15) return t;		//Five cards
  if (tot==25) t+=2*(h.value_number(ten)+h.value_number(jack)+
                     h.value_number(queen)+h.value_number(king)); //Four cards
  else
    if (tot<25) t+=2*h.value_number(tot-16); //eg if 2 aces and total=16 the + 4
  if (tot<17) return t;
  t+=2*pairs_make(h,tot-15);
  return t+2*pairs_make(h,15);
}

int correct_throw(known_hand& h,int& first,int &second,int my_box)
{ int best_yet=0;
  int best_tot=0;
  for (int i=0;i<5;i++)
    for (int j=i+1;j<6;j++)
    { known_hand possible=h;
      possible.remove_card(h[i]);
      possible.remove_card(h[j]);
      int tot=0;
      int box=0;
      for (int l=clubs;l<=spades;l++)
      { if (possible.suit_number(l)==4)
          tot+=46*4+13-h.suit_number(l);
        if (possible.value_number(make_card(l,jack)))
	  tot+=13-h.suit_number(l);
      }
      for (int k=ace;k<=king;k++)
	{ possible.add_card(make_card(clubs,k));
	  tot+=hand_score(possible,0)*(4-h.value_number(k));
          box+=three_score(value(h[i]),value(h[j]),k)*2*(4-h.value_number(k));
	  possible.remove_card(make_card(clubs,k));
	}
      int a_tot=my_box?tot+box:tot-box;
      if (a_tot>=best_yet)
      { first=h[i];
	second=h[j];
	best_yet=a_tot;
        best_tot=tot;
      }
    }
  return (best_tot+23)/46;
}


char *vals="A23456789TJQKO";
char *sus="CDHSG";

int get_card(known_hand &h,known_hand *l)
{ char va,su;
  char ok=0;
  if (auto_play>=forced && l)
  { int rv=h.pick_within(31-l->total());
    if (rv!=-2) return rv;
  }
  char *t,*s;
  do
  {
  if (ok)
  { comment_win.clear();
    if (noise)
      sound_beep(1400);
    disp_puts("Invalid entry, try again");
  }
  ok=1;
  va=b_key("Enter value of card >");
  disp_putc(va);
  t=strchr(vals,(int)va);
  if (auto_play>none)
  { if (va=='G') return -1;
    int rv;
    if (t)
    { rv=h.pick_value(t-vals);
      if (rv!=-1) return rv;
    }
    else
    { s=strchr(sus,(int)va);
      rv=h.pick_suit(s-sus);
      if (rv!=-1) return rv;
    }
  }
  if (!t) continue;
  su=b_key("Enter suit of card >");
  disp_putc(su);
  s=strchr(sus,(int)su);
  if (va=='G') return -1;
  } while (!s || !t);
  return make_card(s-sus,t-vals);
}

void print_card(int c,Window& w,int hidden=0)
{ static char fs[]="\05\04\03\06";
  char tmp=suit(c);
  byte colour=suit(c)==clubs || suit(c)==spades?0x70:0x74;
  if (hidden) colour=0x22;
  w.change_attr(colour);
  w.clear();
  disp_putc(vals[value(c)]);
  w.cursor(2,2);
  disp_putc(fs[tmp]);
  w.cursor(4,4);
  disp_putc(vals[value(c)]);
}

void known_hand::print(char const * const c,int hidden)
{ my_win->clear();
  disp_puts(c);
  for (int i=0;i<number;i++)
  { Window c_win(blue_on_white,1,6*i,5,5,*my_win);
    print_card(cards[i],c_win,hidden);
  }
  my_win->cursor(6,0);
}


void print_box(int crib)
{ score_win.cursor(0,0);
  if (crib==player)
    disp_puts("   ");
  else
    disp_puts("BOX");
  score_win.cursor(6,0);
  if (crib==computer)
    disp_puts("   ");
  else
    disp_puts("BOX");
}

void print_peg(int whose,int value,int character)
{ int yp,xp;
  if (value)
  { if (value>60) value-=60;
    if (value==61)
    { yp=3;
      xp=0;
    }
    else
    { yp=value>30?2:1;
      if (value>30) value-=30;
      xp=((value-1)/5)*6+(value-1)%5;
      if (yp==2) xp=34-xp;
      xp=xp+3;
      if (whose==player) yp=6-yp;
    }
    score_win.cursor(yp,xp);
    disp_putc(character);
  }
}

void add_score(int p,int c)
{ static scores[2]={0,0};
  static lasts[2]={0,0};
  while (p || c)
  { int play=p?player:computer;
    int sv=p?p:c;
    print_peg(play,lasts[play],'');
    lasts[play]=scores[play];
    for (int i=0; i<sv;i++)
    { if (scores[play]!=lasts[play])
        print_peg(play,scores[play],'');
      scores[play]++;
      score_win.change_attr(white_on_brown);
      print_peg(play,scores[play],'');
      score_win.change_attr(black_on_brown);
      if (noise) sound_beep(1200-10*i);
      if (scores[play]==121) break;
    }
    score_win.cursor(3,1);
    disp_puts("                                     ");
    score_win.cursor(3,4);
    if (scores[play]>=121)
      if (play==computer)
      { disp_puts("I win");
        pause();
        exit(0);
      }
      else 
      { disp_puts("You win");
        pause();
        exit(0);
      }
    if (play==player)
      p=0;
    else
      c=0;
  }
}

void account_hand(known_hand& h,char *message,int owner)
{ h.print(message);
  for (int i=0;i<4;i++) deck.random_add_card(h[i]);
  h.add_card(start);
  int t=hand_score(h,1);
  int sc=0; //Used to accumulate suggested score and then as muggins points
  if (scoring==computer || owner==computer)
  { char temp[40];
    sprintf(temp,"The score for this hand is %d",t);
    disp_puts(temp);
  }
  else
  { int ok=0;
    do
    { if (ok)
      { comment_win.clear();
        if (noise) sound_beep(1400);
        if (ok==1)
          disp_puts("Invalid entry");
        else
          disp_puts("Someone's telling porkies!");
      }
      ok=1;
      char d1=b_key("Enter 1st digit of score>");
      d1-='0';
      if (d1>=0 && d1<10)
      { char d2=d1?b_key("Enter 2nd digit of score>")-'0':0;
        sc=(d2>=0 && d2<10)?d1*10+d2:d1;
        if (sc>t)
          ok=2;
        else
        { if (sc<t)
          { comment_win.clear();
            char temp[40];
            sprintf(temp,"Muggins for %d",t-sc);
            disp_puts(temp);
            int hold=sc;
	    sc=t-sc;
            t=hold;
            b_key("Waiting for a keystroke");
            command_win.clear();
          }
          else
            sc=0;
          break;
        }
      }
    } while (1);
  }
  h.return_hand();
  if (owner==player)
    add_score(t,sc);
  else
    add_score(sc,t);
  comment_win.clear();
}

int main(int argc,char *argv[])
{ known_hand my_hand(6);
  known_hand his_hand(6);
  known_hand box(5);
  char crib=computer;
  FILE *cfgfile=fopen("samcrib.cfg","rb");
  if (cfgfile)
  { noise=fgetc(cfgfile);
    scoring=(enum turn)fgetc(cfgfile);
    auto_play=(enum guess)fgetc(cfgfile);
    level=(enum lev)fgetc(cfgfile);
    waiting=fgetc(cfgfile);
    fclose(cfgfile);
  }
  disp_setmode(3);
  disp_open();
  disp_hidecursor();
  prog_screen.add_Window(title_win);
  prog_screen.add_Window(command_win);
  prog_screen.add_Window(player_win);
  prog_screen.add_Window(computer_win);
  prog_screen.add_Window(out_win);
  prog_screen.add_Window(start_win);
  prog_screen.add_Window(comment_win);
  prog_screen.add_Window(score_win);
  my_hand.assign_Window(computer_win);
  his_hand.assign_Window(player_win);
  prog_screen.display();
  title_win.cursor();
  disp_puts("    Cribbage V1.00 (C)1990 Jensen & Partners      F2=Commands");
  start_win.cursor(1,0);
  disp_puts("Start");
  start_win.resize(10,70,5,5);
  for (int i=1;i<62;i++)
  {  print_peg(player,i,'');
     print_peg(computer,i,'');
  }
  do
  { if (crib==player) 
      box.assign_Window(player_win);
    else
      box.assign_Window(computer_win);
    for (int i=0; i<6; i++)
    { my_hand.add_card(deck.extract_card());
      his_hand.add_card(deck.extract_card());
    }
    his_hand.print("Your cards are:-");
    start_win.clear();
    int first,second;
    comment_win.clear();
    disp_puts("Please throw away two cards");
    correct_throw(my_hand,first,second,crib==computer);
    box.add_card(first);
    box.add_card(second);
    my_hand.remove_card(first);
    my_hand.remove_card(second);
    my_hand.print("Computer Hand",1);
    print_box(crib);
    for (int sloop=0;sloop<2;sloop++)
    { int tt;
      while(!his_hand.remove_card(tt=get_card(his_hand,NULL)))
      { comment_win.clear();
        disp_puts("You did not have that card to throw!");
      }
      box.add_card(tt);
      his_hand.print("Your cards are:-");
    }
    comment_win.clear();
    start=deck.extract_card();
    start_win.change_attr(suit(start)&3?0x74:0x70);
    print_card(start,start_win);
    score_win.cursor(4,0);
    if (value(start)==jack)
    { add_score(crib==player?2:0,crib==player?0:2);
      disp_puts("Dealer scores 2 for his heels");
    }
    known_hand remem_my_hand=my_hand;
    known_hand remem_his_hand=his_hand;
    known_hand the_lay(8);
    the_lay.assign_Window(out_win);
    int last_go=crib;
    do
    { int go=last_go==player?computer:player;
      int com_passed=0;
      int pl_passed=0;
      out_win.clear();
      do
      { int score;
	int tt;
        char lay_title[30];
        if (go==computer)
        { tt=play_card(the_lay,my_hand);
          if (tt==-1)
          { com_passed=1;
	    score=-1;
            computer_win.cursor(3,30);
            disp_puts("GO");
          }
          else
          { the_lay.add_card(my_hand[tt]);
	    score=lay_score(the_lay);
            my_hand.remove_card(my_hand[tt]);
            my_hand.print("Computer hand",1);
  	    last_go=computer;
	    sprintf(lay_title,"The lay: Total=%2d",the_lay.total());
            the_lay.print(lay_title,0);
            add_score(0,score);
          }
        }
        else { 
          do
          { comment_win.clear();
	    disp_puts("Please play a card");
	    tt=his_hand.size()?get_card(his_hand,&the_lay):-1;
            score=play_card(the_lay,his_hand);
            if (tt==-1) 
            { if (score==-1) break; }
            else
              if (his_hand.remove_card(tt))
              { the_lay.add_card(tt);
  		score=lay_score(the_lay);
                if (score!=-1) break;
		his_hand.add_card(tt);
		the_lay.remove_card(tt);
	      }
            comment_win.clear();
  	    disp_puts("You cannot do that");
          } while (1);
  	  if (tt==-1)
  	    pl_passed=1;
          else
  	  { sprintf(lay_title,"The lay: Total=%2d",the_lay.total());
            the_lay.print(lay_title,0);
            his_hand.print("Your hand",0);
            add_score(score,0);
  	    last_go=player;
          }
        }
        if (go==computer && !pl_passed) go=player;
        else
          if (go==player && !com_passed) go=computer;
      } while (the_lay.total()<31 && (!pl_passed || !com_passed));
      if (the_lay.total()<31)
        add_score(last_go==player,last_go==computer);
      if (last_go==player || waiting  || auto_play==forced) pause();
      computer_win.cursor(3,30);
      disp_puts("  ");
      the_lay.return_hand();
    } while (my_hand.size()>0 || his_hand.size()>0);
    out_win.clear();
    comment_win.clear();
    if (crib==player)
    { account_hand(remem_my_hand,"My hand was",computer);
      account_hand(remem_his_hand,"Your hand was",player);
      if (scoring==computer || waiting) pause();
      account_hand(box,"The box was",player);
    }
    else
    { account_hand(remem_his_hand,"Your hand was",player);
      account_hand(remem_my_hand,"My hand was",computer);
      pause();
      account_hand(box,"The box was",computer);
    }
    deck.random_add_card(start);
    if (scoring==computer || waiting) pause();
    crib^=1;
  } while (1);
}
