(* Copyright (C) 1991 Jensen & Partners International *)

{Ŀ
  TopSpeed Example Class Library        
                                        
  Basic Collection Classes              
                                        
  See ClassLib.DOC for further details  
 
}

IMPLEMENTATION UNIT Classes;

Function Alloc(size : Word):Address; Forward; { Checked allocate }

CONST
  DebugErrors = False;  { If true, full error messages are printed }

{Ŀ
  Base Object Class                 
 
}

Implementation Of ObjClass;


  Procedure SubDispose;
  begin
    (* No sub-objects in basic class *)
  end;

  Function Copy{:Obj};
  var o : Obj;
  begin
    o := Alloc(SelfSize);
    Move(SELF,o^,SelfSize);
    Copy := o;
  end;

  Function  SelfSize {: Word};
  Begin
    SelfSize := Size(Self);    { Should be redefined for objects using
                                 multiple inheritance (unless the class
                                 derived from ObjClass is first in the
                                 inherits list }
  End;

  Function  SelfAdr {: Address};
  Begin
    SelfAdr := Adr(Self);      { Should be redefined for objects using
                                 multiple inheritance (unless the class
                                 derived from ObjClass is first in the
                                 inherits list }
  End;

begin
end;

Procedure DoDispose(p : Obj; link : Address);
BEGIN
  p^.SubDispose;
  DISPOSE(p);
END;




{Ŀ 
  Base Container Class              
 
}

Implementation Of ContainerClass;

  Function Empty {:BOOLEAN };
  begin
    Empty := Count=0;
  end;

  Procedure SubDispose;
  begin
    ForEach(DoDispose,Nil);
  end;

  Procedure ForEach {(a : ActionProc;link :Address)};
  begin
    ClassException^.Raise(CE_deferred_foreach);
  end;

  Function  Contains{(e : Obj):Boolean};
  begin
    ClassException^.Raise(CE_deferred_contains);
    Contains := False;
  end;

begin
  Count := 0;
end;

{Ŀ 
  List Class                        
 
}

Implementation Of ListElemClass;
begin
  Next := Nil;
  Prev := Nil;
end;


Implementation Of ListClass;

  Procedure Append  {(e : ListElem)};
  begin
    e^.Prev := Last;
    e^.Next := Nil;
    if Last=Nil then First := e          { previously empty }
                else Last^.Next := e;
    Last := e;
    Count := Count+1;
  end;

  Procedure Insert  {(e : ListElem)};
  begin
    e^.Prev := Nil;
    e^.Next := First;
    if First=Nil then Last := e          { previously empty }
                 else First^.Prev := e;
    First := e;
    Count := Count+1;
  end;

  Procedure InsertBefore {(pos,e: ListElem)};
  begin
    If (pos=Nil)or(pos=First) then
      Insert(e)
    else
    begin
      e^.Next := pos;
      e^.Prev := pos^.Prev;
      pos^.Prev^.Next := e;
      pos^.Prev := e;
      Count := Count+1;
    end;
  end;

  Procedure AppendAfter  {(pos,e: ListElem)};
  begin
    If (pos=Nil)or(pos=Last) then
      Append(e)
    else
    begin
      e^.Prev := pos;
      e^.Next := pos^.Next;
      pos^.Next^.Prev := e;
      pos^.Next := e;
      Count := Count+1;
    end;
  end;

  Procedure Remove  {(e : ListElem)};
  begin
    If e=First then
      First := e^.Next
    else
      e^.Prev^.Next := e^.Next;
    If e=Last then
      Last := e^.Prev
    else
      e^.Next^.Prev := e^.Prev;
    e^.Next := Nil;
    e^.Prev := Nil;
    Count := Count-1;
  end;

  Procedure Sort {(Comp : CompareFunc)};

    Procedure MergeSort ( VAR l1,l2,out : ListClass );
    (* merges l1 and l2 into out
       pre-condition:  out is empty
       post-condition: l1 and l2 are empty
    *)

    var
        exit : Boolean;
        e : ListElem;
    begin
      repeat
        exit := True;
        If l1.Count=0 then
          out.Merge(adr(l2))
        else
          If l2.Count=0 then
            out.Merge(adr(l1))
          else
          begin
            exit := FALSE;
            If Comp(l1.First,l2.First)<0 then
            begin
              e := l1.First;
              l1.Remove(e);
            end
            else
            begin
              e := l2.First;
              l2.Remove(e);
            end;
            out.Append(e);
          end;
      UNTIL exit;

    END;

    Procedure DoSort(var out : ListClass; n : word);
    VAR l1,l2 : ListClass;
        e : ListElem;
    (* moves n elements from self to out sorted *)
    BEGIN
      If n>0 then
      begin
        If n=1 then
        begin
          e := First;
          Remove(e);
          out.Append(e);
        end
        else
        begin
          DoSort(l1,n DIV 2);
          DoSort(l2,(n+1)DIV 2);
          MergeSort(l1,l2,out);
        END;
      END;
    END;

  VAR res : ListClass;
  BEGIN
    DoSort(res,Count);
    Count := res.Count;
    First := res.First;
    Last := res.Last;
  END;

  Procedure ForEach {(a : ActionProc; link : Address)};
  var e,n : ListElem;
  begin
    e := First;
    while e<>Nil do
    begin
      n := e^.Next; (* save in case e changed *)
      a(e,link);
      e := n;
    end;
  End;

  Procedure Merge   {(c : List)};
  var
    f : ListElem;
  Begin
    f := c^.First;
    If f<>Nil then
    begin
      If Last<>Nil then
        Last^.Next := f
      else
        First := f;
      f^.Prev := Last;
      Last := c^.Last;
    end;
    Count := Count+c^.Count;
    c^.Count := 0;
    c^.First := Nil;
    c^.Last := Nil;
  End;

  Function Copy    {: Container};
  var e : ListElem;
      c : List;
      n : ListElem;
  begin
    New(c);
    e := First;
    while e<>Nil do
    begin
      n := e^.Copy::ListElem;
      c^.Append(n);
      e := e^.Next;
    end;
    Copy := c;
  End;

  Function  Contains{(e : Obj):Boolean};
  var t,term : ListElem;
  begin
    if e^ is ListElemClass then
    begin
      t := First;
      term := Nil;
      while t<>term do
      begin
        If t=e::ListElem then term := t;
        t := t^.Next;
      end;
      Contains := (t=Nil);
    end
    else
      Contains := False;
  End;



begin
  First := Nil;
  Last := Nil;
end;



{Ŀ
  Stack Class                       
 
}


Implementation of StackClass;

  Function Pop {: StackElem};
  begin
    Pop := First;
    Remove(First);
  end;

begin
end;

{Ŀ
  Queue Class                       
 
}


Implementation of QueueClass;

  Function Dequeue {: QueueElem};
  begin
    Dequeue := First;
    Remove(First);
  end;

begin
end;

{Ŀ
  Vector Class                      
 
}

Implementation of VectorClass;

  Function Ptr{(index:Word):Address};
  var a : Address;
  Begin
    if (index<Low)or(index>High) then
      ClassException^.Raise(CE_vector_index);
    if _Data=Nil then
    begin
      GetMem(_Data,_VESize*(High-Low+1));
      FillChar(_Data^,_VESize*(High-Low+1),0);
    end;
    a := _Data;
    (a:>word):=(a:>word)+_VESize*(index-Low);
    Ptr := a;
  end;

  Procedure Get{(index:Word; var el)};
  begin
    Move(Ptr(index)^,el,_VESize);
  end;

  Procedure Put{(index:Word; el)};
  begin
    Move(el,Ptr(index)^,_VESize);
  end;

  Procedure ForEach {(a : ActionProc; link:Address)};
  (* Only for vector of objects *)
  var
    op : ^Obj;
    i : Word;
  begin
    if _VEObj then
      for i := Low to High do
      begin
        op := Ptr(i);
        If op^<>Nil then
          a(op^,link);
      end;
  end;

  Function  Copy    {: Container};
  var
    c : Vector;
    op1,op2 : ^Obj;
    i : Word;
  begin
    New(c);
    if _VEObj then
    begin
      c^.Init(Low,High,0);
      for i := Low to High do
      begin
        op1 := Ptr(i);
        If op1^<>Nil then
        begin
          op2 := c^.Ptr(i);
          op2^ := op1^^.Copy;
        end;
      end
    end
    else
    begin
      c^.Init(Low,High,_VESize);
      Move(_Data^,c^.Ptr(Low)^,_VESize*(High-Low+1));
    end;
    Copy := c;
  end;

  Function  Contains{(e : Obj):Boolean};
  var
    i : Word;
    op : ^Obj;
  begin
    if _VEObj then
    begin
      i := Low;
      Contains := False;
      while i<=High do
      begin
        op := Ptr(i);
        If op^=e then
        begin
          i := High;
          Contains := True;
        end;
      end
    end
    else
      Contains := False;
  end;


  Procedure Init{ ( lowi,highi : Word;
                    elemsize : Word)}; { elemsize = 0 if array of objects }
  begin
    Low := lowi;
    High := highi;
    _VEObj := False;
    If elemsize=0 then
    begin
      elemsize := Size(Obj); { e.g. pointer size }
      _VEObj := True;
    end;
    _VESize := elemsize;
    _Data := Nil;
  end;


begin
end;

{Ŀ
  Matrix Class                      
 
}

Implementation of MatrixClass;

  Function Ptr{(row,column:Word):Address};
  Var
    v : Vector;
    vp : ^Vector;
  Begin
    vp := VectorClass.Ptr(row);
    if vp^=Nil then
    begin
      New(vp^);
      vp^^.Init(LowColumn,HighColumn,_MESize);
    end;
    Ptr := vp^^.Ptr(column);
  End;

  Procedure Get{(row,column:Word; var el)};
  begin
    Move(Ptr(row,column)^,el,_MESize);
  end;

  Procedure Put{(row,column:Word; el)};
  begin
    Move(el,Ptr(row,column)^,_MESize);
  end;

  Function  Contains{(e : Obj):Boolean};
  var vp : ^vector;
      i  : Word;
  begin
    if _MEObj then
    begin
      i := LowRow;
      Contains := False;
      while i<=HighRow do
      begin
        vp := VectorClass.Ptr(i);
        if vp^<>Nil then
          if vp^^.Contains(e) then
          begin
            i := High;
            Contains := True;
          end;
      end;
    end
    else
      Contains := False;
  end;

  Function  Copy    {: Container};
  var
    c : Matrix;
    op1,op2 : ^Obj;
    i : Word;
    mes : Word;
  begin
    New(c);
    if _MEObj then mes := 0 else mes := _MEsize;
    c^.Init(LowRow,HighRow,LowColumn,HighColumn,mes);
    for i := Low to High do
    begin
       op1 := VectorClass.Ptr(i);
       If op1^<>Nil then
       begin
          op2 := c^.VectorClass.Ptr(i);
          op2^ := op1^^.Copy;
       end;
    end;
    Copy := c;
  end;

  Procedure Init{( lowr,highr : Word;
                  lowc,highc : Word;
                  elemsize : Word)}; { elemsize = 0 if array of objects }
  begin
    LowColumn := lowc;
    HighColumn := highc;
    LowRow := lowr;
    HighRow := highr;
    VectorClass.Init(lowr,highr,0);
    _MEObj := False;
    If elemsize=0 then
    begin
      elemsize := Size(Obj); { e.g. pointer size }
      _MEObj := True;
    end;
    _MESize := elemsize;
  end;


begin
end;



{Ŀ
  Binary Tree Class                 
 
}

Implementation Of TreeElemClass;

  Function Compare{(k : Key):Int16};
  begin
    ClassException^.Raise(CE_deferred_treeelemclass_compare);
    Compare := 0;
  end;

  Function GetKey{():Key};
  begin
    ClassException^.Raise(CE_deferred_treeelemclass_getkey);
    GetKey := NIL;
  end;

begin
  Left := Nil; Right := Nil;
end;


Var lchild : Boolean;
    parent : TreeElem;

Function  DoFind  (t : Tree;
                   k : Key;
                   VAR parent : TreeElem ) : TreeElem;
var
  p,q,term : TreeElem;
  r        : Int16;
  lp       : Boolean;
begin
  q := Nil;
  lp := False;
  p := t^.Root;
  term := Nil;
  while p<>term do
  begin
    r := p^.Compare(k);
    If r < 0 then
    begin
      q := p;
      p := p^.Left;
      lp := True;
    end
    else
      if r > 0 then
      begin
        q := p;
        p := p^.Right;
        lp := False;
      end
      else
        term := p;
  end;
  parent := q;
  lchild := lp;
  DoFind := p;
end;

Procedure DoTreeMerge(e : Obj; link:Address);
begin
  (e:>TreeElem)^.Right := Nil;
  (e:>TreeElem)^.Left := Nil;
  (link:>Tree)^.Insert(e:>TreeElem);
end;






Implementation Of TreeClass;

  Function Find{(k : Key):TreeElem};
  var
    p,q,term : TreeElem;
    r        : Int16;
    lp       : Boolean;
  begin
    q := Nil;
    lp := False;
    p := Root;
    term := Nil;
    while p<>term do
    begin
      r := p^.Compare(k);
      If r < 0 then
      begin
        q := p;
        p := p^.Left;
        lp := True;
      end
      else
        if r > 0 then
        begin
          q := p;
          p := p^.Right;
          lp := False;
        end
        else
          term := p;
    end;
    parent := q;
    lchild := lp;
    Find := p;
  end;

  Procedure Insert{(e : TreeElem)};
  var
    p : TreeElem;
  begin
    p := Find(e^.GetKey);
    if p<>Nil then
    begin
      e^.Left := p^.Left;
      e^.Right := p^.Right;
      DoDispose(p,Nil)
    end
    else
    begin
      e^.Left := Nil;
      e^.Right := Nil;
      Count := Count+1;
    end;
    if parent=Nil then
      Root := e
    else
      if lchild then
        parent^.Left := e
      else
        parent^.Right := e;
  end;



  Function Delete{(k : Key):TreeElem};
  var
    p,q,r,s : TreeElem;
  begin
    p := Find(k);
    if p<>Nil then
    begin
      if p^.Left=Nil then r := p^.Right
      else
      begin
        r := p^.Left;
        if p^.Right<>Nil then
        begin
          s := r;
          while s^.Right <> Nil do
            s := s^.Right;
          s^.Right := p^.Right;
        end;
      end;
      DoDispose(p,Nil);
      if parent=Nil then
        Root := r
      else
        if lchild then
          parent^.Left := r
        else
          parent^.Right := r;
      Count := Count-1;
    end;
    Delete := p;
  end;


  Procedure ForEach {( a : ActionProc; link:Address )};
    Procedure Traverse(o: TreeElem);
    var
      r: TreeElem;
    begin
      while o <> Nil do
      begin
        r := o^.Right;
        Traverse(o^.Left);
        a(o,link);
        o := r;
      end;
    end;
  begin
    Traverse(Root);
  end;

  Procedure Merge {(c : Tree)};
  var p,term : TreeElem;
      k : Key;
      r : Int16;
  begin
    { A better algorithm is left as an exercise to the user! }
    if Root=Nil then
    begin
      Root := c^.Root;
      Count := c^.Count;
    end
    else
      c^.ForEach(DoTreeMerge,SelfAdr);
    c^.Count := 0;
    c^.Root := Nil;
  end;

  Function  Copy    {: Obj};
  var
    t : Tree;

    Function TraverseCopy(o: TreeElem):TreeElem;
    var
      n: TreeElem;
    begin
      n := o^.Copy:>TreeElem;
      if (o^.Left<>Nil) then n^.Left := TraverseCopy(o^.Left);
      If (o^.Right<>Nil) then n^.Right := TraverseCopy(o^.Right);
      TraverseCopy := n;
    end;

  begin
    New(t);
    t^ := Self;
    t^.Root := TraverseCopy(Root);
    Copy := t;
  end;

  Function  Contains{(e : Obj):Boolean};
  begin
    if e^ is TreeElemClass then
      Contains := (Find((e:>TreeElem)^.GetKey)<>Nil)
    else
      Contains := False;
  End;

begin
  Root := Nil;
end;

{Ŀ 
  Dynamic Hash Table Class          
 
}


CONST
   HTmin     = 16;
   HTmax     = 0FF00H DIV SIZE(ADDRESS);

TYPE
   HashArray = ARRAY[0..HTMax] OF HashElem;
   HashArrayPtr = ^HashArray;

Function NewHashTab(n:word):HashArrayPtr;
VAR
  p : HashArrayPtr;
Begin
  GetMem(p,n*SIZE(Address));
  FillChar(p^,n*SIZE(Address),0);
  NewHashTab := p;
end;


Procedure Rehash(var ht : ADDRESS;var curnum : Word;expand : BOOLEAN);
var
  ot,nt : HashArrayPtr;
  os,ns : Word;
  i,h   : Word;
  p     : HashElem;
label 999;
begin
  ot := ht;
  os := curnum;
  If expand then
  begin
    If os=HTmax then
    begin
      ClassException^.Raise(CE_max_hash_table);
    end
    else if os<800H then
      ns := os*2
    else
      ns := os+800H;
    If ns>HTmax then ns := HTmax
  end
  else
  begin
    If os=HTmin then goto 999;
    IF os>1000H then
      ns := os-800H
    else
      ns := os DIV 2;
    If ns<HTmin then ns := HTmin
  end;
  nt := NewHashTab(ns);
  FOR i := 0 TO os-1 do
  begin
    p := ot^[i];
    If p<>Nil then
    begin
      h := p^.HashVal MOD ns;
      while (nt^[h]<>Nil) do
      begin
        h := h+1; If h=ns then h := 0;
      end;
      nt^[h] := p;
    end;
  end;
  FreeMem(ot,os*SIZE(Address));
  curnum := ns;
  ht := nt;
999:
end;

Implementation Of HashElemClass;

  Function Eq{(k : Key):BOOLEAN};
  begin
    ClassException^.Raise(CE_deferred_hashelemclass_eq);
    Eq := False;
  end;

  Function GetKey{():Key};
  begin
    ClassException^.Raise(CE_deferred_hashelemclass_getkey);
    GetKey := Nil;
  end;

begin
  HashVal := MAX(word);
end;

Procedure DoHashMerge(e : Obj; link:Address);
begin
  (link:>HashTable)^.Insert(e:>HashElem);
end;



Implementation Of HashTableClass;

  Function HashKey{(k : Key): Word};
  begin
    ClassException^.Raise(CE_deferred_hashtableclass_hashkey);
    HashKey := 0;
  end;

  Procedure Insert{(e : HashElem)};
  var h,hv:Word;
      k : Key;
      p : HashElem;
  begin
    If HTPtr=Nil then
    begin
      HTSize := HTmin;
      HTPtr := NewHashTab(HTmin);
    end
    else
    begin
      If Count=(HTSize DIV 4)*3 then (* 75% full *)
        Rehash(HTPtr,HTSize,TRUE);
      hv := e^.HashVal;
      If hv=MAX(Word) then
      begin
        hv := HashKey(e^.GetKey);
        If hv=MAX(Word) then hv := 0 ;
        e^.HashVal := hv;
      end;
      h := hv MOD HTSize;
      k := e^.GetKey;
      repeat
        p := (HTPtr:>HashArrayPtr)^[h];
        If (p=Nil) then
        begin
          Count := Count+1;
          (HTPtr:>HashArrayPtr)^[h] := e;
        end
        else
          If (p^.HashVal=hv) AND p^.Eq(k) then
          begin
            DoDispose(p,Nil);
            (HTPtr:>HashArrayPtr)^[h] := e;
            p := Nil;
          end
          else
          begin
            h := h+1;
            If h=HTSize then h := 0;
          end;
      until p=Nil;
    end;
  end;

  Function Find{(k : Key):HashElemPtr};
  var
    h,hv : Word;
    p,q,term   : HashElem;
  begin
    p := Nil;
    If HTPtr<>Nil then
    begin
      hv := HashKey(k);
      h := hv MOD HTSize;
      term := Nil;
      repeat
        p := (HTPtr:>HashArrayPtr)^[h];
        If p<>Nil then
        begin
          If (p^.HashVal=hv) AND p^.Eq(k) then
            term := p
          else
          begin
            h := h+1;
            If h=htsize then h := 0;
          end;
        end;
      until term=p;
    end;
    Find := p;
  end;

  Function Delete{(k : Key):HashElemPtr};
  var
    h,hv,nh : Word;
    p,q,term : HashElem;
  begin
    p := Nil;
    If HTPtr<>Nil then
    begin
      hv := HashKey(k);
      h := hv MOD HTSize;
      term := Nil;
      repeat
        p := (HTPtr:>HashArrayPtr)^[h];
        If p<>Nil then
        begin
          If (p^.HashVal=hv) AND p^.Eq(k) then
            term := p
          else
          begin
            h := h+1;
            If h=HTSize then h := 0;
          end;
        end;
      until term=p;
      if p<>Nil then
      begin
        repeat
          nh := h+1;
          If nh=HTSize then nh := 0;
          q := (HTPtr:>HashArrayPtr)^[nh];
          If (q=Nil)then
          else
            if hv=q^.HashVal then
            begin
              (HTPtr:>HashArrayPtr)^[h] := q;
              h := nh;
            End;
        until q=Nil;
        (HTPtr:>HashArrayPtr)^[h] := Nil;
        Count := Count-1;
        If (Count=(HTSize DIV 4)) then (* 25% full *)
          Rehash(HTPtr,HTSize,FALSE);
      end;
    end;
    Delete := p;
  end;

  Procedure ForEach {( a : ActionProc; link : Address )};
  var i,j : Word;
      p : HashElem;
  begin
    i := 0; j := Count;
    while j<>0 do
    begin
      p := (HTPtr:>HashArrayPtr)^[i];
      If p<>Nil then
      begin
        a(p,link);
        j := j-1;
      end;
      i :=i+1;
    end;
  end;

  Procedure SubDispose;
  begin
    ContainerClass.SubDispose;
    IF HTPtr<>Nil then
    begin
      FreeMem(HTPtr,HTSize*SIZE(Address));
      HTPtr := Nil;
    end;
  end;

  Procedure Merge   {(c : HashTable)};
  begin
    c^.ForEach(DoHashMerge,SelfAdr);
    c^.Count := 0;
    c^.SubDispose;
  end;

  Function  Copy    {: Obj};
  var
    h : HashTable;
    i,j : Word;
    p : HashElem;
  begin
    New(h);
    h^ := Self;
    h^.HTPtr := NewHashTab(HTsize);
    i := 0; j := Count;
    while j<>0 do
    begin
      p := (HTPtr:>HashArrayPtr)^[i];
      If p<>Nil then
      begin
        (h^.HTPtr:>HashArrayPtr)^[i] := p^.Copy::HashElem;
        j := j-1;
      end;
      i :=i+1;
    end;
    Copy := h;
  end;

  Function  Contains{(e : Obj):Boolean};
  begin
    if e^ is HashElemClass then
      Contains := (Find((e:>HashElem)^.GetKey)<>Nil)
    else
      Contains := False;
  End;


begin
  HTPtr := Nil;
end;


Var
  OldHeapError : HeapErrorType;

Function HeapErrorHandler(size : word): int16;
Begin
  If OldHeapError(size)<>2 then
    ClassException^.Raise(CE_heap_full);
  HeapErrorHandler := 2;
End;

Function Alloc{(size : Word):Address};
(* Safe memory allocation *)
Var
  a : Address;
Begin
  OldHeapError := HeapError;
  HeapError := HeapErrorHandler;
  GetMem(a,size);
  HeapError := OldHeapError;
  Alloc := a;
End;

{Ŀ
  Exception Handler Class           
 
}
Implementation of ExceptionClass;

  Procedure ChainTo{(e : Exception)};
  Begin
    first := e^.first;
    next := Nil;
    While e^.next<>Nil do e := e^.next;
    e^.next := Adr(Self);
  End;

  Function Raise{(err:Word):Boolean};
  Begin
    code := err;
    If (next=Nil)or not next^.Raise(err) then
      Raise := Handler
    else
      Raise := True;
  End;

  Procedure Unchain;
  Var
    p : Exception;
  Begin
    p := first;
    While p^.next<>Adr(Self) do p := p^.next;
    p^.next := next;
    next := Nil;
    first := Adr(Self);
  End;

  Function  Handler{:Boolean};
  Begin
    WriteLn('Exception: ',code,', Program terminated');
    Halt;
    Handler := True;
  End;

begin
  first := Adr(Self);
  next := Nil;
  code := 0;
end;

Class LocalExceptionClass Inherits ExceptionClass;
  Function   Handler{:Boolean};
End;

Implementation of LocalExceptionClass;

{Ŀ
  Local Exception Handler           
 
}


  Function  Handler{:Boolean};
  Begin
    WriteLn('Exception: ',code,' Module: Classes.pas ');
{%T DebugErrors}
    Case code of
      CE_ok:
      CE_heap_full:                         Write('Heap Full');
      CE_deferred_foreach:                  Write('ForEach not implemented');
      CE_deferred_contains:                 Write('Contains not implemented');
      CE_vector_index:                      Write('Vector index error');
      CE_deferred_treeelemclass_compare:    Write('TreeElemClass.Compare not implemented');
      CE_deferred_treeelemclass_getkey:     Write('TreeElemClass.GetKey not implemented');
      CE_max_hash_table:                    Write('Maximum Hash table size reached');
      CE_deferred_hashelemclass_eq:         Write('HashElemClass.Eq not implemented');
      CE_deferred_hashelemclass_getkey:     Write('HashElemClass.GetKey not implemented');
      CE_deferred_hashtableclass_hashkey:   Write('HashTableClass.HashKey not implemented');
    End;
    WriteLn;
{%E}
    Halt;
    Handler := True;
  End;

begin
end;

VAR
  LEC : LocalExceptionClass;

begin
  ClassException := Adr(LEC);
end.

