PROGRAM COLUMNP(OUTPUT);

IMPORT PASWIN,PASDOS,HIGHSCRP;

CONST
  InsertLocked=080H;
  CapsLocked=040H;
  NumLocked=020H;
  ScrollLocked=010H;
  LoopIter=20;

TYPE
(*# save*)
(*# data(near_ptr => off) *)
  bp=^BYTE;
(*# restore *)
  delaytype=(Short_delay,Medium_delay,Long_delay);

CLASS Keystate;
  OldStatus:BYTE;
  STATIC PROCEDURE SetUp(Value:BYTE);
  STATIC PROCEDURE Reset;
END;

IMPLEMENTATION OF Keystate;
  PROCEDURE SetUp;
    VAR
      mp:bp;
    BEGIN
(*%F _OS2*)
      mp:=0400017H:>bp;
      OldStatus:=mp^;
      mp^:=Value
(*%E*)
    END;

  PROCEDURE Reset;
    VAR
      mp:bp;
    BEGIN
(*%F _OS2*)
      mp:=0400017H:>bp;
      mp^:=OldStatus
(*%E*)
    END;

  BEGIN
  END;
VAR
  KeyBoard:KeyState;

CLASS Delay;
  lengths,minimums:ARRAY[delaytype] OF INT16;
  STATIC PROCEDURE SetUp(m1,l1,m2,l2,m3,l3:INT16);
  STATIC PROCEDURE TimesPC(Percent:INT16);
  STATIC PROCEDURE DoDelay(x:delaytype);
  STATIC PROCEDURE CheckRange;  { Do not call me }
END;

IMPLEMENTATION OF DELAY;
  PROCEDURE SetUp;
    BEGIN
      lengths[short_delay]:=l3;
      lengths[medium_delay]:=l2;
      lengths[long_delay]:=l1;
      minimums[short_delay]:=m3;
      minimums[medium_delay]:=m2;
      minimums[long_delay]:=m1;
    END;

  PROCEDURE CheckRange;
    VAR
      I:delaytype;
    BEGIN
      FOR I:=short_delay TO long_delay DO
        IF lengths[i]<minimums[i] THEN
          lengths[i]:=minimums[i]
    END;

  PROCEDURE TimesPC;
    VAR
      I:delaytype;
    BEGIN
      FOR I:=short_delay TO long_delay DO
        lengths[I]:=(lengths[I]::integer*percent) div 100;
      CheckRange;
    END;

  PROCEDURE DoDelay;
    BEGIN
      PasDos.Delay(lengths[x])
    END;

  BEGIN
  END;

VAR
  Wait:Delay;

TYPE
  DisplayedIntP=^DisplayedInt;

CLASS DisplayedInt;
  Value:INT16;
  Win:PasWin.PopUpWin;
  Next:DisplayedIntP;
  STATIC PROCEDURE Activate;
  STATIC PROCEDURE SetUp(Title:string[high];foreG,backG:PasWin.Color;x,y:INT16);
  STATIC PROCEDURE Kill;
END;

VAR
  ActivityChain:DisplayedIntP;

IMPLEMENTATION OF DisplayedInt;
  PROCEDURE SetUp;
    BEGIN
      WITH Win DO BEGIN
        Open(x,y,10,1,foreG,backG);
        FrameDraw(PasWin.DoubleFrame,foreG,backG,false);
        FrameWrite(Title,true,1,foreG,backG);
        CursorIsOn:=FALSE;
        PopUp;
      END;
      Next:=ActivityChain;
      ActivityChain:=ADR(SELF);
      Value:=0;
    END;

  PROCEDURE Activate;
    BEGIN
      WITH Win DO BEGIN
        MakeFocus;
        Clear(PasWin.clear_screen);
        PasWin.GoToXY(1,1);
        WRITE(Value:9);
        DeFocus
      END;
      IF next<>NIL THEN
        next^.Activate
    END;

  PROCEDURE Kill;
    BEGIN
      Win.Close;
      IF next<>NIL THEN BEGIN
        next^.Kill;
        next:=NIL
      END
    END;

  BEGIN
  END;

VAR
  Score,Blocks,Level:DisplayedInt;

CLASS TILE;
  data:ARRAY [0..2] OF PasWin.Color;
  STATIC FUNCTION NewTile:PasWin.Color;  { Do not call me }
  STATIC PROCEDURE RFill;
  STATIC PROCEDURE SetUp(x,y,z:PasWin.Color);
  STATIC PROCEDURE SetUpUnknown;
  STATIC PROCEDURE Assign(from:TILE);
  STATIC PROCEDURE Rotate;
  STATIC FUNCTION Score:BOOLEAN;
END;

IMPLEMENTATION OF TILE;

  FUNCTION NewTile;
    VAR
      t:PasWin.Color;
    BEGIN
      t:=(1+RANDOM(6)):>PasWin.Color;
      IF (t=PasWin.Brown) THEN
        t:=PasWin.Yellow
      ELSE
        IF (t=PasWin.Cyan) THEN
          t:=PasWin.LightCyan;
      NewTile:=t;
    END;

  PROCEDURE RFill;
    BEGIN
      data[0]:=NewTile;
      data[1]:=NewTile;
      data[2]:=NewTile
    END;

  PROCEDURE SetUp;
    BEGIN
      data[0]:=x;
      data[1]:=y;
      data[2]:=z
    END;

  PROCEDURE SetUpUnknown;
    BEGIN
      RFill
    END;

  PROCEDURE Assign;
    BEGIN
      data:=from.data
    END;

  PROCEDURE Rotate;
    VAR
      T:PasWin.Color;
    BEGIN
      t:=data[0];
      data[0]:=data[1];
      data[1]:=data[2];
      data[2]:=t
    END;

  FUNCTION Score;
    BEGIN
      IF data[0]=PasWin.LightGray THEN
        Score:=FALSE
      ELSE
        Score:=(data[0]=data[1]) AND (data[1]=data[2])
    END;
  BEGIN
  END;

TYPE
  OutEvent=(touched_bottom,full,nothing);
  InEvent=(rotate,left,right,drop,fall,nout);

CLASS Grid;
  { private: ie keep paws off }
  Table:ARRAY [0..5] OF ARRAY [0..17] OF PasWin.Color;
  CurTile:Tile;
  CurCol,CurY:INT16;
  LastDrop:BYTE;
  Win:PasWin.Win;
  STATIC PROCEDURE PlaceBlock(x,y:BYTE;c:PasWin.Color);
  STATIC PROCEDURE PlaceTile(e:InEvent);
  STATIC PROCEDURE Construct;
  { public: }
  STATIC FUNCTION Enter:OutEvent;
  STATIC FUNCTION Drop1(E:InEvent):OutEvent;
  STATIC FUNCTION Reduce:BYTE;
END;

IMPLEMENTATION OF Grid;

  PROCEDURE PlaceBlock;
    BEGIN
      WITH Win DO BEGIN
        MakeFocus;
        PasWin.GoToXY(2*x+1,18-y);
        PasWin.foreground(c);
        WRITE(0DBH:>CHAR,0DBH:>CHAR);
        table[x,y]:=c;
        DeFocus
      END
    END;

  PROCEDURE PlaceTile;
    VAR
      I,Col:BYTE;
    BEGIN
      IF e=fall THEN
        placeblock(curcol,cury+3,PasWin.LightGray)
      ELSE
        IF (e=left) OR (e=right) THEN BEGIN
          IF e=left THEN
            Col:=CurCol+1
          ELSE
            Col:=CurCol-1;
          FOR I:=0 TO 2 DO
            placeblock(Col,CurY+I,PasWin.LightGray)
        END
        ELSE
          IF e=drop THEN
            FOR I:=CurY+3 TO 17 DO
              placeblock(CurCol,i,PasWin.LightGray);
      FOR I:=0 TO 2 DO
        placeblock(CurCol,CurY+I,CurTile.data[I])
    END;

  FUNCTION Enter;
    BEGIN
      CurTile.RFill;
      CurCol:=3;
      CurY:=15;
      IF table[3][15]=PasWin.LightGray THEN
        Enter:=nothing
      ELSE
        Enter:=full;
      placetile(nout)
    END;

  FUNCTION Drop1;
    BEGIN
      Drop1:=nothing;
      IF ((e<>left) OR (CurCol<>0) AND (table[CurCol-1][CurY]=PasWin.LightGray))
        AND ((e<>right) OR (CurCol<>5) AND (table[CurCol+1][CurY]=PasWin.LightGray)) THEN
        BEGIN
          IF e=rotate THEN
            CurTile.Rotate;
          lastdrop:=0;
          IF e=drop THEN BEGIN
            WHILE (CurY<>0) AND (table[CurCol][CurY-1]=PasWin.LightGray) DO BEGIN
              CurY:=CurY-1;
              LastDrop:=LastDrop+1
            END;
            Drop1:=touched_bottom
          END
          ELSE
            IF e=fall THEN
              CurY:=CurY-1;
          IF e=left THEN
            CurCol:=CurCol-1
          ELSE
            IF e=right THEN
              CurCol:=CurCol+1;
          IF (e=fall) AND ((CurY<0) OR (table[CurCol][CurY]<>PasWin.LightGray)) THEN
            Drop1:=touched_bottom
          ELSE
            placetile(e);
        END
    END;

  FUNCTION Reduce;
    LABEL 1;
    VAR
      Spare:ARRAY[0..5,0..17] OF BOOLEAN;
      CY,I,J:INT16;
      TTile:TILE;
      total:byte;
    BEGIN
      FOR I:=0 TO 5 DO
        FOR J:=0 TO 17 DO
          Spare[I,J]:=FALSE;
      Wait.DoDelay(medium_delay);
      FOR I:=0 TO 17 DO
        FOR J:=0 TO 3 DO BEGIN
          TTile.SetUp(table[j][i],table[j+1,i],table[j+2,i]);
          IF TTile.Score THEN BEGIN
            spare[j][i]:=true;
            spare[j+1][i]:=true;
            spare[j+2][i]:=true
          END
        END;
      FOR I:=0 TO 15 DO
        FOR J:=0 TO 5 DO BEGIN
          TTile.SetUp(table[j][i],table[j,i+1],table[j,i+2]);
          IF TTile.Score THEN BEGIN
            spare[j][i]:=true;
            spare[j][i+1]:=true;
            spare[j][i+2]:=true
          END
        END;
      FOR I:=0 TO 15 DO
        FOR J:=0 TO 3 DO BEGIN
          TTile.SetUp(table[j][i],table[j+1,i+1],table[j+2,i+2]);
          IF TTile.Score THEN BEGIN
            spare[j][i]:=true;
            spare[j+1][i+1]:=true;
            spare[j+2][i+2]:=true
          END
        END;
      FOR I:=0 TO 15 DO
        FOR J:=2 TO 5 DO BEGIN
          TTile.SetUp(table[j][i],table[j-1,i+1],table[j-2,i+2]);
          IF TTile.Score THEN BEGIN
            spare[j][i]:=true;
            spare[j-1][i+1]:=true;
            spare[j-2][i+2]:=true
          END
        END;
      FOR I:=0 TO 17 DO
        FOR J:=0 TO 5 DO
          IF spare[j,i] THEN
            placeblock(j,i,PasWin.Black);
      Wait.DoDelay(medium_delay);
      total:=0;
      FOR J:=0 TO 5 DO BEGIN
        CY:=0;
        I:=0;
        WHILE I<18 DO BEGIN
          WHILE (table[j][i]=PasWin.Black) DO BEGIN
            total:=total+1;
            i:=i+1
          END;
          IF table[j][i]=PasWin.LightGray THEN
            GOTO 1;
          IF CY<>I THEN
            placeblock(j,CY,table[j][i]);
          CY:=CY+1;
          I:=I+1
        END;
        1:
        WHILE (CY<18) AND (table[j][CY]<>PasWin.LightGray) DO BEGIN
          placeblock(j,CY,PasWin.LightGray);
          CY:=CY+1
        END
      END;
      IF total>0 THEN
        Reduce:=total+Reduce
      ELSE
        Reduce:=0
    END;

  PROCEDURE Construct;
    VAR
      I,J:BYTE;
    BEGIN
      WITH Win DO BEGIN
        Open(34,1,12,18,PasWin.LightGray,PasWin.LightGray);
        PasWin.CursorOff;
        WrapOn:=FALSE
      END;
      FOR I:=0 TO 5 DO
        FOR J:=0 TO 17 DO
          Table[I,J]:=PasWin.LightGray;
    END;

  BEGIN
  END;

  VAR Board:Grid;

  PROCEDURE MainLoop;
    LABEL 999;
    VAR
      oe:OutEvent;
      ev:InEvent;
      w,GOT:BYTE;
      key:INT16;
    FUNCTION FullKey:INT16;
      VAR
        A:CHAR;
      BEGIN
        A:=PasDos.RdKey;
        IF A=0:>CHAR THEN
          FullKey:=ORD(PasDos.RdKey)*256
        ELSE
          FullKey:=ORD(A)
      END;

    BEGIN
      oe:=Board.Enter;
      WHILE oe<>full DO BEGIN
        ev:=fall;
        IF w>LoopIter-1 THEN
          w:=0;
        WHILE (w<LoopIter) AND NOT PasDos.KeyPressed DO BEGIN
          Wait.dodelay(short_delay);
          w:=w+1
        END;
        IF w<LoopIter THEN BEGIN
          key:=FullKey;
          CASE Key OF
            04B00H:
              ev:=left;
            04D00H:
              ev:=right;
            04C00H,04800H:
              ev:=rotate;
            05000H:
              ev:=drop;
            000071H:  {Q Key}
              GOTO 999;
            ELSE
          END;
        END;
        IF (Board.Drop1(ev)=touched_bottom) THEN BEGIN
          Got:=Board.Reduce;
          Blocks.Value:=Blocks.Value+got;
          Score.Value:=Score.Value+got*(1+Board.LastDrop DIV 5)*(1+Level.Value DIV 3);
          IF Blocks.Value DIV 50 > Level.Value THEN BEGIN
            Level.Value:=Blocks.Value DIV 50;
            Wait.TimesPC(90)
          END;
          ActivityChain^.Activate;
          oe:=Board.Enter
        END
      END;
      999:
    END;

PROCEDURE RulesScreen;
  VAR
    W:PasWin.FramedWin;
  BEGIN
    W.Open(2,2,19,9,PasWin.LightGray,PasWin.Blue);
    PasWin.CursorOff;
    W.FrameDraw(PasWin.DoubleFrame,PasWin.LightGray,PasWin.Blue,false);
    WRITELN('Get as many lines');
    WRITELN('of three colours asyou can before');
    WRITELN('filling the grid');
    WRITELN;
    WRITELN(CHR(01AH)+CHR(01BH)+'  Move tile');
    WRITELN(CHR(019H)+'   Drop Tile');
    WRITELN(CHR(018H)+'   Rotate Colours');
    WRITE('Q   Quit Game');
    W.DeFocus;
  END;

VAR WholeScreen:PasWin.PopUpWin;
    winners:HIGHSCRP.HighScoreTable;

BEGIN
  WholeScreen.Open(1,1,78,23,PasWin.White,PasWin.Black);
  WholeScreen.FrameDraw(PasWin.DOUBLEFRAME,PasWin.White,PasWin.Black,FALSE);
  PasWin.CursorOff;
  Board.Construct;
  RulesScreen;
  ActivityChain:=NIL;
  Score.SetUp('Score',PasWin.White,PasWin.Green,65,5);
  Blocks.SetUp('Blocks',PasWin.White,PasWin.Red,65,11);
  Level.SetUp('Level',PasWin.Black,PasWin.Cyan,65,17);
  Wait.SetUp(0,30000 div LoopIter,0,5000 div LoopIter,0,1000 div LoopIter);
  KeyBoard.SetUp(0);
  randomize;
  winners.SetUp('column.hig');
  winners.Display;
  MainLoop;
  winners.AddValue(Score.Value);
  winners.Display;
  winners.Close;
  WholeScreen.Close;
  PasWin.CursorOn;
  KeyBoard.Reset;
END.
