/* JPI TopSpeed C demo: Contour curves.            HUR 2/11/89 */
/*                                                             */
/* COPYRIGHT (C) 1989 Jensen and Partners. All Rights Reserved */

#define _JPI_WIN_
#include <stdlib.h>
#include <graph.h>
#include <conio.h>
#include <math.h>
#include <choose.h>

#define LENGTH  30
#define WIDTH   (3*LENGTH/2)

#pragma data ( seg_name => null )   /* put static data in DGROUP */

extern struct videoconfig cv;

static int     Step, Width, Length, Cuts;
static int     Height[WIDTH][LENGTH];
static double  Value [WIDTH][LENGTH];


static double F1 ( double X, double Y )
{
  return ( sin((X*X+Y*Y)/80.0) );
}

static double F2 ( double X, double Y )
{
  return ( sin(X/Y)*cos(Y/5.0) );
}

static double F3 ( double X, double Y )
{
  return ( sin(2.0*cos((X+Y)/5.0)*Y/5.0) );
}

static double F4 ( double X, double Y )
{
  return ( sin(X/2.0)/(1.1+cos((Y-X)/2.0)) );
}

static void Compute ( double (*F) ( double, double) )
{
  int i, j;
  double L,V,U;

  _clearscreen ( _GCLEARSCREEN );
  L = +9.9e100;
  U = -9.9e100;
  _setcolor ( cv.numcolors-1 );
  for ( i = 0 ; i < Width; ++i )
    for ( j = 0; j < Length; ++j )
    {
      V = F ( (double)(i+1), (double)(j+1) );
      Value[i][j] = V;
      if (L > V) L = V;
      if (U < V) U = V;
      _setpixel ( i * Step, j * Step );
      if (kbhit()) return;
    }
  _setcolor ( 0 );
  for ( i = 0 ; i < Width; ++i )
    for ( j = 0; j < Length; ++j )
    {
      Height[i][j] = (int)( 100.0 * (Value[i][j] - L) / (U-L+1e-10) );
      _setpixel ( i * Step, j * Step );
    }
}

static void Draw ( int x, int y )
{
  if (Cuts == 0) _moveto ( x, y );
  else           _lineto ( x, y );
  ++Cuts;
}

#define Median(h,v1,v2,c1,c2) ( v1==v2 ? -1 : (c1*(h-v2)+c2*(v1-h))/(v1-v2) )

static void Square ( int i, int j, int Dist )
{
  int a,b,c,d,l,u,h,x1,y1,x2,y2,m;

  a = Height[i  ][j  ];
  b = Height[i+1][j  ];
  c = Height[i+1][j+1];
  d = Height[i  ][j+1];

  l = u = a;
  if (l > b) l = b; else if (u < b) u = b;
  if (l > c) l = c; else if (u < c) u = c;
  if (l > d) l = d; else if (u < d) u = d;

  h = (l + Dist - 1) / Dist * Dist;
  while (h <= u)
  {
    _setcolor ( 1 + (h / Dist) % (cv.numcolors-1) );
    x1 = i * Step; x2 = x1 + Step;
    y1 = j * Step; y2 = y1 + Step;
    Cuts = 0;

    m = Median ( h, a, b, x1, x2 );
    if (x1 <= m && m <= x2) Draw ( m, y1 );

    m = Median ( h, b, c, y1, y2 );
    if (y1 <= m && m <= y2) Draw ( x2, m );

    m = Median ( h, d, c, x1, x2 );
    if (x1 <= m && m <= x2) Draw ( m, y2 );

    m = Median ( h, a, d, y1, y2 );
    if (y1 <= m && m <= y2) Draw ( x1, m );

    h += Dist;
  }
}


static void Contours ( int Dist )
{
  int i,j;

  _setcolor ( 0 );
  _rectangle ( _GBORDER, 0, 0, Width*Step+1, 0 );
  for ( j = 0; j < Length-1; ++j )
  {
    _setcolor ( 0 );
    _rectangle ( _GFILLINTERIOR, 0, j*Step+1, Width*Step+1, (j+1)*Step+1 );
    for ( i = 0 ; i < Width-1; ++i )
      Square ( i, j, Dist );
  }
}


void contour_demo ()
{
  int d,n;

  clrscr();
  cputs ( "    This routine displays contour curves of various\n" );
  cputs ( "    trigonometric functions.\n\n" );
  cputs ( "    Press a key to start...<ESC> to stop\n" );
  getch();

  Choose ();

  randomize ();
  Step = cv.numypixels / LENGTH;
  do
  {
    ++Step;
    Width = (cv.numxpixels-1) / Step;
  }
  while (Width > WIDTH);
  Length = (cv.numypixels-1) / Step;

  for ( n = 0; ; ++n )
  {
    switch ( n % 4 )
    {
      case 0: Compute ( F1 ); break;
      case 1: Compute ( F2 ); break;
      case 2: Compute ( F3 ); break;
      case 3: Compute ( F4 ); break;
    }
    for ( d = 30; d > 4; d -= 3 )
    {
      if (kbhit()) break;
      Contours ( d );
      delay ( 500 );
    }
    for ( d = 0; d < 9 && !kbhit(); ++d )
      delay ( 500 );
    if (kbhit() && getch() == 27) break;
  }
  _setvideomode ( _DEFAULTMODE );
}
