/* JPI TopSpeed C demo: Space Filling Curves.     HUR 26/10/89 */
/*                                                             */
/* COPYRIGHT (C) 1989 Jensen and Partners. All Rights Reserved */

#define _JPI_WIN_
#include <stdlib.h>
#include <graph.h>
#include <conio.h>
#include <mem.h>
#include <choose.h>

#define MAX        5
#define FALSE      0
#define TRUE       1

#define loop  while(1)

int dX[4] = { +1, 0,-1, 0 };
int dY[4] = {  0,+1, 0,-1 };

int Xlat[4][4] = { { 0,1,2,3 },
                   { 1,0,3,2 },
                   { 2,3,0,1 },
                   { 3,2,1,0 } };

static short Free [MAX+2][MAX+2];

struct { int X,Y; int D,T; } Path [MAX*MAX];

static int Side,Length,Step,Order,X,Y;

extern struct videoconfig cv;

#define XC(X) (X*cv.numypixels/25)
#define YC(Y) ((MAX-Y+1)*cv.numypixels/25)

static void Line ( int L, int C )
{
  _setcolor ( C );
  _moveto ( XC(Path[L-1].X), YC(Path[L-1].Y) );
  _lineto ( XC(Path[L  ].X), YC(Path[L  ].Y) );
}

static void NewSquare ()
{
  int I,J;

  memset ( Free, FALSE, sizeof(Free) );
  for (I = 1; I <= Side; ++I)
    for (J = 1; J <= Side; ++J)
      Free[I][J] = TRUE;
  Length = 0;
  Path[0].X = 1;
  Path[0].Y = 1;
  Free[1][1] = FALSE;
}

static int Try ( int D )
{
  int X,Y;

  X = Path[Length].X;
  Y = Path[Length].Y;
  if (X != Side || Y != 1)
  {
    X += dX[D];
    Y += dY[D];
    if (Free[X][Y])
    {
      Free[X][Y] = FALSE;
      Path[Length].D = D;
      ++Length;
      Path[Length].X = X;
      Path[Length].Y = Y;
      Line ( Length, cv.numcolors-1 );
      return TRUE;
    }
  }
  return FALSE;
}

static void UnDo ()
{
  Free[Path[Length].X][Path[Length].Y] = TRUE;
  Line ( Length, 0 );
  --Length;
}

static void Search ()
{
  int D;

  D = -1;
  loop
  {
    loop
    {
      ++D;
      if (D == 4) break;
      if (Try(D)) return;
    }
    if (Length == 0) return;
    UnDo ();
    D = Path[Length].D;
  }
}

static void NextShape ()
{
  int I;

  for (I = 1; I <= Length; ++I)
    Line ( I, cv.numcolors-1 );
  do
  {
    Search ();
    if (Length == 0) break;
  }
  while (Path[Length].X != Side || Path[Length].Y != 1);
}

static void Prepare ()
{
  int I,B,T;

  Path[Length].D = 0;
  B = TRUE;
  for (I = 0; I <= Length; ++I)
  {
    T = Path[I].D;
    if ((T < 2) != B)
    {
      T = 3-T;
      B = !B;
    }
    Path[I].T = T;
  }
}

static void Display ( int Dir )
{
  int I,D;

  --Order;
  for (I = 0; I < Length; ++I)
  {
    if (Order) Display ( Xlat[Dir][Path[I].T] );
    D = Xlat[Dir][Path[I].D];
    X += Step * dX[D];
    Y -= Step * dY[D];
    _lineto ( X, Y );
  }
  if (Order) Display ( Xlat[Dir][Path[Length].T] );
  ++Order;
}

static void Draw ()
{ int L,M,C;

  Step = 3 + random(4)*random(5);
  M = (cv.numypixels-Step)/Side;
  Order = 0;
  for (L = Step; L < M; L *= Side) ++Order;
  L -= Step;
  X = cv.numxpixels-cv.numypixels;
  Y = cv.numypixels-Step-(cv.numypixels-L)/2;
  C = 1+random(cv.numcolors-1);
  _setcolor( C );
  _moveto ( X, Y );
  Display ( 0 );
  if (random(3) == 0)
  {
    _lineto ( X  , Y+Step );
    _lineto ( X-L, Y+Step );
    _lineto ( X-L, Y      );
    _setcolor ( 1+random(cv.numcolors-1) );
    _floodfill ( X-1, Y+1, C );
  }
}

void curve_demo()
{
  char C;

  clrscr();
  cputs ( "    The left corner shows a basic shape which is\n" );
  cputs ( "    replicated to form the patterns to the right.\n\n" );
  cputs ( "    Use <SPACE> to stop at each pattern.\n" );
  cputs ( "    Use any other key to resume free run.\n\n" );
  cputs ( "    Press a key to start...<ESC> to stop\n" );
  C = getch();

  Choose ();

  randomize ();

  Side = 2;
  NewSquare ();
  do
  {
    _clearscreen ( _GCLEARSCREEN );
    NextShape ();
    if (Length == 0)
      if (Side == MAX)
        break;
      else
      {
        ++Side;
        NewSquare ();
        _clearscreen ( _GCLEARSCREEN );
        NextShape ();
      }
    Prepare ();
    Draw ();
    if (C != ' ') delay ( 1000 );
    if (kbhit())  C = getch ();
    if (C == ' ') C = getch ();
  } while (C != 27);

  _setvideomode(_DEFAULTMODE);
}
