/*
Routines to emulate the Zortech disp library.
Note that although all the main routines are emulate some of the more
esoteric are not, specifically 43 line mode. Note also that this module assumes
the recomendation of the Zortech manual and will work only for text modes.
Currently the 'hidden but documented' global variables do not exist
*/

/*
This emulation uses the JPI Topspeed windowing module as a basis. Note that
this is a very unfair thing to do to it and that one should aim to stop using
this module as soon as possible and move to useing the Topspeed system in
native mode
*/
#include "disp.hpp"
#include <dos.h>
#include <mem.h>
/* The following are dirty hooks into the window library. DO NOT USE */
extern "C"
{
#pragma save
#pragma call(near_call=>on,seg_name=>null)
void        _redrawsection(wintype win, abscoord X1, abscoord Y1, abscoord X2, abscoord Y2);
#pragma restore
}

unsigned disp_inited=0;
static union REGS r;
unsigned char CurrentAttr=0x70; // White on black default

void ioint()
{ int86(0x10,&r,&r);
}

int disp_getmode(void)
{ r.h.ah=0xf;
  ioint();
  return r.h.al;
}

int disp_getattr()
{ return CurrentAttr;
}

void disp_open(void)
{ //Automatic
  disp_inited=1;
}

void disp_close()
{ disp_inited=0;
}

void disp_usebios()
{
}

void disp_flush()
{
}

void disp_setmode(unsigned char)
{
}

void disp_pokew(int row,int col,unsigned attrchar)
{ _fullscreen->Buffer[row*80+col]=attrchar;
  _redrawsection(_fullscreen,col,row,col+1,row+1);
}

unsigned disp_peekw(int row,int col)
{ return _fullscreen->Buffer[row*80+col];
}

void disp_box(int style,int attr, unsigned trow, unsigned lcol,
		unsigned brow, unsigned rcol)
{   unsigned i;
    unsigned attchar;
    static struct
    {	unsigned char ul,ur,ll,lr;	/* upper left, etc.		*/
	unsigned char vert,horz;	/* vertical and horizontal segments */
    } *b,boxchar[] =
    {
	{ 0xC9,0xBB,0xC8,0xBC,0xBA,0xCD },	/* double line		*/
	{ 0xDA,0xBF,0xC0,0xD9,0xB3,0xC4 },	/* single line		*/
	{ 0xDB,0xDB,0xDB,0xDB,0xDB,0xDB },	/* solid box		*/
	{ 213, 184, 212, 190, 179, 205  },	/* double hor, single ver */
	{ 214, 183, 211, 189, 186, 196  },	/* single hor, double ver */
    };

    b = &boxchar[style];
    attr *= 256;

    /* Do corners	*/
    disp_pokew(trow,lcol,attr + b->ul);
    disp_pokew(trow,rcol,attr + b->ur);
    disp_pokew(brow,lcol,attr + b->ll);
    disp_pokew(brow,rcol,attr + b->lr);

    /* Do horizontal segments	*/
    attchar = attr + b->horz;
    for (i = lcol + 1; i < rcol; i++)
    {	disp_pokew(trow,i,attchar);
	disp_pokew(brow,i,attchar);
    }

    /* Do vertical segments	*/
    attchar = attr + b->vert;
    for (i = trow + 1; i < brow; i++)
    {	disp_pokew(i,lcol,attchar);
	disp_pokew(i,rcol,attchar);
    }
}

void disp_move(int row,int col)
{ gotoxy(col+1,row+1);
}

void disp_eeop(void)
{ int cx=wherex();
  int cy=wherey();
  clreol();
  for (int i=cy+1;i<26;i++)
  { gotoxy(1,i);
    clreol();
  }
  gotoxy(cx,cy);
}

void disp_startstand(void)
{ disp_setattr(CurrentAttr>>4|CurrentAttr<<4);
}

void disp_setattr(int attr)
{ CurrentAttr=attr;
  textcolor(Color(attr&0xf));
  textbackground(Color(attr>>4));
}

void disp_setcursortype(int curt)  //I wonder what the window module will think
{ r.h.ah=1;                        //of this
  r.x.cx=curt;
  ioint();
}

void disp_scroll(int n,unsigned top,unsigned left,unsigned bot,unsigned right,
                 unsigned attr)
{ unsigned *areatop=&(_fullscreen->Buffer[top*80+left]);
  char lines=1+bot-top;
  if (n>lines || -n>lines)
    n=0;  // If out of bounds clear screen
  if (n<0)         // Scrolling down screen
  { char lmove=lines+n;
    for (unsigned *finger=areatop+80*(lines-1);lmove;lmove--)
    { memwcpy(finger,finger+80*n,1+right-left);
      finger-=80;
    }
    lines=n;
  }
  else if (n>0)
  { char lmove=lines-n;
    for (unsigned *finger=areatop;lmove;lmove--)
    { memwcpy(finger,finger+80*n,1+right-left);
      finger+=80;
    }
    areatop=areatop+80*(lines-n);
    lines=n;
  }
  for (unsigned *finger=areatop;lines;--lines)
  { memwset(finger,' '|attr<<8,1+right-left);
    finger+=80;
  }
  _redrawsection(_fullscreen,left,top,right,bot);
}

void disp_peekbox(unsigned *buff,unsigned top,unsigned left,unsigned bot,
                  unsigned right)
{ char nl=1+bot-top,l=0;
  char wid=1+right-left;
  for (unsigned *finger=buff;l<nl;l++,finger+=wid)
    memwcpy(finger,&_fullscreen->Buffer[(top+l)*80+left],wid);
}

void disp_pokebox(unsigned *buff,unsigned top,unsigned left,unsigned bot,
                  unsigned right)
{ char nl=1+bot-top,l=0;
  char wid=1+right-left;
  for (unsigned *finger=buff;l<nl;l++,finger+=wid)
    memwcpy(&_fullscreen->Buffer[(top+l)*80+left],finger,wid);
  _redrawsection(_fullscreen,left,top,right,bot);
}

void disp_fillbox(unsigned attrchar,unsigned top,unsigned left,unsigned bot,
                  unsigned right)
{ unsigned amount=(1+right-left)*(bot-top+1);
  unsigned *temp=new unsigned [amount];
  memwset(temp,attrchar,amount);
  disp_pokebox(temp,top,left,bot,right);
  delete [] temp;
}
