#include "menu.hpp"
#ifdef __OS2__
#include "OS2Glue.hpp"
#else
#include <bios.h>
#endif

#define Enter 0x1C0D
#define DownArrow 0x5000
#define DownArrowR 0x50E0
#define UpArrow 0x4800
#define UpArrowR 0x48E0

MenuItem::MenuItem(unsigned char Q,char *name,long retval,menufunc *f)
{ next=0;
  Quick=Q;
  heading= new char [strlen(name)+1];
  strcpy(heading,name);
  Datum=retval;
  func=f;
}

MenuItem::~MenuItem()
{ if (next)
  { delete next;
    next=0;
  }
  if (heading)
    delete [] heading;
}

char MenuItem::operator[](long D)  // This returns the number of the item matching D
{ if (D==Datum)        // A negative item is returned for missing element
    return 0;
  else
    if (next)
      return (*next)[D]+1;
    else
      return -128;
}

char MenuItem::operator[](unsigned char x) // Find by selchar
{ if (x==Quick)
    return 0;
  else
    if (next)
      return (*next)[x]+1;
    else
      return -128;
}

MenuItem & MenuItem::operator[](char x)   // Recursion is neat but slow
{ if (x)
    return (*next)[char(x-1)];
  else
    return *this;
}

void MenuItem::operator+=(MenuItem &it)   // Add menuitem to end of list
{ if (next)
    (*next)+=it;
  else
    next=&it;
}

Menu::~Menu()
{ if (head) delete head;
}

void Menu::blank()
{ head=0;
  nitems=0;
  maxlen=0;
  x1=y1=-1;
  back=LightGray;   // The following should be conditional on color screen
  fore=Blue;
  selback=Cyan;
  selfore=Blue;
  highback=LightGray;
  highfore=Red;
  highselback=Cyan;
  highselfore=Red;
}

void Menu::additem(char *x) // Add item using position as datum
{ additem(x,NULL,'\0');
}

void Menu::additem(char *x,menufunc *func) // Add item with procedure
{ additem(x,func,'\0');
}

void Menu::additem(char *x,unsigned char c)
{ additem(x,(menufunc *)NULL,c);
}

void Menu::additem(char *x,menufunc *f,unsigned char c)
{ MenuItem &MIP=*new MenuItem((unsigned char)(c & 0x5F),x,nitems,f);
  if (head)
    (*head)+=MIP;
  else
    head=&MIP;
  nitems++;
  char t=strlen(x);
  if (f)
    t++;
  if (t>maxlen)
    maxlen=t;
}

void Menu::outline(MenuItem &x,char Current)
{ Color NormBack=Current?selback:back;
  Color NormFore=Current?selfore:fore;
  char high=x.Quick;
  char Textlen=strlen(x.heading);
  textcolor(NormFore);
  textbackground(NormBack);
  for (char i=0;i<Textlen;i++)
  { if ((x.heading[i]&0x5f)==high)
    { textcolor(Current?highselfore:highfore);
      textbackground(Current?highselback:highback);
      putch(x.heading[i]);
      high=0;
      textcolor(NormFore);
      textbackground(NormBack);
    }
    else
      putch(x.heading[i]);
  }
  for (i=Textlen;i<maxlen-1;i++)
    putch(' ');
  if (Textlen<maxlen)
    putch(x.func?'\x10':' ');
}

void Menu::draw(char cpos)
{ clrscr();
  for (char i=0;i<nitems;i++)
  { outline((*head)[i],i==cpos);
    if (i<nitems-1) putch('\n');
  }
}

long Menu::operator()(long starting) // Execute menu
{ char X1=x1>=0?x1:(ScreenWidth-maxlen-1)/2;
  char Y1=y1>=0?y1:(ScreenDepth-nitems-1)/2;
  window W(X1,Y1,X1+maxlen+1,Y1+nitems+1,SINGLEFRAME,fore,back,fore,back);
  setwrap(0);
  cursoroff();
  char cpos=(*head)[starting];
  if (cpos<0)
    cpos=0; // Ignore false starting values
  draw(cpos);
  do
  { char ocpos=cpos;
    unsigned key=bioskey(0);
    switch (key)
    { case DownArrow:
      case DownArrowR:
        if (++cpos>=nitems)
          cpos=0;
        break;
      case UpArrow:
      case UpArrowR:
        if (--cpos<0)
          cpos=nitems-1;
        break;
      default:
      { char t=(*head)[(unsigned char)(key & 0x5F)];
        if (t>=0)
          cpos=t;
        else
          break;
        draw(cpos);
      }
      case Enter:
        MenuItem &hit=(*head)[cpos];
        if (hit.func)
          return hit.func();
        else
          return hit.Datum;
    }
    if (cpos!=ocpos)
    { gotoxy(1,ocpos+1);
      outline((*head)[ocpos],0);
      gotoxy(1,cpos+1);
      outline((*head)[cpos],1);
    }
  } while (1);
}
