#ifndef precise_maths
#define precise_maths
#include <iostreamp.hpp>
#pragma link(precise)
#define nodebug

#ifdef debug
extern long mem_out;
#endif

#if (defined M_I86SM)|(defined M_I86CM)
#define RET 0xC3
#else
#define RET 0xCB
#endif

#pragma save, call(inline=>on)
static unsigned aam(signed char x) =
{ 0xD4, 0x0A, RET
};
#pragma restore

class digit
{ private:
    signed char value;
    void norm()
    { unsigned t=aam(value);
      carry=t>>8;
      value = (signed char)t;
    }
    void sub_norm()
    { if (value<0)
      { value+=10;
        digit::carry=1;
      }
      else
        norm();
    }
  public:
    static signed char carry;  // Note that a signed carry math is used
    void operator=(digit const& y)
    { value=y.value;
    }
    digit()
    { value=0;
    }
    digit(int x)
    { value=aam(x);
    }
    void operator+=(digit &y)
    { value+=y.value+carry;
      norm();
    }
    digit digit::operator+(digit &y)
    { digit t=*this;
      t+=y;
      return t;
    }
    void add(digit &y)
    { value+=y.value;
      norm();
    }
    void sub(digit &y)
    { value-=y.value;
      sub_norm();
    }
    void operator-=(digit &y)
    { value-=(carry+y.value);
      sub_norm();
    }
    void mul(digit &y)
    { value*=y.value;
      norm();
    }
    void operator*=(digit &y) // Not very clean
    { value*=y.value;
      value+=carry;
      norm();
    }
    operator char()
    { return value; }
};

class precise_unsigned
{ private:
    unsigned buflen;
    unsigned bufuse;
    digit *buffer;
    void inc_space(unsigned);
    char compare(precise_unsigned const&y) const; // -ve if host less +ve if host greater
    void copy(precise_unsigned &x);
  public:
    precise_unsigned()
    { buflen=0;
      bufuse=0;
      buffer=0;
    };
    precise_unsigned(long x);
    precise_unsigned(precise_unsigned &x)
    { copy(x);
    }
    void operator=(precise_unsigned &x)
    { delete [] buffer;
#ifdef debug
        mem_out-=buflen;
#endif
      copy(x);
    }
    ~precise_unsigned()
    { delete [] buffer;
#ifdef debug
        mem_out-=buflen;
#endif
    }
    void operator+=(precise_unsigned &x);
    friend precise_unsigned operator+(precise_unsigned &x,precise_unsigned &y);
    void operator-=(precise_unsigned &x);
    friend precise_unsigned operator-(precise_unsigned &x,precise_unsigned &y);
    void operator*=(precise_unsigned &x);
    friend precise_unsigned operator*(precise_unsigned &x,precise_unsigned &y);
    void operator*=(unsigned char x);
    void pow10(unsigned p)
    { if (p)
      { inc_space(bufuse+p);
        memmove(buffer+p,buffer,bufuse);
        memset(buffer,0,p);
        bufuse+=p;
      }
    }
    friend inline char operator==(precise_unsigned const&x,precise_unsigned const&y);
    friend inline char operator!=(precise_unsigned const&x,precise_unsigned const&y);
    friend inline char operator>=(precise_unsigned const&x,precise_unsigned const&y);
    friend inline char operator<=(precise_unsigned const&x,precise_unsigned const&y);
    friend inline char operator>(precise_unsigned const&x,precise_unsigned const&y);
    friend inline char operator<(precise_unsigned const&x,precise_unsigned const&y);
    friend ostream& operator<<(ostream &out,precise_unsigned &x);
};

inline char operator==(precise_unsigned const&x,precise_unsigned const&y)
{ return x.compare(y)==0;
}
inline char operator!=(precise_unsigned const&x,precise_unsigned const&y)
{ return x.compare(y)!=0;
}
inline char operator>=(precise_unsigned const&x,precise_unsigned const&y)
{ return x.compare(y)>=0;
}
inline char operator<=(precise_unsigned const&x,precise_unsigned const&y)
{ return x.compare(y)<=0;
}
inline char operator>(precise_unsigned const&x,precise_unsigned const&y)
{ return x.compare(y)>0;
}
inline char operator<(precise_unsigned const&x,precise_unsigned const&y)
{ return x.compare(y)<0;
}

#endif