(* Copyright (C) 1991 Jensen & Partners International *)

{Ŀ
  TopSpeed Example Class Library        
                                        
  Basic Stream Classes                  
                                        
  See ClassLib.DOC for further details  
 
}

IMPLEMENTATION UNIT Streams;
IMPORT PasDos,CoreMem;

CONST
  DebugErrors = False;  { If true, full error messages are printed }


Const
  MemBuffSize = 4000H;
  Null        = 0FFFFH;

Type
  conv = record
           low : word;
           high : word;
         end;


{Ŀ
  Base Stream Class                 
  (not used directly)               
 
}



Implementation of Stream;

  Procedure Seek {(p : Integer)};
  Begin
    if p>0 then pos := p else pos := 0;
  end;

  Function Read {(VAR v):Boolean};
  Begin
    Read := ReadBytes(v,SIZE(v))=SIZE(v);
  end;

  Procedure Write {(v)};
  Begin
    WriteBytes(v,SIZE(v));
  end;

  Function ReadBytes {(VAR v; nbytes : word):Word};
  { implemented by client }
  Begin
    StreamException^.Raise(SE_deferred_readbytes);
    ReadBytes := 0;
  end;

  Procedure WriteBytes {(v; nbytes)};
  { implemented by client }
  Begin
    StreamException^.Raise(SE_deferred_writebytes);
  end;

  Function ReadBytesAt {(atpos:Integer; VAR v; nbytes : word):Word};
  Var
    oldpos : Integer;
    r : Word;
  Begin
    oldpos := pos;
    Seek(atpos);
    r := ReadBytes(v,nbytes);
    Seek(oldpos);
    ReadBytesAt := r;
  end;

  Procedure WriteBytesAt {(atpos:Integer;v; nbytes)};
  Var
    oldpos : Integer;
  Begin
    oldpos := pos;
    Seek(atpos);
    WriteBytes(v,nbytes);
    Seek(oldpos);
  end;

Begin
  strmsize := 0;
  pos := 0;
end;

Type
  buffarray = array[0..99] of farAddress;
  buffptr = ^buffarray;

Type A24 = array [0..23] OF Byte;
{# save, call(reg_param=>(si,ax,di,es,cx),reg_saved=>(ds,dx,st1,st2))}

{ very fast inline-expanded routine to do a block move }

Inline Procedure FastMove(src,dst : FarAddress; count : Word) =

  A24(
            0E3H,10H,            {     jcxz l2   }
            8CH,0DBH,            {     mov bx,ds }
            8EH,0D8H,            {     mov ds,ax }
            0F7H,0C6H,01H,00H,   {     test si,1 }
            74H,02H,             {     jz l1     }
            0A4H,                {     movsb     }
            49H,                 {     dec cx    }
            0D1H,                { l1: shr cx,1  }
            0E9H,0F3H,0A5H,      {     rep movsw }
            13H,0C9H,            {     adc cx,cx }
            0F3H,0A4H,           {     rep movsb }
            8EH,0DBH             {     mov ds,bx }
            );                   { l2:           }

{# restore }




Implementation of MemStream ;


  Function Open {(maxsize : Integer):Boolean};
  Var
    a : FarAddress;
    l : Word;
  Begin
    Close;
    pos := 0;
    strmsize := maxsize;
    Open := True;
  end;



  Procedure Close;
  var
    i : word;
    bp : buffptr;
  Begin
    bp := _buflist;
    if bp<>Nil then
    Begin
      i := _npages;
      while i>0 do
      Begin
        i := i-1;
        if bp^[i]<>FarNil then
          CoreMem._ffree(bp^[i]);
      end;
      FreeMem(bp,_npages*Size(FarAddress));
      _buflist := Nil;
    end;
    _npages := 0;
    pos := 0;
    strmsize := 0;
  end;

  Function ReadBytes {(VAR v; nbytes : word):Word};
  var
    p,q : FarAddress;
    s,l : Word;
  Begin
    if pos+nbytes>strmsize then nbytes := strmsize-pos;
    q := FarAdr(v);
    s := nbytes;
    while s<>0 do
    Begin
      p := _GetPageAddr(l);
      if l>s then l := s;
      FastMove(p,q,l);
      q := Ptr(Seg(q^),Ofs(q^)+l);
      s := s-l;
      pos := pos+l;
    end;
    ReadBytes := nbytes;
  end;

  Procedure WriteBytes {(v; nbytes)};
  var
    p,q : FarAddress;
    s,l : Word;
  Begin
    q := FarAdr(v);
    s := nbytes;
    while s<>0 do
    Begin
      p := _GetPageAddr(l);
      if l>s then l := s;
      FastMove(q,p,l);
      q := Ptr(Seg(q^),Ofs(q^)+l);
      s := s-l;
      pos := pos+l;
    end;
    if pos>strmsize then strmsize := pos;
  end;

  Function _GetPageAddr{(Var len: word):FarAddress};
  { returns buffer address and length to end of buffer }
  { allocates any buffers needed from far heap }
  var
    bp : buffptr;
    a  : FarAddress;
    page,offset : Word;
  Begin
    a := FarNil;
    page := pos div MemBuffSize;
    offset := pos mod MemBuffSize;
    if page>=_npages then
    Begin
      GetMem(bp,Size(FarAddress)*(page+1));
      Move(_buflist^,bp^,Size(FarAddress)*_npages);
      _buflist := bp;
      while page>=_npages do
      Begin
        bp^[_npages] := FarNil;
        _npages := _npages+1;
      end;
    end
    else
      bp := _buflist;
    len := MemBuffSize-offset;
    a := bp^[page];
    if a=FarNil then
    Begin
      bp^[page] := CoreMem._fmalloc(MemBuffSize);
      a := bp^[page];
    end;
    if a<>FarNil then
      a := Ptr(Seg(a^),Ofs(a^)+offset);
    if a=FarNil then
      StreamException^.Raise(SE_memstream_heapfull);
    _GetPageAddr := a;
  end;


Begin
  _buflist := Nil;
  Close;
end;


Implementation of FileStream;

  Procedure Open {(mode:StrmOpenMode; name : string[max])};
  Var r  : PasDos.Registers;
      fn : array[0..64] of char;
  Begin
    Close;
    filename := name;
    StrToZ(name,fn);
    r.DS := Seg(fn);
    r.DX := Ofs(fn);
    r.CX := 0;
    If mode=StrmOpen then
      r.AX := 3D00H { open for read }
    else
      if mode=StrmCreate then
        r.AX := 3C00H  { create }
      else
        r.AX := 3D02H; { open for read and write }
    PasDos.MsDos(r);
    if r.Flags mod 2 = 1 then
    Begin
      errcode := r.AX;
      r.AX := Null;
    end
    else
    Begin
      errcode := 0 ;
      r.CX := 0;
      r.DX := 0;
      r.AX := 4202H; { seek absolute from end }
      r.BX := _handle;
      PasDos.MsDos(r) ;
      strmsize := ((r.DX::Integer) << 16)+r.AX;
      if mode = StrmAppend then
        pos := strmsize
      else
        Seek(0);
    end;
    _handle := r.AX;
    if (errcode<>0)AND errcheck then
      StreamException^.Raise(SE_filestream_open);
  end;


  Procedure Close;
  Var r  : PasDos.Registers;
  Begin
    If _handle=Null then errcode := 0
    else
    Begin
      r.BX := _handle;
      r.AH := 3EH; { close file }
      PasDos.MsDos(r);
      if r.Flags mod 2 = 1 then errcode := r.AX else errcode := 0 ;
      if (errcode<>0)AND errcheck then
        StreamException^.Raise(SE_filestream_close);
      pos := 0;
      strmsize := 0;
    end;
  end;

  Procedure CloseErase;
  Var r  : PasDos.Registers;
      fn : array[0..64] of char;
  Begin
    Close;
    If errcode=0 then
      If filename<>'' then
      Begin
        StrToZ(filename,fn);
        r.DS := Seg(fn);
        r.DX := Ofs(fn);
        r.AH := 41H;
        PasDos.MsDos(r);
        if r.Flags mod 2 = 1 then errcode := r.AX
        else errcode := 0 ;
      end;
    if (errcode<>0)AND errcheck then
      StreamException^.Raise(SE_filestream_closeerase);
  end;

  Procedure Seek {(p : Integer)};
  Var r  : PasDos.Registers;
  Begin
    Stream.Seek(p);
    r.CX := (pos:>conv).high;
    r.DX := (pos:>conv).low;
    r.AX := 4200H; { seek absolute }
    r.BX := _handle;
    PasDos.MsDos(r) ;
    if r.Flags mod 2 = 1 then
    begin
      errcode := r.AX;
      if errcheck then
        StreamException^.Raise(SE_filestream_seek);
    end
    else
      errcode := 0 ;
  end;

  Function ReadBytes {(VAR v; nbytes : word):Word};
  Var r  : PasDos.Registers;
  Begin
    r.DS := Seg(v);
    r.DX := Ofs(v);
    r.AH := 3FH; { file read }
    r.BX := _handle;
    r.CX := nbytes;
    PasDos.MsDos(r) ;
    if r.Flags mod 2 = 1 then
    begin
      errcode := r.AX;
      if errcheck then
        StreamException^.Raise(SE_filestream_readbytes);
      r.AX := 0;
    end
    else
      errcode := 0 ;
    pos := pos+r.AX;
    ReadBytes := r.AX;
  end;

  Procedure WriteBytes {(v; nbytes:word)};
  Var r  : PasDos.Registers;
  Begin
    { check extent }
    r.DS := Seg(v);
    r.DX := Ofs(v);
    r.AH := 40H; { file read }
    r.BX := _handle;
    r.CX := nbytes;
    PasDos.MsDos(r) ;
    if r.Flags mod 2 = 1 then
    Begin
      errcode := r.AX;
      r.AX := 0;
    end
    else
    Begin
      if r.AX<>r.CX then errcode := 0F0H else errcode := 0;
    end;
    if (errcode<>0)and errcheck then
      StreamException^.Raise(SE_filestream_writebytes);
    pos := pos+r.AX;
    if pos>strmsize then strmsize := pos;
  end;
Begin
  errcheck := true;
  _handle := Null;
  filename := '';
  Close;
end;

Const
  EmsFrameSize = 4000H;
Var
  Int67[67H*4]      : FarAddress;
  Int67Seg[67H*4+2] : Word;


Implementation of EmsStream;


  Function Open{(maxsize : Integer):Boolean};
  TYPE
    b8 = ARRAY [0..7] OF CHAR;
  VAR
    r       : PasDos.Registers;
    i       : Word;
    emssize : Word;
{#save,data(near_ptr=>off)}
    p       : ^b8;
{#restore }
  Begin
    Close;
    Open := False;
    IF (Int67<>FarNil) then
    Begin
      { Check for Ems installed }
      p := Ptr(Int67Seg,10);
      if (p^[0]='E')AND(p^[1]='M')AND(p^[2]='M')AND(p^[3]='X')AND(p^[4]='X')AND
         (p^[5]='X')AND(p^[6]='X')AND(p^[7]='0') then
      Begin
        r.AH := 40H;
        PasDos.Intr(67H,r);
        if r.AH=0 then
        Begin
          r.AH := 41H;
          PasDos.Intr(67H,r);
          if (r.AH=0) then
          Begin
            _emsPageFr := r.BX ;
            emssize := maxsize div EmsFrameSize;
            r.AH := 42H;
            PasDos.Intr(67H,r); (* find how many pages free *)
            if r.BX>=emssize then
            Begin
              r.AH := 43H;
              r.BX := emssize ;
              PasDos.Intr(67H,r); (* allocate ems pages *)
              _emshandle := r.DX ;
              _npages := emssize;
              Open := True;
            end;
          end;
        end ;
      end;
    end ;
  end;

  Procedure Close;
  { closes stream if open }
  VAR r : PasDos.Registers;
  Begin
    If _emshandle<>Null then
    Begin
      r.AH := 45H;
      r.DX := _emshandle;
      PasDos.Intr(67H,r);
      _emshandle := Null ;
    end ;
    _npages := 0;
    pos := 0;
    strmsize := 0;
  end;

  Function _GetPageAddr{(Var len: word):FarAddress};
  { returns ems buffer address and length to end of buffer }
  var
    r : PasDos.Registers;
    o : Word;
  Begin
    r.AX := 4400H;
    r.BX := pos div EmsFrameSize;
    r.DX := _emshandle;
    PasDos.Intr(67H,r);
    if r.AH<>0 then
      StreamException^.Raise(SE_ems_pagemap);
    o := pos mod EmsFrameSize;
    _GetPageAddr := Ptr(_emspagefr,o) ;
    len := EmsFrameSize-o;
  End;


Begin
  _emshandle := Null;
  Close;
end; (* EmsFileObj *)

Class LocalExceptionClass Inherits Classes.ExceptionClass;
  Function   Handler{:Boolean};
End;

Implementation of LocalExceptionClass;

{Ŀ
  Local Exception Handler           
 
}


  Function  Handler{:Boolean};
  Begin
    WriteLn('Exception: ',code,', Module: Streams.pas ');
{%T DebugErrors}
    Case code of
      SE_ok:
      SE_heap_full:             Write('heap full');
      SE_deferred_readbytes:    Write('ReadBytes not implemented');
      SE_deferred_writebytes:   Write('WriteBytes not implemented');
      SE_memstream_heapfull:    Write('MemStream, heap full');
      SE_filestream_open:       Write('FileStream, Open failed');
      SE_filestream_close:      Write('FileStream, Close failed');
      SE_filestream_closeerase: Write('FileStream, CloseErase failed');
      SE_filestream_seek:       Write('FileStream, Seek failed');
      SE_filestream_readbytes:  Write('FileStream, ReadBytes failed');
      SE_filestream_writebytes: Write('FileStream, WriteBytes failed');
      SE_ems_pagemap:           Write('EMS page map failure');
    End;
    WriteLn;
{%E}
    Halt;
    Handler := True;
  End;

begin
end;

VAR
  LEC : LocalExceptionClass;

begin
  StreamException := Adr(LEC);
end.
