#pragma data(stack_size => 20000)
/* Last updated 20/12/90                       */

/*-------------------------------------------------------------------------*
*                                                                          *
*  windemo.c - window and process demonstration program.     		   *
*                                                                          *
*  COPYRIGHT (C) 1989-90 Jensen and Partners. All Rights Reserved          *
*                                                                          *
*--------------------------------------------------------------------------*/

#define _JPI_WIN_

#include <stddef.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <process.h>
#include <task.hpp>
#include <textwin.hpp>
#include <iostream.hpp>
#include <iomanip.hpp>
#include <coniostr.hpp>
#ifdef __OS2__
#include "OS2Glue.hpp"
#else
#include <bios.h>
#endif

#ifndef _mthread
#error Must be made in MT model
#endif

typedef enum kfs { RShift,  LShift,  Ctrl,  Alt,  Scroll,  Num,  Cap,  Ins } KBFlagSet;

typedef struct {
    unsigned x,  y;
} WCoor;

typedef WCoor WW[30];

void P1(void);
void P2(void);
void P3(void);
void P4(void);
void P5(void);
void P6(void);
void P7(void);
void P8(void);
void P9(void);

void SetC(Color cc, Color bwc);
void Intro(void);
void IntroText(void);
void RandPalette(PaletteDef p);
void Rmove(window &W);
void RChangeSize(window &W) ;
void DrawBar (unsigned x, unsigned y, unsigned val, int on);
void Demo(void);
void CheckBW(void);

int _isbw;
semaphore DS;
const WW ww={
    {0, 0},
    {5, 0},
    {10, 0},
    {15, 0},
    {20, 0},
    {10, 2},
    {10, 4},
    {10, 6},
    {10, 8},
    {10, 10},
    {5, 10},
    {0, 10},
    {35, 4},
    {35, 2},
    {35, 0},
    {40, 0},
    {45, 0},
    {50, 2},
    {50, 4},
    {45, 6},
    {40, 6},
    {35, 6},
    {35, 8},
    {35, 10},
    {65, 0},
    {65, 2},
    {65, 4},
    {65, 6},
    {65, 8},
    {65, 10}
};

class SimpleTask : private task
{ public:
    SimpleTask(void (*func)(void))
    { init(func,2000,0,1);
    }
    ~SimpleTask(){}
};

void CheckScrollLock()
{   Lock();
    while (bioskey(2)&(1<<Scroll));
    Unlock();
    return;
}

void SetC(Color cc, Color bwc)
{
    if( _isbw)
        cc = bwc;
    textcolor(cc);
    return;
}


void IntroText()
{
    SetC(White,  LightGray);
    conout<<"\n       This is a demonstration of the ";
    SetC(Yellow, White);
    conout << "Window" ;
    SetC(White, LightGray);
    conout <<" and " ;
    SetC(Yellow, White) ;
    conout << "Process" ;
    SetC(White, LightGray);
    conout << " Modules\n\n" ;
    SetC(LightGray, LightGray);
    conout <<"           9 time-sliced processes are created,  which output to, \n";
    conout <<"           move,  re-size and re-colour  4 overlapping windows\n\n";
    SetC(White, White);
    conout << "     Press any key to start,  Esc to finish ";
    return;
}

int ext,  chnge;
windef DescDef={0, 15, 75, 24, White, Black,FALSE, TRUE, FALSE, TRUE,
                DOUBLEFRAME, Blue, LightGray};

void Intro()
/* Display startup and demonstration description */
{

   windef WD;
   window *Win[30];
   unsigned i;
   window *DescWin;
   unsigned ddelay;

   const int  targetX = 32 ;
   const int  targetY =  5 ;

    for( i = 0; i <= 29; i++ )
    { Color Fore,Back,FrameFore,FrameBack;
        if( _isbw)
        {
            if( i&1 )
            {
                Back = Black ;
                Fore = White ;
            }
            else
            {
                Back = LightGray ;
                Fore = Black ;
            }
        }
        else
        {
            Back = Color(i&7) ;
            Fore = Color((15-i)&7) ;
        }
        FrameBack= !_isbw && Back == Black ? Green : Back;
        FrameFore= FrameBack==LightGray ? Black : LightGray;
        Win[i] = new window(ww[i].x,ww[i].y,ww[i].x+9,ww[i].y+3,SINGLEFRAME,
                            FrameFore,FrameBack,Fore,Back) ;
        Win[i]->settitle("JPI",  CenterUpperTitle);
        Win[i]->use();
        conout << "TopSpeed" << "  C++";
    }
    if( _isbw )
        DescDef.FrameFore = Black ;
    DescWin = new window(DescDef) ;
    DescWin->settitle(" JPI TopSpeed C++ : Window Demonstration ", CenterUpperTitle);
    IntroText();
    ddelay = 100;
    i = 0 ;
    while( !kbhit() )
    {
        CheckScrollLock();
        if( ddelay > 10 )
            ddelay-=2;
        else if (ddelay > 1 )
            --ddelay;
        delay(ddelay);
        Win[i]->putontop() ;
        i = (i+1)%30 ;
    }
    do
    {
        ext=TRUE;
        for( i = 0; i <= 29; i++ )
        {
            Win[i]->info(WD) ;
            chnge = TRUE ;
            if(WD.X1 < targetX-3 )
            {
                WD.X1+=3;
                WD.X2+=3;
            }
            else if (WD.X1 > targetX+3)
            {
                WD.X1-=3;
                WD.X2-=3;
            }
            else if(WD.Y1==targetY)
                chnge = FALSE ;
            if(WD.Y1 < targetY )
            {
                ++WD.Y1;
                ++WD.Y2;
            }
            else if(WD.Y1 > targetY)
            {
                --WD.Y1;
                --WD.Y2;
            }
            if( chnge )
            {
                Win[i]->change(WD.X1, WD.Y1, WD.X2, WD.Y2) ;
                ext=FALSE;
            }
        }
    }
    while(ext == FALSE);
    for(i = 0; i < 30; i++)
    {
      delete Win[i];
        delay(1);
     }
     delete DescWin;
     if( getch()==27 )
         exit(0);
     return;
}

/* Actual demo starts here */
/* ----------------------- */

window *W1,  *W2,  *W4;
palettewindow *W3;
int Initialize;
PaletteDef pal;
typedef  Color BW3[3];
const BW3 bw3 ={Black,  LightGray,  White};

/* Random Colouring */

void RandPalette (PaletteDef p)
{
    unsigned i;

    for(i = 0; i < PALETTEMAX ; i++)
    {
        if(_isbw)
            p[i].Fore = bw3[random(3)] ;
        else
            p[i].Fore = random(16);
        do
        {
            if(_isbw)
                p[i].Back = bw3[random(3)];
            else
                p[i].Back = random(8);
        }while(p[i].Back == p[i].Fore);
    }
    return;
}

/* Random Movement */

void Rmove(window &W)
{
    unsigned dx1, dy1;
    unsigned ix, iy, nx, ny;
    windef WDef;

    while(1)
    {
        W.info(WDef);
        do
        {
            dx1 = WDef.X1+(random(16)-8)*2 ;
            dy1 = WDef.Y1+random(16)-8 ;
        }
        while(!( dx1 < 80 &&  dy1 < 24 && dx1+(WDef.X2-WDef.X1) <= 80
                 && dy1+(WDef.Y2-WDef.Y1) <= 24));
        ix = dx1>WDef.X1?2:0;
        iy = dy1>WDef.Y1?2:0;
        while( dx1 != WDef.X1 && dy1 != WDef.Y1)
        {
            if(dx1 != WDef.X1)
                nx=WDef.X1+ix-1;
            if(dy1 != WDef.Y1)
                ny = WDef.Y1+iy-1;
            W.change(nx, ny, nx+(WDef.X2-WDef.X1), ny+(WDef.Y2-WDef.Y1)) ;
            W.info(WDef);
        }
    }
}

/* Random Change Size */

void RChangeSize(window &W)
/* this randomly changes the size of a window */
{
    unsigned nx, ny;
    windef WDef;
    W.info(WDef);
    do
    {
        nx = WDef.X2+(random(16)-8)*2 ;
        ny = WDef.Y2+random(16)-8 ;
    } while(!(nx-WDef.X1 < 70 && ny-WDef.Y1 < 18 && nx-WDef.X1 > 10 &&
              ny-WDef.Y1 > 4));
    W.change(WDef.X1, WDef.Y1, nx, ny) ;
}


/* Processes */

/* P1 prints out numbers in Hex,  Dec and Binary */

void P1()
{
    unsigned i;

    W1->use() ;
    while(1)
    {
        for(i = 0; i < 1000; i++ )
        {   char S[16];
            char T[5];
            char U[5];
            ultoa(i, S, 2);
            ultoa(i, T, 10);
            ultoa(i, U, 16);
            conout << setw(10) << S << setw(5) << T << setw(5) << U << endl << flush;
        }
    }
}

/* P2 displays out a sideways barchart */

void DrawBar (unsigned x, unsigned y, unsigned val, int on)
{   textcolor(!(y&1) ? LightGray :_isbw ? White : Red);
    if(val > 0)
    {
        --val;
        gotoxy(x+val/2, y) ;
        conout << (char)(on ? (val&1 ? '':''):(val&1 ?'':' '));
    }
    return;
}


void P2()
{
    unsigned i, j;
    static unsigned V[10];
    windef WD;

    W2->use() ;
    i = 0 ;
    for(i = 1; i <= 9; i++ )
    {
        V[i] = random(50);
        for(j = 0; j <= V[i]; j++)
        {
            DrawBar(1, i, j, TRUE) ;
        }
    }
    while(1)
    {
        for(i = 1; i <= 9; i++)
        {
            W2->info(WD);
            if( i < WD.X2-WD.X1-2 )
            {
                if((random(2)==0 ) && ( V[i]<50 ))
                {
                    ++V[i];
                    DrawBar(1, i, V[i], TRUE);
                }
                else if( V[i] > 0 )
                {
                    DrawBar(1, i, V[i], FALSE);
                    --V[i];
                }
            }
        }
    }
}

/* P3 displays a character grid with randomly changing pallette colours */

void P3()
{
    unsigned X, Y, D;
    char CH;

    CH = '*' ;
    D = 0 ;
    W3->use() ;
    while(1)
    {
        for( X = 1; X <= 8; X++ )
        {
            for( Y = 1; Y <= 8; Y++ )
            {
                setpalettecolor((Y*7+X+D)&7);
                gotoxy(X, Y);
                conout << CH;
                gotoxy(1, 10);
                setpalettecolor(0) ;
                conout << setw(2) << X << ", " << Y;
            }
        }
        RandPalette( pal ) ;
        W3->setpalette(pal) ;
        ++D;
    }
}

/* P4 displays some text */

void P4()
{   char cont=1;
    unsigned i;

    W4->use() ;
    do
    {
        for(i = 1; cont && i <= 100; i++)
        {
            conout <<"This is a test of 9 time-sliced processes,  and also of window writing,  moving,  rearranging and resizing." << flush;
            conout << " Press any key to terminate. " << flush;
            CheckScrollLock();
            if(kbhit())
                cont = 0;
        }
    } while(cont);
    DS.clear();
    while (1) Delay(10);
}

/* P5 Re-sizes and Re-arranges text */

void P5()
{
    while(1)
    {
        switch(random(4))
        {
            case 0 :
                W1->putontop();
                break;
            case 1 :
                W2->putontop();
                break;
            case 2 :
                W3->putontop();
                break;
            case 3 :
                W4->putontop();
        }
        switch(random(4))
        {
            case 0 :
                RChangeSize(*W1);
                break;
            case 3 :
                RChangeSize(*W4);
        }
    }
}

/* P6 Moves window 1 */

void  P6()
{
    Rmove(*W1);
    return;
}

/* P7 Moves window 2 */

void  P7()
{
    Rmove(*W2);
    return;
}

/* P8 Moves window 2 */

void  P8()
{
    Rmove(*W3);
    return;
}

/* P9 Moves window 2 */

void  P9()
{
    Rmove(*W4);
    return;
}


void Demo()
{
    Color Back;
    randomize();
    Initialize = TRUE ;
    clrscr();
    cursoroff();
    Back=_isbw?Black:Red;
    W1=new window(10,2,60,8,SINGLEFRAME,White,Back,White,Back);
    W1->settitle(" Process 1 ", CenterUpperTitle);
    W2 = new window(40,2,70,12,DOUBLEFRAME,Black,_isbw?LightGray:Cyan,
                     White,_isbw?Black:Blue);
    W2->settitle(" Process 2 ", CenterUpperTitle);
    RandPalette(pal) ;
    pal[0].Fore= _isbw?LightGray:Yellow ;
    pal[0].Back= Black ;
    pal[1].Fore= White ;
    pal[1].Back= _isbw?Black:Green ;
    W3=new palettewindow(pal ,40,6,50,17,SINGLEFRAME,Black,_isbw?LightGray:Cyan,
                         White,_isbw?Black:Blue) ;
    W3->settitle(" Process 3 ", CenterUpperTitle);
    W4 = new window(10,10,70,20,DOUBLEFRAME,White,_isbw ? Black:Green,Black,LightGray) ;
    W4->settitle(" Process 4 ", CenterUpperTitle);

    StartScheduler();
    SimpleTask p1[]={P1,P2,P3,P4,P5,P6,P7,P8,P9};
    DS.set_and_wait();
    StopScheduler();
    window *lamb=new window(0,0,79,24);  // Must leave a window that is not auto-closed
    clrscr();
    gotoxy(1, SCREENDEPTH-1);
    cursoron();
    return;
}

void CheckBW()
{
   _isbw = FALSE;
   return;
}


int main()
{   clrscr();
    CheckBW();
    Intro();
    Demo();
    return(0);
}


