#include "winkey.hpp"
#include <string.h>
#include <stdio.h>
#include <disp.hpp>
#include <stdlib.h>
#ifdef __OS2__
#include "os2glue.hpp"
#else
#include <bios.h>
#endif

byte norm_screen=0x1f,invert_screen=0x71,exclaim_screen=0x4f;

/* Standard initialiser for Windows note Window shrinks if boxed*/
Window::Window(byte a,int y,int x,int w,int h,boolean box)
{ if (h<3 || w<3) box=false;
  if (box) {
    top_left_row=y+1;
    top_left_col=x+1;
    width=w-2;
    height=h-2;
  }
  else
  { top_left_row=y;
    top_left_col=x;
    width=w;
    height=h;
  }
  boxed=box;
  attr=a;
  cu_row=0;
  cu_col=0;
  next=NULL;
  showing=1;
  under=0;
}

Window::Window(byte a,int y,int x,int w,int h,Window& win)
{ top_left_row=win.top_left_row+y;
  top_left_col=win.top_left_col+x;
  width=w;
  height=h;
  boxed=0;
  attr=a;
  cu_row=0;
  cu_col=0;
  next=NULL;
  showing=1;
  under=0;
}

Window::~Window()
{ if (under)
    pop_down();
}

void Window::print_word(char *w)
{ int word_length=strlen(w);
  if (cu_col+word_length>width) 
    newline();
  else
    cursor();
  disp_puts(w);
  cu_col+=word_length;
  cursor();
}

void Window::resize(int y,int x,int w,int h,char b)
{ boxed=(b==2?boxed:b);
  if (boxed) {
    top_left_row=y+1;
    top_left_col=x+1;
    width=w-2;
    height=h-2;
  }
  else
  {
    top_left_row=y;
    top_left_col=x;
    width=w;
    height=h;
  }
}

/* Scroll routine insists cursor still on screen */
void Window::scroll(int amount)
{ disp_scroll(amount,top_left_row,top_left_col,height+top_left_row-1,
              width+top_left_col-1,attr);
  cu_row-=amount;
}

void Window::clear()
{ disp_scroll(height,top_left_row,top_left_col,height+top_left_row-1,
              width+top_left_col-1,attr);
  if (boxed) 
    disp_box(0,attr,top_left_row-1,top_left_col-1,
	     top_left_row+height,top_left_col+width);
  cursor(0,0);
}

void Window::cursor()
{ disp_move(top_left_row+cu_row,top_left_col+cu_col);
  disp_setattr(attr);
  if (showing)
    disp_flush();
}

void Window::cursor_follow(boolean on)
{ showing=on;
  cursorstate(showing?release:turn_off);
}

/* newline moves cursor to current Window */
void Window::newline()
{ if (++cu_row>=height) scroll(1);
  cu_col=0;
  cursor();
}

/* Pop up a Window */
byte Window::pop_up()
{ under=malloc((height+2)*(width+2)*sizeof(unsigned));
  if (under)
  { if (boxed)
      disp_peekbox(under,top_left_row-1,top_left_col-1,top_left_row+height,
       	 top_left_col+width);
    else
      disp_peekbox(under,top_left_row,top_left_col,top_left_row+height-1,
      		   top_left_col+width-1);
  }
  else
    return 1;
  clear();
  return 0;
}

void Window::pop_down()
{ if (boxed)
    disp_pokebox(under,top_left_row-1,top_left_col-1,
                 top_left_row+height,top_left_col+width);
  else
    disp_pokebox(under,top_left_row,top_left_col,
                 top_left_row+height-1,top_left_col+width-1);
  free(under);
  under=0;
}
 
/* add a Window to the chain maintained by the screen */
void whole_screen::add_Window(Window& w)
{ *tail=&w;
  tail=&w.next;
}

void whole_screen::display()
{ Window *finger=&my_Window;
  while (finger)
  { finger->clear();
    finger=finger->next;
  }
}

/* Note that a screen Window is owned by each keyboard handler */
key_handler::key_handler()
{ file_name=0;
  recording=0;
  buffer_pos=0;
  current.length=0;
  info_win=new Window(norm_screen,23);
}

key_handler::key_handler(char *s)
{ FILE *quickfile=fopen(s,"rb");
  file_name=malloc(strlen(s)+1);
  strcpy(file_name,s);
  if (quickfile)
  { if (!fread(&quick[0],sizeof(quick),1,quickfile))
      memset(&quick[0],0,sizeof(quick));
    fclose(quickfile);
  }
  else memset(&quick[0],0,sizeof(quick));
  recording=0;
  buffer_pos=0;
  current.length=0;
  info_win=new Window(norm_screen,23);
}

void key_handler::macro_def(unsigned Fn)
{ recording=1;
  recording_into=Fn-1;
  quick[recording_into].length=0;
  info_win->clear();
  disp_printf("Recording quick key ShiftF%d -- Press Escape to end",Fn);
}

void key_handler::view_macros(Window &text_win)
{ text_win.clear();
  disp_printf("Quick keys currently defined");
  text_win.newline();
  for (int i=0;i<10;i++)
    if (quick[i].length)
    { disp_printf("F%2d:",1+i);
      for (int j=0; j<quick[i].length; j++)
	disp_putc(quick[i].text[j]);
      text_win.newline();
    }
}

void key_handler::save_macros()
{ if (!file_name) return;
  FILE *quickfile=fopen(file_name,"wb");
  if (quickfile)
  { fwrite(&quick[0],sizeof(quick),1,quickfile);
    fclose(quickfile);
  }
  else display_message("Unable to create macro file");
}
	      

/* Routine to get key, also expands quick keys as it goes */
int key_handler::get_key()
{ int ret;
  char ub;
  if (current.length-buffer_pos)
    ret=current.text[buffer_pos++];
  else
    ret=bioskey(0);
  ub=(ret>>8);
  if (ub>=0x54 && ub<=0x5d) /* Is a function key */
  { memcpy(&current,&quick[ub-0x54],sizeof(current));
    buffer_pos=0;
    return get_key();
  }
  if (recording)
    if (ub==1)
    { recording=false;
      info_win->clear();
    }
    else
    { if (quick[recording_into].length>9)
      { display_message("Maximum quick key length exceeded");
	recording=false;
      }
      quick[recording_into].text[quick[recording_into].length++]=ret;
    }
  return ret;
}

static Window DisplayWindow=Window(norm_screen,24);

void display_message(char *message_string)
{ char sl=strlen(message_string)+18;
  Window DW(exclaim_screen,10,(80-sl)/2,sl,3,true);
  DW.pop_up();
  cursorstate(turn_off);
  disp_printf("%s : Press any key",message_string);
  bioskey(0);
  cursorstate(release);
}

void display_info(const char *message_string)
{ DisplayWindow.clear();
  disp_puts(message_string);
}

char display_query(char* message_string)
{ int ch;
  char sl=strlen(message_string)+12;
  Window DW(exclaim_screen,10,(80-sl)/2,sl,3,true);
  DW.pop_up();
  DW.print_word(message_string);
  DW.print_word(" : (y/n)?");
  cursorstate(turn_on);
  do
  { ch=bioskey(0)>>8;
  } while (ch!=0x15 && ch!=0x31);
  cursorstate(release);
  return ch==0x15;
}

/* Routine to input string from keyboard with echo to current cursor position */

char key_handler::edit_string(Window &within,char *buffer, int max_len)
{ char ckey;
  int where_in=strlen(buffer);
  within.print_word(buffer);
  cursorstate(turn_on);
  within.cursor();
  while ((ckey=get_key())!=0x0d)
  { switch (ckey)
    { case 8:
        if (where_in)
        { within.move_cursor(0,-1);
          disp_putc(' ');
          within.cursor();
	  buffer[--where_in]='\0';
	}
	break;
      case 27:
        cursorstate(release);
        return 0;
        break;
      default:
	if (where_in<max_len)
	{ disp_putc(ckey);
          within.move_cursor(0,1);
	  buffer[where_in++]=ckey;
	}
	break;
    }
  }
  buffer[where_in]='\0';
  cursorstate(release);
  return 1;
}

char get_filename(key_handler &key,char *prompt,char *name,char *ext,boolean exists)
{ unsigned fwidth=strlen(prompt)+filenamelength+2;
  if (fwidth>80)
    fwidth=80;
  char margin=(80-fwidth)/2;
  Window filewin(norm_screen,10,margin,fwidth,3,true);
  char temp[filenamelength];
  strcpy(temp,name);
  filewin.pop_up();
  char esc;
  do
  { filewin.clear();
    filewin.print_word(prompt);
    if (esc=!key.edit_string(filewin,temp,filenamelength-4))
      break;
    strcat(temp,ext);
    FILE *test=fopen(temp,"r");
    char opened=test!=0;
    if (opened)
      fclose(test);
    if (exists && !opened)
    { if (display_query("File does not exist, create"))
      { test=fopen(temp,"w");
        if (test)
        { fclose(test);
          break;
        }
        else
          display_message("Unable to open file");
        break;
      }
    }
    else if (!exists && opened)
    { if (display_query("File already exists, destroy"))
	break;
    }
    else
      break;
  } while (1);
  filewin.pop_down();
  if (!esc)
    strcpy(name,temp);
  return !esc;
}

void cursorstate(curstate show)
{ static curstate cstate[50]={turn_on};
  static char level=0;
  char action=show!=release && show!=cstate[level] ||
              show==release && cstate[level]!=cstate[level-1];
  if (show==release)
    level--;
  else
    cstate[++level]=show;
  if (action)
    if (cstate[level]==turn_on)
      disp_showcursor();
    else
      disp_hidecursor();
}
