/* Release 3.00 */
/*-------------------------------------------------------------------------*
*                                                                          *
*  bios.h -  header file for ROM BIOS access.                              *
*                                                                          *
*  COPYRIGHT (C) 1989..1991 Jensen and Partners. All Rights Reserved       *
*                                                                          *
*--------------------------------------------------------------------------*/

#ifndef _BIOS_INC_
#define _BIOS_INC_

#ifdef __cplusplus
extern "C" {
#endif
#ifndef _WINDOWS
int	 bioscom(int _mode, char _byte, int _port);
int	 biosdisk(int _mode, int _drive, int _head, int _track, int _sector,
			int _secnum, void *_buffer);
int	 bioskey(int _mode);
int	 biosprint(int _mode, int _byte, int _port);
#endif			
int	 biosequip(void);
int	 biosmemory(void);
long	 biostime(int _mode, long _newtime);
#ifdef __cplusplus
}
#endif

#define _COM_INIT       0   /* init serial port */
#define _COM_SEND       1   /* send character */
#define _COM_RECEIVE    2   /* receive character */
#define _COM_STATUS     3   /* get serial port status */

/* serial port initializers. */

/* character size initializers */

#define _COM_CHR7       2   /* 7 bits characters */
#define _COM_CHR8       3   /* 8 bits characters */

/* stop bit values */

#define _COM_STOP1      0   /* 1 stop bit */
#define _COM_STOP2      4   /* 2 stop bits */

/*  parity initializers */

#define _COM_NOPARITY   0   /* no parity */
#define _COM_ODDPARITY  8   /* odd parity */
#define _COM_EVENPARITY 24  /* even parity */

/*  baud rate initializers */

#define _COM_110        0       /* 110 baud */
#define _COM_150        32      /* 150 baud */
#define _COM_300        64      /* 300 baud */
#define _COM_600        96      /* 600 baud */
#define _COM_1200       128     /* 1200 baud */
#define _COM_2400       160     /* 2400 baud */
#define _COM_4800       192     /* 4800 baud */
#define _COM_9600       224     /* 9600 baud */


/* constants for BIOS disk support */

/* disk services */

#define _DISK_RESET     0   /* reset disk controller */
#define _DISK_STATUS    1   /* get disk status */
#define _DISK_READ      2   /* read disk sectors */
#define _DISK_WRITE     3   /* write disk sectors */
#define _DISK_VERIFY    4   /* verify disk sectors */
#define _DISK_FORMAT    5   /* format disk track */

/* struct used to send/receive information to/from the BIOS disk services */

#ifndef __STDC__
struct diskinfo_t {
    unsigned drive;
    unsigned head;
    unsigned track;
    unsigned sector;
    unsigned nsectors;
    void far *buffer;
    };
#endif
/* constants for BIOS keyboard support */

/* keyboard services */

#define _KEYBRD_READ        0   /* read next character from keyboard */
#define _KEYBRD_READY       1   /* check for keystroke */
#define _KEYBRD_SHIFTSTATUS 2   /* get current shift key status */


/* constants for BIOS printer support */

/* printer services */

#define _PRINTER_WRITE  0   /* write character to printer */
#define _PRINTER_INIT   1   /* intialize printer */
#define _PRINTER_STATUS 2   /* get printer status */


/* constants for BIOS time of day support */

/* time of day services */

#define _TIME_GETCLOCK  0   /* get current clock count */
#define _TIME_SETCLOCK  1   /* set current clock count */


#ifdef __cplusplus
extern "C" {
#endif
#ifndef _WINDOWS
unsigned  _bios_keybrd(unsigned _cmd);
unsigned  _bios_printer(unsigned _cmd, unsigned _port, unsigned _data);
unsigned  _bios_serialcom(unsigned _cmd, unsigned _port, unsigned _data);
unsigned  _bios_disk(unsigned _cmd, struct diskinfo_t *_info);
#endif
unsigned  _bios_equiplist(void);
unsigned  _bios_memsize(void);
unsigned  _bios_timeofday(unsigned _mode, long *_time);
#ifdef __cplusplus
}
#endif

#endif