/* Release 3.00 */
/*-------------------------------------------------------------------------*
*                                                                          *
*  clib.h -  internal header file for C library.                           *
*                                                                          *
*  COPYRIGHT (C) 1989..1991 Jensen and Partners. All Rights Reserved       *
*                                                                          *
*--------------------------------------------------------------------------*/

#ifndef _CLIB_INC_
#define _CLIB_INC_

#ifndef __cplusplus
#pragma call( seg_name => null)    /* C Library uses _TEXT and _DATA  by */
#pragma data( seg_name => null)    /* default. May be over-ridden locally. */
#endif

#pragma warn(wall => on)           /* enable all warnings */

#define _STATIC static             /* hide local stuff */

#define INCL_SUB                   /* bring in OS2 sub systems */
#define INCL_DOSPROCESS            /* bring in OS2 processes */

#define _IO_MTF
#define _CT_MTF

#define LONGDOUBLE long double
#define FLOORL(x) floorl(x)
#define LOG10L(x) log10l(x)

#include <alloc.h>
#ifdef _USE_BIOS
#include <bios.h>
#endif
#ifdef _USE_CONIO
#include <conio.h>
#endif
#include <ctype.h>
#include <dir.h>
#include <errno.h>
#include <dos.h>
#include <fcntl.h>
#ifdef _USE_FLOAT
#include <float.h>
#endif
#ifdef _USE_GRAPH
#include <graph.h>
#endif
#ifndef _NO_IO
#include <io.h>
#endif
#include <limits.h>
#include <locking.h>
#include <locale.h>
#include <mem.h>
#include <math.h>
#include <process.h>
#include <search.h>
#include <setjmp.h>
#include <share.h>
#include <signal.h>
#include <stat.h>
#include <stdarg.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef _USE_TIME
#include <time.h>
#include <timeb.h>
#endif
#ifdef _USE_WINDOW
#include <window.h>
#endif

#ifdef _USE_OS2
#include <os2kernl.h>              /* only include API if making OS2 lib */
#endif

#ifdef __cplusplus
#pragma module(init_prio=>25)
#endif

/***************************************************************************
*                                                                          *
*  Internal Data Types.                                                    *
*                                                                          *
****************************************************************************/

typedef struct	{			    /* structure for DOS function 4B */
		     unsigned env;
		     char   far *com;
		     void   far *FCB5;
		     void   far *FCB6;

		} pblock;

typedef void (*procptr)(void);

#ifdef _USE_WINDOW
#ifndef _WINDOWS
typedef  struct usl *UseListPtr;
typedef  struct usl {
                       UseListPtr Next;
#ifdef _OS2
                       int        Proc;
#else
                       void      *Proc;
#endif
                       wintype    Wind;
                     } UseListLink;
#endif
#endif


#pragma save
#ifndef _DLL
#pragma call(near_call => on)
#endif

typedef struct sfc {               /* control structure for vscanner */
    FILE *_stream;                 /* file pointer */
    int  (*_vs_fill)(FILE *st);    /* filler function pointer */
    int _vs_count;                 /* character count */
    int _s_args;                   /* argument count */
    jmp_buf _EOFhandler;           /* EOF jump environment */
} SF ;

typedef struct  pf {
    FILE *_stream;                           /* file pointer */
    int (*_vp_flush)(FILE *st);        /* output function pointer */
    int _vp_ret;                             /* character count */
                   } PF;

#ifdef __cplusplus
extern "C" {
#endif
extern int   (*_vs_fill)(FILE *st);    /* filler function pointer */
extern int   (*_vp_flush)(FILE *st);   /* output function pointer */

int         _filler(FILE *st);
int         _flusher(FILE *st);
#ifdef __cplusplus
}
#endif

#pragma restore


typedef  char  far *fva_list;

#define fva_start(ap, plist) ap = (fva_list)&plist + sizeof(plist)
#define fva_arg(ap, t)       ((t far *)(ap += sizeof(t)))[-1]
#define fva_end(ap)          ap = NULL

/***************************************************************************
*                                                                          *
*  Internal Function Prototypes - Default Calling convention               *
*                                                                          *
****************************************************************************/

#ifdef __cplusplus
extern "C" {
#endif

#ifndef _WINDOWS
#ifdef _USE_WINDOW
void     _initwin(void);           /* low level windows initialization */
void     _checkwindow(wintype win);
#endif
#endif

int      _dup(int handle);	   /* low level I/O */
int      _dup2(int h1, int h2);	

int      _os2_open(const char *path, int access, int permission, int mode);
int      _getfmode(const char *path);
void     _ioerr(int code);
unsigned lsl(unsigned sel);

void        _exit1(void);
void        _exit2(void);
void        _exitf(void);

#ifndef _WINDOWS
int            _g_putch(int c);
char          *_g_cgets(char *s);
int            _g_charoutput(int c);
char          *_g_strinput(char *s);
char          *_d_cgets(char *s);
int            _d_putch(int c);
int            _w_putch(int c);
char          *_w_cgets(char *string);
extern int   (*_i_putch)(int c);
extern char *(*_i_cgets)(char *s);
int            _activepage(void);
#endif

char *_ld_format(char *fmt_buf, LONGDOUBLE num, int flag, int prec);
char *_v_format(unsigned long l, char *s);

int  _matherr(struct exception *x);
int  _flsbuf(int  ch, FILE *st);
int  _filbuf(FILE *st);

#ifdef __cplusplus
}
#endif
/***************************************************************************
*                                                                          *
*  Internal Function Prototypes - Near Calling convention.                 *
*                                                                          *
****************************************************************************/
#ifdef __cplusplus
extern "C" {
#endif

#pragma save
#pragma call(near_call => on)

#ifdef _USE_TIME
clock_t        _clock_time(void);
size_t         _set_date(char *s, size_t scount, size_t lim, char *st);
unsigned long  _t_calc(struct date dte, struct time tme);
int            _t_to_struct(unsigned long val, struct date *dte, struct time *tme);
struct tm    * _t_to_tm(unsigned long clock);
unsigned long  _t_norm(struct tm *timeptr);
unsigned long  _ytable(int n);
#endif

void        StreamLock(FILE *st);
void        StreamUnlock(FILE *st);

#ifndef _OS2
va_list          _rreal_out(va_list _v_argptr, int flag, int wid, int prec);
va_list          _real_pdef(va_list _v_argptr, int flag, int wid, int prec);
va_list          _rreal_in(int pmod, int smod, unsigned flag, int width, va_list _v_argptr);
va_list          _real_sdef(int pmod, int smod, unsigned flag, int width, va_list _v_argptr);
#else
va_list          _rreal_out(va_list _v_argptr, int flag, int wid, int prec, PF *control);
va_list          _real_pdef(va_list _v_argptr, int flag, int wid, int prec, PF *control);
va_list          _rreal_in(int pmod, int smod, unsigned flag, int width, va_list _v_argptr, SF *control);
va_list          _real_sdef(int pmod, int smod, unsigned flag, int width, va_list _v_argptr, SF *control);
#endif

char  far *_ff_fmt(char far *_fmt_buf, LONGDOUBLE num, int flag, int prec);
char      *_f_fmt(char *_fmt_buf, LONGDOUBLE num, int flag, int prec);
char      *_fptoa(unsigned long, char *s);

double      _atof(const char *s, int flag, int *pos);

int         _pf_entry(FILE *st);
void        _pf_exit(FILE *st, int flag);
void        _vp_setbuf(FILE *st, char *buf);
void        _vp_freebuf(FILE *st);

FILE       *_alloc_stream(FILE *stream, int handle, int flag);
void        _flushall(void);
int         _free_stream(void);
int         _open_stream(const char *path, int *fptr, int mode);
int         _scan_type(const char *type, int *fptr, int *mptr);
int         _valid_stream(FILE *st);

#ifndef _WINDOWS
#ifdef _USE_WINDOW
void        _bell(void);
void        _bufferspacefill(wintype win, unsigned pos, unsigned len);
void        _clipframe(wintype win);
void        _clipxy(wintype win, relcoord *X, relcoord *Y );
#ifdef _OS2
int         _currentproc(void);
#else
void       *_currentproc(void);
#endif
wintype     _curwin(void);
void        _drawframe(wintype win);
void        _displaybeneath(wintype win, wintype NW);
void        _disposetitle(wintype win);
void        _getpal(wintype win, PaletteDef Pal);
void        _igotoxy(wintype win, relcoord X, relcoord Y);
wintype     _makewindow(windef *WD);
void        _mergewindows(wintype s, wintype d);
void        _puttitle(wintype win, unsigned mode, unsigned row);
void        _redrawsection(wintype win, abscoord X1, abscoord Y1, abscoord X2, abscoord Y2);
void        _redrawwindow(wintype win);
void        _redrawwindowpane(wintype win);
void        _resetcursor(void);
void        _setpal(wintype win, PaletteDef Pal);
void        _takeoffstack (wintype win);
void        _unlinkcursor(wintype win);
void        _updatescreen(wintype win, abscoord X, abscoord Y, unsigned Len );
void        _windowwrite(wintype win, relcoord x, relcoord y, unsigned Len,
                         const void *str, int frame);
void        _writec(wintype win, char c);
void        _writeout(char s[], int ss);
#endif
#endif
#pragma restore
#ifdef __cplusplus
}
#endif

/***************************************************************************
*                                                                          *
*  Internal Function Prototypes - Default call, but small save set.        *
*                                                                          *
****************************************************************************/
#ifdef _USE_WINDOW
#ifdef __cplusplus
extern "C" {
#endif

#ifndef _WINDOWS
#pragma save
#pragma call(reg_saved=>(ds, di, si,st1,st2))
unsigned char _getscreendepth(void);
void         _buffertoscreen(abscoord X, abscoord Y, const void *buffer, unsigned len);
void         _bufferwrite(void *buffer, const void *src, unsigned len, unsigned attr);
void         _initscreentype(int issnow);
void         _palxlat(unsigned buff[], void *a, unsigned len, unsigned char attr[]);
void         _screentobuffer(abscoord X, abscoord Y, void *buffer, unsigned len);
#pragma restore
#endif
#ifdef __cplusplus
}
#endif
#endif
/***************************************************************************
*                                                                          *
*  Internal Function Prototypes - near call = reg_param,                   *
*  but small save set.                                                     *
*                                                                          *
****************************************************************************/

#ifdef __cplusplus
extern "C" {
#endif

#pragma save
#pragma call(near_call=>on, reg_saved=>(ds, di, si,st1,st2), reg_param=>(ax,bx,cx,dx), c_conv=>off)

#ifdef _OS2
int    _vgetc(SF *control);
int    _vungetc(int c, SF *control);
#else
int    _vgetc(void);
int    _vungetc(int c);
#endif

void   _ljustify(char far *s, int flag, int width);
void   _lstr_out(char far *s, int flag, int width, int precision);

#ifdef _OS2
void   _justify(char *s, int flag, int width, PF *control);
void   _str_out(char *s, int flag, int width, int precision, PF *control);
void   _vputc(int ch, PF *control);
#else
void   _justify(char *s, int flag, int width);
void   _str_out(char *s, int flag, int width, int precision);
void   _vputc(int ch);
#endif

#pragma restore

#ifdef __cplusplus
}
#endif

/***************************************************************************
*                                                                          *
*  Internal Function Prototypes - near call non-standard reg_param,        *
*  and small save set.                                                     *
*                                                                          *
****************************************************************************/

#ifdef __cplusplus
extern "C" {
#endif
#pragma save
#pragma call(reg_param=>(dx,ax), reg_saved=>(ds, di, si,st1,st2))
unsigned char _exists(const char *path);
#pragma restore
#ifdef __cplusplus
}
#endif

/***************************************************************************
*                                                                          *
*  Internal Function Prototypes - near call, and small save set.           *
*                                                                          *
****************************************************************************/

#ifdef __cplusplus
extern "C" {
#endif

#pragma save
#pragma call(near_call => on, reg_saved=>(ds, di, si,st1,st2))
void        _hswap(void huge *b1, void huge *b2, size_t width);
void        _swap(void *b1, void *b2, size_t width);
#pragma restore

#ifdef __cplusplus
}
#endif

/***************************************************************************
*                                                                          *
*  Internal Function Prototypes - c convention and reg_param.              *
*                                                                          *
****************************************************************************/

#ifdef __cplusplus
extern "C" {
#endif
#pragma save, call(seg_name=>null,reg_param=>(ax,bx,cx,dx,st0,st6,st5), c_conv=>off )
extern LONGDOUBLE  _acvt(int flag, char *mbuf, int exp, int len);
extern void        _lcvt(LONGDOUBLE x, int n, int far *dec, int far *sign, char far * buf);
extern void        _cvt(LONGDOUBLE x, int n, int *dec, int *sign, char * buf);
extern void        _seterrno(int type);
#pragma restore
#ifdef __cplusplus
}
#endif

#ifdef __cplusplus
extern "C" {
#endif
#pragma save, call(seg_name=>null)
extern LONGDOUBLE  _pow_int(LONGDOUBLE x,int y);
extern LONGDOUBLE  _round( LONGDOUBLE x );
extern void        _bcd(LONGDOUBLE x,char near * p);
#pragma restore
#ifdef __cplusplus
}
#endif

/***************************************************************************
*                                                                          *
*  Internal Function Prototypes - far call, c convention, stack param.     *
*                                                                          *
****************************************************************************/

#ifdef __cplusplus
extern "C" {
#endif
#pragma save
#pragma call(reg_param =>(), near_call => off, c_conv => on, reg_saved => (ds, si,di,st1,st2))
void _FatalError(unsigned long address, int code);
unsigned long _FatalErrorPos(void);
#pragma restore

#ifndef _WINDOWS
#pragma save
#pragma call(reg_param => (), c_conv => off, reg_saved=>(ds, si, di,st1,st2))
#pragma name(prefix => "")
PROCESS  SYSTEM$CurrentProcess(void);
void     SYSTEM$initprocess(void);
void     SYSTEM$NewPriority(unsigned _p);
unsigned SYSTEM$CurrentPriority(void);
void     CoreProc$NewPriority(unsigned _p);
unsigned CoreProc$CurrentPriority(void);
void     SYSTEM$IOTRANSFER(void **P1, void **P2, unsigned I);
void     SYSTEM$NEWPROCESS(procptr PP, void *A, unsigned S, void **P1);
void     SYSTEM$TRANSFER(void **P1, void **P2);
void     SYSTEM$Out(unsigned P, unsigned char V);
#pragma restore
#endif
#ifdef __cplusplus
}
#endif
/***************************************************************************
*                                                                          *
*  Internal Process Function Prototypes.                                   *
*                                                                          *
****************************************************************************/
#ifdef __cplusplus
extern "C" {
#endif
#pragma save
#pragma call(reg_saved=>(ds, di, si, st1, st2))
int     _beget(const char *path, char *args[], char *envs[], int flag);
int     _exec(const char *path, pblock *params);
int     _execve(const char *path, char *args[], char *envs[], int flag);
int     _get_retcode(void);
void    _setenv(void);
#pragma call(reg_param=>(di, es, si, ax, cx), reg_saved=>(ds, bx, dx, st1, st2))
void    _env_copy(char far *dest, char far *source, int num);
#pragma restore
#ifdef __cplusplus
}
#endif

/***************************************************************************
*                                                                          *
*  Internal Procedure variables.                                           *
*                                                                          *
****************************************************************************/

#ifdef __cplusplus
extern "C" {
#endif
extern void (*_exit_stk[32])(void); 		 /* at/onexit function stack */

#pragma save
#pragma check(index=>off)
extern void  (**_exit_ptr)(void);
extern void  (**_exit_top)(void);
#pragma restore

extern void  (*_exit_file)(void);
extern void  (*_exit_io)(void);
extern void  (*_exit_tmp)(void);

#pragma save
#pragma call(near_call=>on)

fva_list  _freal_out(fva_list _v_argptr, int flag, int wid, int prec);

#ifdef _OS2
extern va_list (*_real_in)(int pmod, int smod, unsigned flag, int width, va_list _v_argptr, SF *control);
extern va_list (*_real_out)(va_list _v_argptr, int flag, int wid, int prec, PF *control);
#else
extern va_list (*_real_in)(int pmod, int smod, unsigned flag, int width, va_list _v_argptr);
extern va_list (*_real_out)(va_list _v_argptr, int flag, int wid, int prec);
#endif
#pragma restore

#pragma save
#pragma call(reg_saved=>(di,si,ds,st1,st2))
extern int (* _shr_mem)(void);
#pragma restore

extern void  (*_Continue)(void);

#pragma save
#pragma call(reg_saved=>(ds,di,si,st1,st2))
extern void  (*_exit_proc)(void);
#pragma restore
#ifdef __cplusplus
}
#endif

/***************************************************************************
*                                                                          *
*  Internal Data Declarations.                                             *
*                                                                          *
****************************************************************************/

#ifdef __cplusplus
extern "C" {
#endif

extern char         _cvt_buf[64];    /* real number conversion buffer */
extern char         _ctype[257];     /* ctype data */
extern struct lconv _lconv;
#ifdef _USE_TIME
extern time_t       _proc_time;	     /* time at beginning of process */
#endif
extern char         _sys_errbuf[128];/* error message buffer */
extern char         *sys_errlist[];  /* error messages */
extern int           sys_nerr;	     /* number of system error strings */
extern double       _tens[32];
extern char         _upperHex;       /* conversion case */

extern jmp_buf    _EOFhandler;	     /* EOF error jump environment */
extern unsigned   _fmask;
extern            _iosetup;	     /* initialisation flag */
extern char      *_tmpfiles[];       /* temp file name array */
extern int        _open_init;
extern int        _p_init;
extern int        _s_args;           /* argument count */
extern FILE      *_stream;           /* file pointer */
extern int        _ungotchar;        /* temporary storage for console I/O */
extern int        _vp_ret;           /* character count */
extern int        _vs_count;         /* character count */
extern int        _open_max;         /* maximum number of open files */
extern char       *_tmpfiles[];
extern FILE       *_tmpfptrs[];

#ifndef _WINDOWS
#ifdef _USE_WINDOW
extern wintype       _cursorstack;   /* windows data */
extern unsigned      _cursorlines;
extern wintype       _fullscreen;
extern UseListPtr    _uselist;
extern wintype       _windowstack;
extern int           _winctrl;
extern int           _winlocks;
extern unsigned char _winsetup;
#endif

extern unsigned      _ProcIds[];     /* thread data */

#ifdef _OS2
extern unsigned      _LastProc;
extern unsigned char _Started;
extern unsigned long _LockSem;
extern unsigned      _LockCount;
extern unsigned      _LockedThread;
#else
extern Task          _cp;  /* currently active task + ready queue */
extern Task          _dq;  /* queue of delayed tasks */
extern Task          _wq;  /* queue of tasks that have reached thier delay time
                              but haven't been placed on the ready queue */

extern PROCESS           _SchedProc;
extern void            * _SchedStack;
extern unsigned char     _Started;
extern unsigned char volatile Stop;
extern unsigned volatile _LockNestMonster;
extern unsigned volatile _SchedTime;
extern int volatile      _NextThread;
extern unsigned char     _proc_init;
#endif
#ifdef _mthread
extern struct thread     _threadTable[MAXTHREAD];
#endif
extern int               _threadTotal;
#endif

extern char         *environ[];		  /* environment array */
extern char         *_envbuf;	          /* environment buffer */
extern unsigned int  _env;                /* environment segment */
extern char          _argbuf[128];	  /* command line buffer */
extern char          _sigTrans[];         /* signal translation */
extern char          _raiseTrans[];
extern unsigned char _env_init;
extern unsigned      _assign_init_magic;
#ifdef __cplusplus
}
#endif

/*****************************************************************************/
/* CPP Definitions.                                                          */
/*****************************************************************************/

typedef void *(*ctor_t)(void *);
typedef void (*dtor_t)(void *);

#pragma save, call(reg_param=>(), reg_saved=>(ds,si,di,st1,st2), near_call=>on, c_conv=>on)
#ifdef __cplusplus
extern "C" {
#endif
void   _call_ctors(ctor_t, char *obj_ptr, int vbnum, int n, int siz);
void   _call_dtors(dtor_t, char *obj_ptr, int vbnum, int n, int siz);
#ifdef __cplusplus
}
#endif
#pragma restore

/*****************************************************************************/
/* Macro Definitions.                                                        */
/*****************************************************************************/

#define TRUE       1
#define FALSE      0

#define _CTLZ    0x01A
#define _O_LRD   0x400
#define _WRITE	     2
#define _NODEFMODE   4
#define _READ	     8

#define _MNEG        1
#define _ENEG        2
#define _STD         8
#define _HMAX       54
#define _HEX         1
#define _OCT         2
#define _NEG         4
#define _OFL         8

#define _DAYSECS       86400L	  /* seconds in year */
#define _SEVENTIES 315532800L     /* seconds between 1970 and 1980 */

#define InternalError(c) _FatalError(_FatalErrorPos(), (c))

#define PF_VBSIZE 256
#define PF_TEMP_BUF 128

#define DEC_POINT _lconv.decimal_point[0]

#define PF_LEFT	     0x0001	      /* justification */
#define PF_SIGN	     0x0002	      /* force leading PF_SIGN */
#define PF_SPACE     0x0004	      /* force leading PF_SPACE */
#define PF_HASH	     0x0008	      /* PF_HASH flag */
#define PF_NOPOINT   0x0010	      /* supress radix point */
#define PF_UPPER     0x0020	      /* PF_UPPER case type */
#define PF_IS_SIGNED 0x0040	      /* signed data-type */
#define PF_ZFLAG     0x0080	      /* format with 0 */
#define PF_HEX	     0x0100	      /* PF_HEX integer  */
#define PF_OCT	     0x0200	      /* octal integer */
#define PF_GTYPE     0x0400	      /* g real type */
#define PF_EXP	     0x0800	      /* e real type */
#define PF_PSET	     0x1000	      /* precison set */
#define PF_LDMOD     0x2000	      /* long real */
#define PF_CPP       0x4000	      /* C++ formatting rules */

#define SF_WILD	     0x0004	      /* any int type */
#define SF_SUPRESS   0x0008	      /* star flag */
#define SF_UPPER     0x0010	      /* upper case type */
#define SF_NSIGNED   0x0020	      /* unsigned data-type */
#define SF_HEX	     0x0040	      /* hex integer  */
#define SF_OCT	     0x0080	      /* octal integer */
#define SF_LDMOD     0x0100           /* LONGDOUBLE */

/*****************************************************************************/
/* Graphics Macro Definitions.                                               */
/*****************************************************************************/

#ifdef _USE_GRAPH

#ifndef _WINDOWS
#define  ON                  _txcolor
#define  OFF                   0
#define  CGA320Width         320
#define  CGA640Width         640
#define  CGADepth            200
#define  CGANumColor         4
#define  EGA640Width         640
#define  EGA320Width         320
#define  EGA200Depth         350
#define  EGA350Depth         350
#define  EGA480Depth         350
#define  EGANumColor         16
#define  VGAWidth            320
#define  VGADepth            200
#define  VGANumColor         16
#define  HercWidth           720
#define  HercDepth           348
#define  HercNumColor        2

#define ODD(c) ((c)&1)

#define CGA320_ADDRESS    ((bs <0xB800> *)((x>>2)+((0x2000-40)*(y&1))+(40*y)))
#define CGA640_ADDRESS    ((bs <0xB800> *)((x>>3)+((0x2000-40)*(y&1))+(40*y)))

#define HEADER_SIZE    4
#define FILL_MASK_SIZE 8
#define GRAPHICS       1
#define TEXT           0
#define GR_LEFT       -1
#define GR_RIGHT       1
#define GR_UP         -1
#define GR_DOWN        1
#define MAXMODE        _MRES256COLOR

/*****************************************************************************/
/* Type and structure Definitions.                                           */
/*****************************************************************************/

typedef  unsigned char bs;
typedef  bs far *bp;
typedef  bs HercMapType[HercDepth/4][HercWidth/8];

struct intersect {
    short x;
    short y;
};

struct floodstart {
    short best_quad;
    short quad_type;
    short flag;
};

struct video_type {
    unsigned char sys0;
    unsigned char dis0;
    unsigned char sys1;
    unsigned char dis1;
};

/*****************************************************************************/
/* Data Definitions.                                                         */
/*****************************************************************************/

#ifdef __cplusplus
extern "C" {
#endif
extern int _depth;
extern int _width;
extern struct xycoord _origin;
extern struct xycoord _clip_tl;
extern struct xycoord _clip_br;
extern struct rccoord _text_tl;
extern struct rccoord _text_br;
extern struct videoconfig _current_video;
extern struct xycoord _current_graph;
extern struct rccoord _current_text;
extern struct floodstart _fstart;

extern short _lastmode;
extern short _defaultmode;
extern short _active_page;
extern short _visual_page;
extern short _page_size;
extern long _bkcolor;
extern short _fgcolor;
extern short _txcolor;
extern short _mode_changed;
extern unsigned char _cursor_state;
extern unsigned char _wrap_state;
extern unsigned char _display_state;
extern short _cursor_on;
extern short _scr_attr;
extern long _col_table[];
extern char _EGA_table[];
extern short _EGAStartPlane;
extern short _EGAPlaneShift;
extern unsigned _EGAtranslate;
extern unsigned char _EGA64K;

extern unsigned char *_current_mask;
extern unsigned char _fill_mask[FILL_MASK_SIZE];
extern unsigned short _current_linestyle;
extern unsigned short _cursor_lock;
extern int far *_EGAScreen;
extern int _ungotchar;
extern int _p_init;
extern long _CGAPal[4][3];
#ifdef __cplusplus
}
#endif

/*****************************************************************************/
/* Graphics Function prototypes.                                             */
/*****************************************************************************/

#ifdef __cplusplus
extern "C" {
#endif
#pragma save
#pragma call(reg_param => (), c_conv=>on, reg_saved=>(ds, di, si,st1,st2))

extern void (*_get)(void far *image, short x1, short y1, short x2, short y2);
extern void (*_put)(short x, short y, const void far *image, int action);

int           _getvideomode(void);

void          _txt_out(int ch);
void          _gr_out(int ch);
void          _setbiosmode(int mode);
void          _setlimits(struct videoconfig *video);
void          _getsystem(struct video_type  *video);
int           _getmemory(void);
void          _clear_Herc(void);
void          _clear_CGA(void);
void          _clear_EGA(int row, int col);

void          _CGA320Get(void far *image, short x1, short y1, short x2, short y2);
void          _CGA320Put(short x, short y, const void far *image, int action);
void          _CGA640Get(void far *image, short x1, short y1, short x2, short y2);
void          _CGA640Put(short x, short y, const void far *image, int action);
void          _EGAGet(void far *image, short x1, short y1, short x2, short y2);
void          _EGAPut(short x, short y, const void far *image, int action);
void          _VGAGet(void far *image, short x1, short y1, short x2, short y2);
void          _VGAPut(short x, short y, const void far *image, int action);
void          _HercGet(void far *image, short x1, short y1, short x2, short y2);
void          _HercPut(short x, short y, const void far *image, int action);

void          _graph_bfill(void far *base, int ch, int numwords, unsigned mask);
void          _graph_wfill(void far *base, int ch, int numwords);
#pragma restore
#ifdef __cplusplus
}
#endif

/*****************************************************************************/
/* Function prototypes - near/reg params.                                    */
/*****************************************************************************/

#ifdef __cplusplus
extern "C" {
#endif
#pragma save
#pragma call(reg_param => (ax,bx,cx,dx,st0,st6,st5,st4,st3), reg_saved=>(ds, di, si,st1,st2), c_conv=>off)

extern void  (*_hscan)(short *xl, short *xr, short y, short border);
extern void  (*_plot)(short x, short y, short c);
extern short (*_point)(short x, short y);
extern void  (*_hline)(short x, short y, short x2, short FillCol);
extern void  (*_line)(short x1, short y1, short x2, short y2, unsigned fmask);

void           _InitCGA(int mode);
void           _InitEGA(int mode);
void           _InitVGA256(void);
void           _InitHerc(void);

void           _setcur(int row, int col, int page);
void           _get_in_range(short *a0, short *b0);
void           _draw_ellipse(short  x0, short y0, short a0, short b0,
                             short fill);
int            _draw_arc(short  x0, short y0, short a0, short b0,
                         short startx, short starty, short endx, short endy);
void           _lag_fill(short xl, short xr, short y, int direction,
                         short llim, short rlim, short border);
void           _hscanline(short *xlp, short *xrp, short y, short border);
bs             _byte(bs low, bs high);

void           _CGA320Plot(short x, short y, short c);
void           _CGA640Plot(short x, short y, short c);
short          _CGA320Point(short x, short y);
short          _CGA640Point(short x, short y);
void           _CGA320Line (short x1, short y1, short x2, short y2, unsigned fmask);
void           _CGA320HLine ( short x, short y, short x2, short c);
void           _CGA320HScan(short *xl, short *xr, short y, short border);
void           _CGA640Line (short x1, short y1, short x2, short y2, unsigned fmask);
void           _CGA640HLine ( short x, short y, short x2, short c);
void           _CGA640HScan(short *xl, short *xr, short y, short border);
void           _EGAPlot( short x, short y, short c);
void           _EGA2Plot( short x, short y, short c);
void           _EGAPlot1( short x, short y, short c);
short          _EGAPoint(short x, short y);
short          _EGA2Point(short x, short y);
void           _EGAHLine ( short x, short y, short x2, short c);
void           _EGA2HLine ( short x, short y, short x2, short c);
void           _EGALine(short x1, short y1, short x2, short y2, unsigned fmask);
void           _EGA2Line(short x1, short y1, short x2, short y2, unsigned fmask);
void           _EGAHScan(short *xl, short *xr, short y, short border);
short          _VGAPoint(short x, short y);
void           _VGAPlot( short x, short y, short c);
void           _VGAHLine ( short x, short y, short x2, short c);
void           _VGALine(short x1, short y1, short x2, short y2, unsigned fmask);
void           _VGAHScan(short *xl, short *xr, short y, short border);
void           _HercPlot(short x, short y, short c);
short          _HercPoint(short x, short y);
void           _HercLine(short x1, short y1, short x2, short y2, unsigned fmask);
void           _HercHLine (short x, short y, short x2, short c);
void           _GenericHScan(short *xl, short *xr, short y, short border);
void           _HercGraphMode(void);
void           _HercTextMode(void);

int            _get_arc(char *quadrant, short x, short y, short sx, short xy,
                        short ex, short ey);
int            _arc_plot(int quadrant, int flag, int clipflag, short xp, short yp,
                         short startx, short endx, short starty, short endy);
struct intersect _get_vec(short x0, short y0, short a0, short b0, short vx, short vy);
void           _resetEGA(void);
void           _ellipse_plot(short x, short y);
short          _get_fillstart(short *x, short *y, short ox, short oy,
                              short startx, short starty, short endx, short endy);
void           _set_fillstart(char quadrant[]);
short          _col_xlat(long color);
short          _EGAcol_xlat(long color);
void           _g_enter(void);
void           _g_exit(void);
void           _gcur(short state, short x, short y, short page);
void           _scroll(void);
void           _gmemset(void far *buf, unsigned word, int num);
void           _gmemwset(void far *buf, unsigned word, int num);
void far      *_gmemcpy(void far *dest, void far *source, int num);
void           _clear_blink(void);
void           _setbc(void);
void           _setr10(void);
short          _gr_abs_value(short x);
void           _VGAmaskline(unsigned ptr, short c, short n, unsigned mask, unsigned bit);
short          _CGA320leftscan(unsigned ptr, short border, short lim, short left);
short          _CGA320rightscan(unsigned ptr, short border, short lim, short right);
short          _CGA640leftscan(unsigned ptr, short border, short lim, short left);
short          _CGA640rightscan(unsigned ptr, short border, short lim, short right);
short          _EGAleftscan(bs far *ptr, short border, short lim, short left);
short          _EGArightscan(bs far *ptr, short border, short lim, short right);
short          _VGAleftscan(unsigned ptr, short border, short lim, short left);
short          _VGArightscan(unsigned ptr, short border, short lim, short right);
short          _EGAxlat(short c);
#pragma restore

extern int (*_g_charout)(int c);
extern char *(*_g_strin)(char *s);

#pragma save
#pragma call(near_call=>off)
void      _initgraph(void);
#pragma restore
#ifdef __cplusplus
}
#endif

#endif
#endif
#endif






