/* Release 3.00 */
/*-------------------------------------------------------------------------*
*                                                                          *
*  fio.h -  header file for Modula-2 file IO procedures                    *
*                                                                          *
*  COPYRIGHT (C) 1989..1991 Jensen and Partners. All Rights Reserved.      *
*                                                                          *
*--------------------------------------------------------------------------*/

#ifndef _M2FIO_INC
#define _M2FIO_INC

#pragma project("#set tsm2='on'")

#pragma save
#pragma name(prefix=>"")
#pragma call(c_conv=>off)

#define MaxOpenFiles   15
#define DiskFull       0x0F0H
#define StandardInput  0
#define StandardOutput 1
#define ErrorOutput    2
#define AuxDevice      3
#define PrinterDevice  4

#ifndef _CHARSET_DEF
#define _CHARSET_DEF
typedef int CHARSET;
#endif

#ifdef __cplusplus
extern "C" {
#endif
extern short int FIO@EOF;
extern short int FIO@IOcheck;
extern CHARSET   FIO@Separators;
extern short int FIO@OK;
extern short int FIO@ChopOff;
extern short int FIO@Eng;

int      FIO$Open(unsigned NameH, const char *Name);
int      FIO$Append(unsigned NameH, const char *Name);
int      FIO$Create(unsigned NameH, const char *Name);
void     FIO$Close(int F);
void     FIO$AssignBuffer(int F, unsigned BufH, void *Buf);
int      FIO$Exists(unsigned NameH, const char *Name);
void     FIO$Erase(unsigned NameH, const char *Name);
void     FIO$Rename(unsigned NameH, const char *Name, unsigned NewNameH, const char *NewName);
void     FIO$Truncate(int F);
long     FIO$GetPos(int F);
void     FIO$Seek(int F, unsigned long Pos);
long     FIO$Size(int F);
int      FIO$IOresult(void);

void     FIO$WrChar(int F, char V);
void     FIO$WrBool(int F, short int V, int Length);
void     FIO$WrShtInt(int F, int V, int Length);
void     FIO$WrInt(int F, int V, int Length);
void     FIO$WrLngInt(int F, long V, int Length);
void     FIO$WrShtCard(int F, unsigned short V, int Length);
void     FIO$WrCard(int F, unsigned V, int Length);
void     FIO$WrLngCard(int F, unsigned long V, int Length);
void     FIO$WrShtHex(int F, unsigned short V, int Length);
void     FIO$WrHex(int F, unsigned V, int Length);
void     FIO$WrLngHex(int F, unsigned long V, int Length);
void     FIO$WrReal(int F, float V, int Length);
void     FIO$WrLngReal(int F, double V, int Length);
void     FIO$WrFixReal(int F, double V, unsigned Precision, int Length);
void     FIO$WrFixLngReal(int F, double V, unsigned Precision, int Length);

void     FIO$WrStr(int F, unsigned VH, const char *V);
void     FIO$WrStrAdj(int F, unsigned SH, const char *S, int Length);
void     FIO$WrCharRep(int F, char V, unsigned Count);
void     FIO$WrLn(int F);
void     FIO$WrBin(int F, unsigned BufH, const void *Buf, unsigned Count);

char     FIO$RdChar(int F);
int      FIO$RdBool(int F);
int      FIO$RdShtInt(int F);
int      FIO$RdInt(int F);
long     FIO$RdLngInt(int F);
unsigned FIO$RdShtCard(int F);
unsigned FIO$RdCard(int F);
unsigned long  FIO$RdLngCard(int F);
unsigned short RdShtHex(int F);
unsigned FIO$RdHex(int F);
unsigned long FIO$RdLngHex(int F);
float    FIO$RdReal(int F);
double   FIO$RdLngReal(int F);

void     FIO$RdStr(int F, unsigned VH, char *V);
void     FIO$RdItem(int F, unsigned VH, char *V);
unsigned FIO$RdBin(int F, unsigned BufH, void *Buf, unsigned Count);

void     FIO$ChDir(unsigned NameH, const char *Name);
void     FIO$MkDir(unsigned NameH, const char *Name);
void     FIO$RmDir(unsigned NameH, const char *Name);
void     FIO$GetDir(unsigned short Drive, unsigned NameH, char *Name);

#define FIO_READONLY  0x00H
#define FIO_HIDDEN    0x01H
#define FIO_SYSTEM    0x02H
#define FIO_VOLUME    0x04H
#define FIO_DIRECTORY 0x08H
#define FIO_ARCHIVE   0x10H

typedef struct _DirEntry{
    short          rsvd[21];
    unsigned short attr;
    unsigned       time;
    unsigned       date;
    unsigned long  size;
    char           Name[13];
} DirEntry;

short int  FIO$ReadFirstEntry(unsigned DirNameH,unsigned short Attr, DirEntry *D);
short int  FIO$ReadNextEntry(DirEntry *D);

unsigned char FIO$GetDrive(void);

unsigned char FIO$SetDrive(unsigned char Drive);

unsigned long FIO$GetCurrentDate(void);

unsigned long FIO$GetFileDate(int f);

void          FIO$SetFileDate(int f, unsigned long d);

typedef struct {
  unsigned char Year, Month, Day;
  unsigned char Hour, Min, Sec;
} FileStamp;

unsigned char FIO$GetFileStamp(int f, FileStamp *b);

unsigned char FIO$ThreadEOF(void);

unsigned char FIO$ThreadOK(void);


#ifdef __cplusplus
}
#endif

#pragma restore
#endif
