/* Release 3.00 */

/*-------------------------------------------------------------------------*
*                                                                          *
*  graph.h -  structure, macro and function declarations for graphics      *
*                                                                          *
*  COPYRIGHT (C) 1989..1991 Jensen and Partners. All Rights Reserved.      *
*                                                                          *
*--------------------------------------------------------------------------*/

#ifndef _GRAPH_INC_
#define _GRAPH_INC_

#pragma project("#set cgraph='jpi'")

/* Filling functions take a combination of two colors
   encoded as Color1 + NumColor * Color2 */

/* Device independant routines
   provided as procedure variables
   (see below)
*/

struct videoconfig {
    short numxpixels;       /* number of pixels on X axis */
    short numypixels;       /* number of pixels on Y axis */
    short numtextcols;      /* number of text columns available */
    short numtextrows;      /* number of text rows available */
    short numcolors;        /* number of actual colors */
    short bitsperpixel;     /* number of bits per pixel */
    short numvideopages;    /* number of available video pages */
    short mode;             /* current video mode */
    short adapter;          /* active display adapter */
    short monitor;          /* active display monitor */
    short memory;           /* adapter video memory in K bytes */
};

struct xycoord {
    short xcoord;
    short ycoord;
};

struct rccoord {
    short row;
    short col;
};

#define _DEFAULTMODE    -1      /* restore screen to original mode */
#define _TEXTBW40        0      /* 40 x 25 text, 16 grey */
#define _TEXTC40         1      /* 40 x 25 text, 16/8 color */
#define _TEXTBW80        2      /* 80 x 25 text, 16 grey */
#define _TEXTC80         3      /* 80 x 25 text, 16/8 color */
#define _MRES4COLOR      4      /* 320 x 200, 4 color */
#define _MRESNOCOLOR     5      /* 320 x 200, 4 grey */
#define _HRESBW          6      /* 640 x 200, BW */
#define _TEXTMONO        7      /* 80 x 25 text, BW */
#define _HERCMONO        8      /* 720 x 348, BW for HGC */
#define _MRES16COLOR    13      /* 320 x 200, 16 color */
#define _HRES16COLOR    14      /* 640 x 200, 16 color */
#define _ERESNOCOLOR    15      /* 640 x 350, BW */
#define _ERESCOLOR      16      /* 640 x 350, 4 or 16 color */
#define _VRES2COLOR     17      /* 640 x 480, BW */
#define _VRES16COLOR    18      /* 640 x 480, 16 color */
#define _MRES256COLOR   19      /* 320 x 200, 256 color */

#define _MDPA       0x0001      /* Monochrome Display Adapter (MDPA) */
#define _CGA        0x0002      /* Color Graphics Adapter     (CGA)  */
#define _EGA        0x0004      /* Enhanced Graphics Adapter  (EGA)  */
#define _VGA        0x0008      /* Video Graphics Array       (VGA)  */
#define _MCGA       0x0010      /* MultiColor Graphics Array  (MCGA) */
#define _HGC        0x0020      /* Hercules Graphics Card     (HGC)  */

#define _MONO       0x0001      /* Monochrome */
#define _COLOR      0x0002      /* Color (or Enhanced emulating color) */
#define _ENHCOLOR   0x0004      /* Enhanced Color */
#define _ANALOG     0x0018      /* Analog */

#define _GBORDER         2      /* draw outline only */
#define _GFILLINTERIOR   3      /* fill using current fill mask */

#define _GTOP            1
#define _GNOTOP          0
#define _GCLEARSCREEN    0
#define _GVIEWPORT       1
#define _GWINDOW         2

#define _GCURSOROFF      0
#define _GCURSORON       1

#define _GWRAPOFF        0
#define _GWRAPON         1

#define _GPSET           3
#define _GPRESET         2
#define _GAND            1
#define _GOR             0
#define _GXOR            4

#define _BLACK           0x000000L
#define _BLUE            0x2A0000L
#define _GREEN           0x002a00L
#define _CYAN            0x2a2a00L
#define _RED             0x00002aL
#define _MAGENTA         0x2a002aL
#define _BROWN           0x00152aL
#define _WHITE           0x2a2a2aL
#define _GRAY            0x151515L
#define _LIGHTBLUE       0x3F1515L
#define _LIGHTGREEN      0x153f15L
#define _LIGHTCYAN       0x3f3f15L
#define _LIGHTRED        0x15153fL
#define _LIGHTMAGENTA    0x3f153fL
#define _LIGHTYELLOW     0x153f3fL
#define _BRIGHTWHITE     0x3f3f3fL

#define _MODEFOFF        0L
#define _MODEFOFFTOON    1L
#define _MODEFOFFTOHI    2L
#define _MODEFONTOOFF    3L
#define _MODEFON         4L
#define _MODEFONTOHI     5L
#define _MODEFHITOOFF    6L
#define _MODEFHITOON     7L
#define _MODEFHI         8L

#define _MODE7OFF        0L
#define _MODE7ON         1L
#define _MODE7HI         2L

#ifdef __cplusplus
extern "C" {
#endif
short far          _setvideomode(short _mode);
void far           _enable_herc(void);
short far          _setactivepage(short _page);
short far          _setvisualpage(short _page);
struct videoconfig far * far _getvideoconfig(struct videoconfig far * _v);
struct xycoord far _setlogorg(short _x, short _y);
struct xycoord far _getlogcoord(short _x, short _y);
struct xycoord far _getphyscoord(short _x, short _y);
void far           _setcliprgn(short _x1, short _y1, short _x2, short _y2);
void far           _setviewport(short _x1, short _y1, short _x2, short _y2);
void far           _clearscreen(short _area);
struct xycoord far _moveto(short _x, short _y);
struct xycoord far _getcurrentposition(void);
short far          _lineto(short _x, short _y);
short far          _rectangle(short _fill, short _x1, short _y1, short _x2,
                              short _y2);
short far          _ellipse(short _fill, short _x1, short _y1, short _x2,
                            short _y2);
short far          _arc(short _x1, short _y1, short _x2, short _y2,
                        short _x3, short _y3, short _x4, short _y4);
short far          _pie(short _fill, short _x1, short _y1, short _x2, short _y2,
                        short _x3, short _y3, short _x4, short _y4);
short far          _cube(short _fill, short _top, short _x1, short _y1, short _x2,
                         short _y2, short _depth);
short far          _polygon(short _fill, short _n, short far _px[],
                            short far _py[]);
short far          _setpixel(short _x, short _y);
short far          _getpixel(short _x, short _y);
short far          _floodfill(short _x, short _y, short _boundary);
short far          _stackfill(short _x, short _y, short _boundary);
short far          _setcolor(short _col);
short far          _getcolor(void);
void far           _setlinestyle(unsigned short _mask);
unsigned short far _getlinestyle(void);
void far           _setfillmask(const unsigned char far * _mask_array);
unsigned char far * far _getfillmask(unsigned char far * _mask_array);
long far           _setbkcolor(long _col);
long far           _getbkcolor(void);
long far           _remappalette(short _pixel, long _col);
short far          _remapallpalette(long far * _color_array);
short far          _selectpalette(short _num);
void far           _settextwindow(short _x1, short _y1, short _x2, short _y2);
void far           _outtext(const char far * _str);
short far          _wrapon(short _flag);
short far          _displaycursor(short _flag);
struct rccoord far _settextposition(short _row, short _col);
struct rccoord far _gettextposition(void);
short far          _settextcolor(short _col);
short far          _gettextcolor(void);
void far           _getimage(short _x1, short _y1, short _x2, short _y2,
                             char far * _buffer);
void far           _putimage(short _x, short _y, const char far * _buffer,
                             short _action);
long far           _imagesize(short _x1, short _y1, short _x2, short _y2);
#ifdef __cplusplus
}
#endif

#endif
