/* Release 3.00 */
/*-------------------------------------------------------------------------*
*                                                                          *
*  graphics.h -  structure, macro and function declarartions for BGI       *
*                graphics                                                  *
*                                                                          *
*  COPYRIGHT (C) 1989..1991 Jensen and Partners. All Rights Reserved.      *
*                                                                          *
*--------------------------------------------------------------------------*/

#ifndef _BGI_GRAPH_DEF
#define _BGI_GRAPH_DEF

#pragma project("#set cgraph='borland'")

/*  Error Codes From graphresult */
enum graphics_errors {		
    grOk = 0,
    grNoInitGraph = -1,
    grNotDetected = -2,
    grFileNotFound = -3,
    grInvalidDriver = -4,
    grNoLoadMem = -5,
    grNoScanMem = -6,
    grNoFloodMem = -7,
    grFontNotFound = -8,
    grNoFontMem = -9,
    grInvalidMode = -10,
    grError = -11,
    grIOerror = -12,
    grInvalidFont = -13,
    grInvalidFontNum = -14,
    grInvalidVersion = -18
};

/* Graphics Drivers */
enum graphics_drivers {
    DETECT,                                  /* autodetect */
    CGA, MCGA, EGA, EGA64, EGAMONO, IBM8514, /* 1 - 6 */
    HERCMONO, ATT400, VGA, PC3270,	     /* 7 - 10 */
    CURRENT_DRIVER = -1                      /* default */
};


/* Graphics Modes For Each Driver */
enum graphics_modes {		
    CGAC0      = 0,  /* 320x200 palette 0; 1 page	*/
    CGAC1      = 1,  /* 320x200 palette 1; 1 page	*/
    CGAC2      = 2,  /* 320x200 palette 2: 1 page	*/
    CGAC3      = 3,  /* 320x200 palette 3; 1 page	*/
    CGAHI      = 4,  /* 640x200 1 page			*/
    MCGAC0     = 0,  /* 320x200 palette 0; 1 page	*/
    MCGAC1     = 1,  /* 320x200 palette 1; 1 page	*/
    MCGAC2     = 2,  /* 320x200 palette 2; 1 page	*/
    MCGAC3     = 3,  /* 320x200 palette 3; 1 page	*/
    MCGAMED    = 4,  /* 640x200 1 page			*/
    MCGAHI     = 5,  /* 640x480 1 page			*/
    EGALO      = 0,  /* 640x200 16 color 4 pages	*/
    EGAHI      = 1,  /* 640x350 16 color 2 pages	*/
    EGA64LO    = 0,  /* 640x200 16 color 1 page 	*/
    EGA64HI    = 1,  /* 640x350 4 color  1 page 	*/
    EGAMONOHI  = 0,  /* 640x350 64K on card, 1 page - 256K on card, 4 pages */
    HERCMONOHI = 0,  /* 720x348 2 pages 		*/
    ATT400C0   = 0,  /* 320x200 palette 0; 1 page	*/
    ATT400C1   = 1,  /* 320x200 palette 1; 1 page	*/
    ATT400C2   = 2,  /* 320x200 palette 2; 1 page	*/
    ATT400C3   = 3,  /* 320x200 palette 3; 1 page	*/
    ATT400MED  = 4,  /* 640x200 1 page			*/
    ATT400HI   = 5,  /* 640x400 1 page			*/
    VGALO      = 0,  /* 640x200 16 color 4 pages	*/
    VGAMED     = 1,  /* 640x350 16 color 2 pages	*/
    VGAHI      = 2,  /* 640x480 16 color 1 page 	*/
    PC3270HI   = 0,  /* 720x350 1 page			*/
    IBM8514LO  = 0,  /* 640x480 256 colors		*/
    IBM8514HI  = 1   /* 1024x768 256 colors		*/
};

/* Colors For Setpalette and Setallpalette */

#ifndef _BGI_COLORS
#define _BGI_COLORS

enum COLORS {
    BLACK,		    /* dark colors */
    BLUE,
    GREEN,
    CYAN,
    RED,
    MAGENTA,
    BROWN,
    LIGHTGRAY,
    DARKGRAY,		    /* light colors */
    LIGHTBLUE,
    LIGHTGREEN,
    LIGHTCYAN,
    LIGHTRED,
    LIGHTMAGENTA,
    YELLOW,
    WHITE
};

#endif

enum CGA_COLORS {
    CGA_LIGHTGREEN     = 1,		/* Palette C0 Color Names	*/
    CGA_LIGHTRED       = 2,
    CGA_YELLOW	       = 3,

    CGA_LIGHTCYAN      = 1,		/* Palette C1 Color Names	*/
    CGA_LIGHTMAGENTA   = 2,
    CGA_WHITE	       = 3,

    CGA_GREEN	       = 1,		/* Palette C2 Color Names	*/
    CGA_RED	       = 2,
    CGA_BROWN	       = 3,

    CGA_CYAN	       = 1,		/* Palette C3 Color Names	*/
    CGA_MAGENTA        = 2,
    CGA_LIGHTGRAY      = 3
};

enum EGA_COLORS {
    EGA_BLACK		 =  0,	    /* dark colors */
    EGA_BLUE		 =  1,
    EGA_GREEN		 =  2,
    EGA_CYAN		 =  3,
    EGA_RED		 =  4,
    EGA_MAGENTA 	 =  5,
    EGA_BROWN		 =  20,
    EGA_LIGHTGRAY	 =  7,
    EGA_DARKGRAY	 =  56,     /* light colors */
    EGA_LIGHTBLUE	 =  57,
    EGA_LIGHTGREEN	 =  58,
    EGA_LIGHTCYAN	 =  59,
    EGA_LIGHTRED	 =  60,
    EGA_LIGHTMAGENTA	 =  61,
    EGA_YELLOW		 =  62,
    EGA_WHITE		 =  63
};

/* Linestyles for get/setlinestyle */

enum line_styles {
    SOLID_LINE   = 0,
    DOTTED_LINE  = 1,
    CENTER_LINE  = 2,
    DASHED_LINE  = 3,
    USERBIT_LINE = 4	/* User defined line style */
};

/* Line Widths for get/setlinestyle */
enum line_widths {		
    NORM_WIDTH  = 1,
    THICK_WIDTH = 3
};

enum font_names {
    DEFAULT_FONT	= 0,	/* 8x8 bit mapped font */
    TRIPLEX_FONT	= 1,	/* "Stroked" fonts */
    SMALL_FONT	= 2,
    SANS_SERIF_FONT = 3,
    GOTHIC_FONT	= 4
};

#define HORIZ_DIR	0	/* left to right */
#define VERT_DIR	1	/* bottom to top */

#define USER_CHAR_SIZE	0	/* user-defined char size */

/* Fill Patterns for get/setfillstyle */

enum fill_patterns {		
    EMPTY_FILL,		/* fills area in background color */
    SOLID_FILL,		/* fills area in solid fill color */
    LINE_FILL,		/* --- fill */
    LTSLASH_FILL,	/* /// fill */
    SLASH_FILL,		/* /// fill with thick lines */
    BKSLASH_FILL,	/* \\\ fill with thick lines */
    LTBKSLASH_FILL, 	/* \\\ fill */
    HATCH_FILL,		/* light hatch fill */
    XHATCH_FILL,	/* heavy cross hatch fill */
    INTERLEAVE_FILL,	/* interleaving line fill */
    WIDE_DOT_FILL,	/* Widely spaced dot fill */
    CLOSE_DOT_FILL, 	/* Closely spaced dot fill */
    USER_FILL		/* user defined fill */
};

/* BitBlt Operators For putimage */
enum putimage_ops {		
    COPY_PUT,		/* MOV */
    XOR_PUT,		/* XOR */
    OR_PUT, 		/* OR  */
    AND_PUT,		/* AND */
    NOT_PUT 		/* NOT */
};

/* Horizontal and Vertical Justification for settextjustify */
enum text_just {		
    LEFT_TEXT	= 0,
    CENTER_TEXT	= 1,
    RIGHT_TEXT	= 2,

    BOTTOM_TEXT	= 0,
    TOP_TEXT	= 2
};


#define MAXCOLORS 15

struct palettetype {
    unsigned char size;
    signed char colors[MAXCOLORS+1];
};

struct linesettingstype {
    int linestyle;
    unsigned upattern;
    int thickness;
};

struct textsettingstype {
    int font;
    int direction;
    int charsize;
    int horiz;
    int vert;
};

struct fillsettingstype {
    int pattern;
    int color;
};

struct pointtype {
    int x;
    int y;
};

struct viewporttype {
    int left, top, right, bottom;
    int clip;
};

struct arccoordstype {
    int x, y;
    int xstart, ystart, xend, yend;
};

#ifdef __cplusplus
extern "C" {
#endif
#pragma save
#pragma call(reg_saved => (ds,si,di,st1,st2), reg_param=>(), c_conv=>on)
void far       arc(int x, int y, int stangle, int endangle, int radius);
void far       bar(int left, int top, int right, int bottom);
void far       bar3d(int left, int top, int right, int bottom, int depth,
                     int topflag);
void far       circle(int x, int y, int radius);
void far       cleardevice(void);
void far       clearviewport(void);
void far       closegraph(void);
void far       detectgraph(int far *graphdriver,int far *graphmode);
void far       drawpoly(int numpoints, int far *polypoints);
void far       ellipse(int x, int y, int stangle, int endangle, int xradius, int yradius);
void far       fillellipse( int x, int y, int xradius, int yradius );
void far       fillpoly(int numpoints, int far *polypoints);
void far       floodfill(int x, int y, int border);
void far       getarccoords(struct arccoordstype far *arccoords);
void far       getaspectratio(int far *xasp, int far *yasp);
int  far       getbkcolor(void);
int  far       getcolor(void);
struct palettetype * far  getdefaultpalette(void);
char * far     getdrivername(void);
void far       getfillpattern(char far *pattern);
void far       getfillsettings(struct fillsettingstype far *fillinfo);
int  far       getgraphmode(void);
void far       getimage(int left, int top, int right, int bottom,
	                void far *bitmap);
void far       getlinesettings(struct linesettingstype far *lineinfo);
int  far       getmaxcolor(void);
int  far       getmaxmode(void);
int  far       getmaxx(void);
int  far       getmaxy(void);
char * far     getmodename( int mode_number );
void far       getmoderange(int graphdriver, int far *lomode, int far *himode);
unsigned far   getpixel(int x, int y);
void far       getpalette(struct palettetype far *palette);
int far        getpalettesize(void);
void far       gettextsettings(struct textsettingstype far *texttypeinfo);
void far       getviewsettings(struct viewporttype far *viewport);
int far        getx(void);
int far        gety(void);
void far       graphdefaults(void);
char * far     grapherrormsg(int errorcode);
void far       _graphfreemem(void far *ptr, unsigned size);
void far * far _graphgetmem(unsigned size);
int far        graphresult(void);
unsigned far   imagesize(int left, int top, int right, int bottom);
void far       initgraph(int  far *graphdriver, int  far *graphmode,
                         char far *pathtodriver);
int far        installuserdriver(char far *name, int /* huge */ (*detect)(void));
int far        installuserfont( char far *name );
void far       line(int x1, int y1, int x2, int y2);
void far       linerel(int dx, int dy);
void far       lineto(int x, int y);
void far       moverel(int dx, int dy);
void far       moveto(int x, int y);
void far       outtext(char far *textstring);
void far       outtextxy(int x, int y, char far *textstring);
void far       pieslice(int x, int y, int stangle, int endangle, int radius);
void far       putimage(int left, int top, void far *bitmap, int op);
void far       putpixel(int x, int y, int color);
void far       rectangle(int left, int top, int right, int bottom);
void far       restorecrtmode(void);
void far       sector( int X, int Y, int StAngle, int EndAngle, int XRadius, int YRadius );
void far       setactivepage(int page);
void far       setallpalette(struct palettetype far *palette);
void far       setaspectratio( int xasp, int yasp );
void far       setbkcolor(int color);
void far       setcolor(int color);
void far       setfillpattern(char far *upattern, int color);
void far       setfillstyle(int pattern, int color);
unsigned far   setgraphbufsize(unsigned bufsize);
void far       setgraphmode(int mode);
void far       setlinestyle(int linestyle, unsigned upattern, int thickness);
void far       setpalette(int colornum, int color);
void far       setrgbpalette(int colornum, int red, int green, int blue);
void far       settextjustify(int horiz, int vert);
void far       settextstyle(int font, int direction, int charsize);
void far       setusercharsize(int multx, int divx, int multy, int divy);
void far       setviewport(int left, int top, int right, int bottom, int clip);
void far       setvisualpage(int page);
void far       setwritemode( int mode );
int  far       textheight(char far *textstring);
int  far       textwidth(char far *textstring);

int            registerbgidriver(void (*driver)(void));
int far        registerfarbgidriver(void far *driver);

/*  These "functions" are NOT user-callable
    They are there just so you can link in graphics drivers */

void           CGA_driver(void);
void           EGAVGA_driver(void);
void           IBM8514_driver(void);
void           Herc_driver(void);
void           ATT_driver(void);
void           PC3270_driver(void);

extern int far CGA_driver_far[];
extern int far EGAVGA_driver_far[];
extern int far IBM8514_driver_far[];
extern int far Herc_driver_far[];
extern int far ATT_driver_far[];
extern int far PC3270_driver_far[];

int            registerbgifont(void (*font)(void));
int far        registerfarbgifont(void far *font);

/* These "functions" are NOT user-callable,
   they are there just so you can link in fonts */

void           triplex_font(void);
void           small_font(void);
void           sansserif_font(void);
void           gothic_font(void);

extern int far triplex_font_far[];
extern int far small_font_far[];
extern int far sansserif_font_far[];
extern int far gothic_font_far[];
#pragma restore
#ifdef __cplusplus
}
#endif
#endif
