/* Release 3.00 */
/*-------------------------------------------------------------------------*
*                                                                          *
*  mgraph.h -  header file for Modula-2 graphics procedures                *
*                                                                          *
*  COPYRIGHT (C) 1989..1991 Jensen and Partners. All Rights Reserved.      *
*                                                                          *
*--------------------------------------------------------------------------*/

#ifndef _M2GRAPH_INC
#define _M2GRAPH_INC

#pragma project("#set tsm2='on'")

#pragma save
#pragma call(ds_entry => null)
#pragma name(prefix=>"")
#pragma call(c_conv=>off)


#ifdef __cplusplus
extern "C" {
#endif
void     Graph$GraphMode(void);
void     Graph$TextMode(void);

void     Graph$Plot(unsigned x, unsigned y, unsigned Color);
/* set a single pixel to given color */

unsigned Graph$Point(unsigned x, unsigned y);
/* read a pixel of the screen */

void     Graph$HLine(unsigned x, unsigned y, unsigned x2, unsigned FillColor);
/* draw a horizontal line of given color (used for filling) */


void     Graph$Line(unsigned x1, unsigned y1, unsigned x2, unsigned y2, unsigned Color);
/* draw an arbitrary line in given color */

void     Graph$Circle(unsigned x0, unsigned y0, unsigned r, unsigned c);
/* centre x0,y0; radius r */

void     Graph$Disc(unsigned x0, unsigned y0, unsigned r, unsigned FillColor);
/* filled circle, centre x0,y0; radius r */

void     Graph$TrueCircle(unsigned x0, unsigned y0, unsigned r, unsigned c);
/* centre x0,y0; radius r , adjusted to be "round" for display */

void     Graph$TrueDisc(unsigned x0, unsigned y0, unsigned r, unsigned FillColor);
/* filled circle, centre x0,y0; radius r, adjusted to be "round" for display */

void     Graph$Ellipse( unsigned x0, unsigned y0,  /* center */
                        unsigned a0, unsigned b0,  /* semi-axes */
                        unsigned c,
                        int fill) ; /* wether filled */

void     Graph$Polygon(unsigned n, unsigned px[], unsigned py[], unsigned FillColor);
/* filled polygon of n points */


/*
   Initialization routines.
   Called to setup display output type.
*/
void     Graph$InitCGA(void);
void     Graph$InitEGA(void);
void     Graph$InitVGA(void);
void     Graph$InitHerc(void);
void     Graph$InitGraph(void);
#ifdef __cplusplus
}
#endif

/*
  procedure variables for device dependant routines
*/

/*
    Device specific constants and routines
    equivalent to the variables Width, Depth and NumColors
*/

/* MSDOS Extensions */



typedef struct _videoconfig {
    short numxpixels;       /* number of pixels on X axis */
    short numypixels;       /* number of pixels on Y axis */
    short numtextcols;      /* number of text columns available */
    short numtextrows;      /* number of text rows available */
    short numcolors;        /* number of actual colors */
    short bitsperpixel;     /* number of bits per pixel */
    short numvideopages;    /* number of available video pages */
    short mode;             /* current video mode */
    short adapter;          /* active display adapter */
    short monitor;          /* active display monitor */
    short memory;           /* adapter video memory in K bytes */
} VideoConfig;

typedef struct  {
    short xcoord;
    short ycoord;
} GraphCoords;

typedef struct {
    short row;
    short col;
} TextCoords;

#define _DEFAULTMODE    -1      /* restore screen to original mode */
#define _TEXTBW40       0       /* 40 x 25 text, 16 grey */
#define _TEXTC40        1       /* 40 x 25 text, 16/8 color */
#define _TEXTBW80       2       /* 80 x 25 text, 16 grey */
#define _TEXTC80        3       /* 80 x 25 text, 16/8 color */
#define _MRES4COLOR     4       /* 320 x 200, 4 color */
#define _MRESNOCOLOR    5       /* 320 x 200, 4 grey */
#define _HRESBW         6       /* 640 x 200, BW */
#define _TEXTMONO       7       /* 80 x 25 text, BW */
#define _HERCMONO       8       /* 720 x 348, BW for HGC */
#define _MRES16COLOR    13      /* 320 x 200, 16 color */
#define _HRES16COLOR    14      /* 640 x 200, 16 color */
#define _ERESNOCOLOR    15      /* 640 x 350, BW */
#define _ERESCOLOR      16      /* 640 x 350, 4 or 16 color */
#define _VRES2COLOR     17      /* 640 x 480, BW */
#define _VRES16COLOR    18      /* 640 x 480, 16 color */
#define _MRES256COLOR   19      /* 320 x 200, 256 color */

#define _MDPA       0x0001      /* Monochrome Display Adapter (MDPA) */
#define _CGA        0x0002      /* Color Graphics Adapter     (CGA)  */
#define _EGA        0x0004      /* Enhanced Graphics Adapter  (EGA)  */
#define _VGA        0x0008      /* Video Graphics Array       (VGA)  */
#define _MCGA       0x0010      /* MultiColor Graphics Array  (MCGA) */
#define _HGC        0x0020      /* Hercules Graphics Card     (HGC)  */

#define _MONO       0x0001      /* Monochrome */
#define _COLOR      0x0002      /* Color (or Enhanced emulating color) */
#define _ENHCOLOR   0x0004      /* Enhanced Color */
#define _ANALOG     0x0018      /* Analog */

#define _GBORDER         2      /* draw outline only */
#define _GFILLINTERIOR   3      /* fill using current fill mask */

#define _GTOP            1
#define _GNOTOP          0
#define _GCLEARSCREEN    0
#define _GVIEWPORT       1
#define _GWINDOW         2

#define _GCURSOROFF      0
#define _GCURSORON       1

#define _GWRAPOFF        0
#define _GWRAPON         1

#define _GPSET           3
#define _GPRESET         2
#define _GAND            1
#define _GOR             0
#define _GXOR            4

#define _BLACK           0x000000L
#define _BLUE            0x2A0000L
#define _GREEN           0x002a00L
#define _CYAN            0x2a2a00L
#define _RED             0x00002aL
#define _MAGENTA         0x2a002aL
#define _BROWN           0x00152aL
#define _WHITE           0x2a2a2aL
#define _GRAY            0x151515L
#define _LIGHTBLUE       0x3F1515L
#define _LIGHTGREEN      0x153f15L
#define _LIGHTCYAN       0x3f3f15L
#define _LIGHTRED        0x15153fL
#define _LIGHTMAGENTA    0x3f153fL
#define _LIGHTYELLOW     0x153f3fL
#define _BRIGHTWHITE     0x3f3f3fL

#define _MODEFOFF        0L
#define _MODEFOFFTOON    1L
#define _MODEFOFFTOHI    2L
#define _MODEFONTOOFF    3L
#define _MODEFON         4L
#define _MODEFONTOHI     5L
#define _MODEFHITOOFF    6L
#define _MODEFHITOON     7L
#define _MODEFHI         8L

#define _MODE7OFF        0L
#define _MODE7ON         1L
#define _MODE7HI         2L

#ifdef __cplusplus
extern "C" {
#endif
extern int Graph$FillState;

void       Graph$GetVideoConfig(VideoConfig *v);
void       Graph$SetClipRgn(unsigned x1, unsigned y1, unsigned x2,
                            unsigned y2);
long       Graph$GetBkColor(long);
void       Graph$GetFillMask(unsigned makH, unsigned char *Mask);
unsigned   Graph$GetLinestyle(void);
long       Graph$SetBkColor(long Color);
void       Graph$SetFillMask(unsigned MaskH, unsigned char *Mask);
void       Graph$SetLinestyle(unsigned Mask);
int        Graph$DisplayCursor(int Toggle);
unsigned   Graph$GetTextColor(void);
TextCoords Graph$GetTextPosition(void);
unsigned   Graph$SetTextColor(unsigned Color);
TextCoords Graph$SetTextPosition(unsigned row, unsigned col);
void       Graph$SetTextWindow(unsigned r1, unsigned c1, unsigned r2,
                               unsigned c2);
int        Graph$Wrapon(int Opt);
void       Graph$OutText(unsigned textH, char *Text);
int        Graph$SetVideoMode(unsigned Mode);
unsigned   Graph$SetActivePage(unsigned Page);
unsigned   Graph$SetVisualPage(unsigned Page);
void       Graph$ClearScreen(unsigned Area);
void       Graph$Rectangle(unsigned x1, unsigned y1, unsigned x2, unsigned y2, unsigned Color, unsigned Fill);
void       Graph$Arc(unsigned x1, unsigned y1, unsigned x2, unsigned y2, unsigned x3, unsigned y3,
                     unsigned x4, unsigned y4, unsigned Color);
void       Graph$Pie(unsigned x1, unsigned y1, unsigned x2, unsigned y2, unsigned x3, unsigned y3,
                     unsigned x4, unsigned y4, unsigned Colr, unsigned  Fill);
void       Graph$FloodFill(unsigned x, unsigned y, unsigned Color, unsigned Boundary);
void       Graph$StackFill(unsigned x, unsigned y,  unsigned Color, unsigned Boundary);
long       Graph$RemapPalette(unsigned Pixel, unsigned long Color);
unsigned   Graph$RemapAllPalette(unsigned ColH, long *Colarray);
unsigned   Graph$SelectPalette(unsigned Palnum);
void       Graph$GetImage(unsigned x1, unsigned y1, unsigned x2, unsigned y2, void *Buffer);
void       Graph$PutImage(unsigned x, unsigned y, void *Buffer, int Action);
long       Graph$ImageSize(unsigned x1, unsigned y1, unsigned x2, unsigned y2);
void       Graph$Cube(int top, unsigned x1, unsigned y1, unsigned x2, unsigned y2, unsigned depth, unsigned  Color, int Fill);
#ifdef __cplusplus
}
#endif

#pragma restore

#endif

