/* Release 3.00 */
/*-------------------------------------------------------------------------*
*                                                                          *
*  mouse.h -  structure and function definitions for MSDOS and OS2         *
*             mouse support                                                *
*                                                                          *
*  COPYRIGHT (C) 1989..1991 Jensen and Partners. All Rights Reserved.      *
*                                                                          *
*--------------------------------------------------------------------------*/
#ifndef _MOUSE_INC_
#define _MOUSE_INC_


struct _ms_data {
    int left_pressed;
    int middle_pressed;
    int right_pressed;
    int actions;
    int row;
    int col;
};

struct _ms_range {
    int max_col;
    int min_col;
    int max_row;
    int min_row;
};

struct _ms_motion {
    int vert;
    int horiz;
};


struct _ms_graphcur {
    int row;
    int col;
    unsigned int screen_mask[16];
    unsigned int cursor_mask[16];
};

struct _hardware {
    int start_scan;
    int end_scan;
};

struct _software {
    unsigned screen_mask;
    unsigned cursor_mask;
};

union _ms_type {
    struct _hardware h;
    struct _software s;
};

struct _ms_textcur {
    int type;
    union _ms_type c;
};

struct _ms_sense {
    int h_speed;
    int v_speed;
    int threshold;
};

#ifndef __STDC__
struct _ms_interrupt {
    void (far * func)(void);
    unsigned mask;
};
#endif

#define _MS_ON    1
#define _MS_OFF   0
#define _MS_HIDE  0
#define _MS_SHOW  1

#define _MS_MIDDLE 2
#define _MS_LEFT   0
#define _MS_RIGHT  1

#define _MS_HARD  1
#define _MS_SOFT  0


#ifdef __cplusplus
extern "C" {
#endif
int      _ms_reset(void);
void     _ms_cursor(int mode);
unsigned _ms_driversize(void);
void     _ms_getmotion(struct _ms_motion *_mp);
int      _ms_getpage(void);
void     _ms_getpress(int button, struct _ms_data *_mp);
void     _ms_getrelease(int button, struct _ms_data *_mp);
void     _ms_getsensitivity(struct _ms_sense *_mp);
void     _ms_getstatus(struct _ms_data *_mp);
void     _ms_lightpen(int _mode);
void     _ms_restoredriver(void *_buffer);
void     _ms_savedriver(void *_buffer);
void     _ms_setdouble(int _threshhold);
void     _ms_setgraphcursor(struct _ms_graphcur *_mp);
void     _ms_setinterrupt(struct _ms_interrupt *_mp);
void     _ms_setmickeys(int _vert, int _horiz);
void     _ms_setpage(int _page);
void     _ms_setposition(int _row, int _col);
void     _ms_setrange(struct _ms_range *_mp);
void     _ms_setsensitivity(struct _ms_sense *_mp);
void     _ms_settextcursor(struct _ms_textcur *_mp);
void     _ms_swapinterrupt(struct _ms_interrupt *_mp);
void     _ms_updatescreen(int _x1, int _y1, int _x2, int _y2);
#ifdef __cplusplus
}
#endif

#endif


