/* Release 3.00 */
/*-------------------------------------------------------------------------*
*                                                                          *
*  mwindow.h -  header file for Modula-2 window  procedures                *
*                                                                          *
*  COPYRIGHT (C) 1989..1991 Jensen and Partners. All Rights Reserved.      *
*                                                                          *
*--------------------------------------------------------------------------*/

#ifndef _M2WIN_INC
#define _M2WIN_INC

#pragma project("#set tsm2='on'")

#pragma save
#pragma name(prefix=>"")
#pragma call(c_conv=>off)

#define TRUE  1
#define FALSE 0

#define  SCREENWIDTH 80
#define  SCREENDEPTH 25
#define  PALETTESIZE 10
#define  PALETTEMAX  PALETTESIZE

#ifndef _WINTYPES_DEF
#define _WINTYPES_DEF
typedef struct WDS *wintype;

typedef unsigned  relcoord;
typedef unsigned  abscoord;

typedef enum { Black,     Blue,         Green,      Cyan,
        Red,       Magenta,      Brown,      LightGray,
        DarkGray,  LightBlue,    LightGreen, LightCyan,
        LightRed,  LightMagenta, Yellow,     White } Color ;

typedef  char titlestr[SCREENWIDTH];
typedef  char framestr[9];

typedef  struct {
    abscoord X1, Y1, X2, Y2;        /* outer co-ordinates of opposite corners */
    Color Foreground, Background;   /* not Used if Palette */
    int CursorOn;                   /* if cursor active    */
    int WrapOn;                     /* if EOL wrap enabled */
    int Hidden;                     /* if window on view   */
    int FrameOn;                    /* if frame            */
    framestr FrameDef;              /* only Used if frame  */
    Color FrameFore,
    FrameBack;                      /* only Used if frame
                                    and not Palette Window */
} windef;

typedef enum { NoTitle,
              LeftUpperTitle,CenterUpperTitle,RightUpperTitle,
              LeftLowerTitle,CenterLowerTitle,RightLowerTitle } TitleMode;

#define SINGLEFRAME  "Ŀ"
#define DOUBLEFRAME  "ͻͼ"

typedef  struct {
    Color Fore, Back;
} PaletteColorDef ;

typedef PaletteColorDef PaletteDef[PALETTEMAX];

#define  NORMALPALETTECOLOR  0
#define  FRAMEPALETTECOLOR   1

typedef  unsigned *bufferptr;

typedef  struct WDS{
    unsigned Guard;
    wintype Next;
    wintype CursorChain;
    windef WDef;
    abscoord XA, YA, XB, YB;
    unsigned OWidth, ODepth;
    unsigned Width, Depth;
    char *Title;
    TitleMode TMode;
    relcoord CurrentX, CurrentY;
    int IsPalette;
    unsigned char PalAttr[PALETTEMAX];
    int CurPalColor;
    bufferptr Buffer;
    void *UserRecord;
}  WinDescriptor;
#endif

#ifdef __cplusplus
extern "C" {
#endif

extern windef fullscreendef;
extern wintype _fullscreen;

wintype  Window$Open(windef WD);
void     Window$SetTitle(wintype _win, char _NewTitle[], TitleMode _Mode );
void     Window$SetFrame(wintype _win, framestr _Frame, Color _Fore, Color _Back);
void     Window$Use(wintype _win);
void     Window$PutOnTop(wintype _win);
void     Window$PutBeneath(wintype _win, wintype _WA);
void     Window$Hide(wintype _win);
void     Window$Change(wintype _win, abscoord _X1, abscoord _Y1, abscoord _X2, abscoord _Y2 );
void     Window$Snapshot(void);
void     Window$Close(wintype *_win);
wintype  Window$Used(void);
wintype  Window$Top(void);
int      Window$ObscuredAt(wintype _win, relcoord _X, relcoord _Y);
wintype  Window$At(abscoord _X, abscoord _Y);
void     Window$GotoXY(relcoord X, relcoord _Y);
relcoord Window$WhereX(void);
relcoord Window$WhereY(void);
void     Window$ConvertCoords(wintype _win, relcoord _X, relcoord _y, abscoord *_XO, abscoord *_YO);
void     Window$ClrScr(void);
void     Window$CursorOn(void);
void     Window$CursorOff(void);
void     Window$InsLine(void);
void     Window$DelLine(void);
void     Window$ClrEol(void);
void     Window$TextColor(Color _c);
void     Window$TextBackground (Color _c);
void     Window$SetWrap(int _on);
void     Window$Info(wintype _win, windef *_WD);
void     Window$RdBufferLn(wintype _win, relcoord _X, relcoord _Y, void *_Dest, unsigned _Len);
void     Window$WrBufferLn(wintype _win, relcoord _X, relcoord _Y, void *_src, unsigned _Len);

/* Palette Windows                                                     */


wintype  Window$PaletteOpen(windef *_WD, PaletteDef _Pal);
void     Window$SetPalette(wintype _win, PaletteDef _Pal);
int      Window$PaletteColor(void);
void     Window$PSetPaletteColor (int _c);
int      Window$PPaletteColorUsed(wintype _win, int _pc);
#ifdef __cplusplus
}
#endif


#pragma restore
#endif
