/* Release 3.00 */
/*-------------------------------------------------------------------------*
*                                                                          *
*  paslib.h -  header file for PasLib                                      *
*                                                                          *
*  COPYRIGHT (C) 1989..1991 Jensen and Partners. All Rights Reserved.      *
*                                                                          *
*--------------------------------------------------------------------------*/

#ifndef _PASLIB_INC
#define _PASLIB_INC

#include <stdio.h>

#pragma project("#set tspas='on'")

#pragma save
#pragma name(prefix=>"")
#pragma call(c_conv=>off)
#pragma name(upper_case=>on)

#ifdef __cplusplus
extern "C" {
#endif
typedef  char String80[81];           /* Byte 0 contains length */
typedef  char maxstring[256];         /* Byte 0 contains length */
typedef  String80 convstring;         /* Byte 0 contains length */
typedef  String80 namestring;         /* Byte 0 contains length */
typedef  enum  {rm,wm} ifilemode;
typedef  enum  {Initialized, TextFile, TempFile, EOLinBuffer, Device, Assigned, ACrt, App} States;

typedef struct  _filetype {
                char *bufferptr;        /* must be first field of RECORD          */
                unsigned objectsize;    /* must be second field of RECORD         */
                unsigned char state;    /* must be third 3 field of RECORD, zero  */
                                        /* length name indicates a temporary file */
                unsigned char mode;
                unsigned handle;
                String80 name;
                FILE *stream;
              } filetype;


extern filetype  PASLIB$_input;
extern filetype  PASLIB$_outut;
extern unsigned char  PASLIB$winio;
extern unsigned char  PASLIB$ClearOnAllocate;
extern unsigned char  PASLIB$IOCheck;
extern unsigned char  PASLIB$IsoFormatting;
extern int (*  PASLIB$HeapError)(unsigned);
extern long  PASLIB$RandSeed;
extern unsigned PASLIB$IOBufferSize;
extern unsigned PASLIB$FileMode;

#define  DefaultBufferSize  512

void  PASLIB$_allocate(char **p, unsigned s);
void  PASLIB$_deallocate(char *p, unsigned s);
void  PASLIB$HALT(void);

void  PASLIB$_wrint(long i, int w);
void  PASLIB$_wrbool(unsigned char b, int w);
void  PASLIB$_wrchar(char c, int w);
void  PASLIB$_wrastr(unsigned size, unsigned low, unsigned high, char * s, int w);
void  PASLIB$_wrreal(long double r, int w, int d);
void  PASLIB$_wrcstr(int l, const char *s, int w);
void  PASLIB$_wrln(void);
void  PASLIB$_page(void);
unsigned char  PASLIB$_eoln(void);
unsigned char  PASLIB$_eof(void);

void  PASLIB$_f_wrint(filetype *f,long i, int w);
void  PASLIB$_f_wrbool(filetype *f,unsigned char b, int w);
void  PASLIB$_f_wrchar(filetype *f,char c, int w);
void  PASLIB$_f_wrastr(filetype *f,unsigned size, unsigned low, unsigned high, char * s, int w);
void  PASLIB$_f_wrreal(filetype *f,long double r, int w, int d);
void  PASLIB$_f_wrcstr(filetype *f,int l, const char *s, int w);
void  PASLIB$_f_wrln(filetype *f);
void  PASLIB$_f_wrbin(filetype *f, const void *v);
void  PASLIB$_f_page(filetype *f);
unsigned char  PASLIB$_f_eoln(filetype *f);
unsigned char  PASLIB$_f_eof(filetype *f);

void  PASLIB$_rdint(int *i);
void  PASLIB$_rdchar(char *c);
void  PASLIB$_rdbool(unsigned char *b);
void  PASLIB$_rdstr(unsigned char size, maxstring s, int len);
void  PASLIB$_rdreal(long double *d);
void  PASLIB$_rdln(void);

void  PASLIB$_f_rdint(filetype *f,int *i);
void  PASLIB$_f_rdchar(filetype *f,char *c);
void  PASLIB$_f_rdbool(filetype *f,unsigned char *b);
void  PASLIB$_f_rdstr(filetype *f,unsigned char size, maxstring, int len);
void  PASLIB$_f_rdreal(filetype *f,long double *d);
void  PASLIB$_f_rdln(filetype *f);
void  PASLIB$_f_rdbin(filetype *f, void *v);

void  PASLIB$_rewrite(filetype *f, unsigned rs, unsigned char IsText);
void  PASLIB$_append(filetype *f, unsigned rs, unsigned char IsText);
void  PASLIB$_reset(filetype *f, unsigned rs, unsigned char IsText);

void  PASLIB$assign(filetype *f, unsigned char size, namestring name);
void  PASLIB$put(filetype *f);
void  PASLIB$get(filetype *f);
void  PASLIB$close(filetype *f);

/* I/O extensions */

long PASLIB$FilePos(filetype *f);
long PASLIB$FileSize(filetype *f);
void PASLIB$Seek(filetype *f, long pos);
void PASLIB$Flush(filetype *f);
unsigned PASLIB$IOresult(void);

void   PASLIB$Rename(filetype *f, unsigned char size, char * newname );
void   PASLIB$Truncate(filetype *f);
unsigned char PASLIB$Exists(filetype *f);
void   PASLIB$Erase(filetype *f);

typedef struct _DateTime {
  unsigned Year, Month, Day, Hour, Min, Sec;
} DateTime;

void  PASLIB$PackTime(DateTime *dt, long *Time);
void  PASLIB$UnpackTime(long Time, DateTime *DT);
void  PASLIB$GetFTime(char *f, long *Time);
void  PASLIB$SetFTime(char *f, long *Time);

void  PASLIB$GetFAttr(char *f, unsigned *Attr);
void  PASLIB$SetFAttr(char *f, unsigned Attr);

void  PASLIB$SetIOR(unsigned code);

/******************/

unsigned char  PASLIB$_strcompare(char *s1, char *s2);

void  PASLIB$_bindfile(filetype *f, unsigned n);

long double   PASLIB$sin(long double r);
long double   PASLIB$cos(long double r);
long double   PASLIB$exp(long double r);
long double   PASLIB$log10(long double r);
long double   PASLIB$pow(long double x, long double y);
long double   PASLIB$ln(long double r);
long double   PASLIB$sqrt(long double r);
long double   PASLIB$arctan(long double r);
long double   PASLIB$frac(long double r);
long double   PASLIB$sqr(long double r);

/* Math Extensions */

long double   PASLIB$arccos(long double x);
long double   PASLIB$arcsin(long double x);
long double   PASLIB$arctan2(long double x, long double  y);
long double   PASLIB$cosh(long double x);
long double   PASLIB$sinh(long double x);
long double   PASLIB$tanh(long double x);
long double   PASLIB$tan(long double x);

long double   PASLIB$frexp(long double x, int *expo);

long double   PASLIB$hypot(long double x, long double y);

void     PASLIB$randomize(void);
unsigned PASLIB$random(unsigned range);

void  PASLIB$FreeMem(char **p, unsigned size);
void  PASLIB$GetMem(char **p, unsigned size);
long  PASLIB$MaxAvail(void);
long  PASLIB$MemAvail(void);

void      PASLIB$FillChar(void *x, unsigned count, unsigned char ch);
void      PASLIB$Move(void *source, void *dest, unsigned count);
unsigned  PASLIB$Compare(void *Source, void *Dest, unsigned Len);
unsigned  PASLIB$ScanR(void *Dest, unsigned Count, unsigned char Value);
unsigned  PASLIB$ScanL(void *Dest, unsigned Count, unsigned char Value);
unsigned  PASLIB$ScanNeR(void *Dest, unsigned Count, unsigned char Value);
unsigned  PASLIB$ScanNeL(void *Dest, unsigned Count, unsigned char Value);

void PASLIB$IntToStr(long V, unsigned char size, char *convstring, unsigned Base, unsigned char *OK);

long PASLIB$StrToInt(unsigned char size, const char *S, unsigned Base, unsigned char *OK);

long double PASLIB$StrToReal(unsigned char size, const char *S, unsigned char *OK);

void PASLIB$RealToStr(long double RN, int Digits, int *Width, unsigned char size,
                   const char * S, unsigned char *OK);

void     PASLIB$Caps(unsigned char size, char * s1);
unsigned PASLIB$CharPos(unsigned char size, const char * Source, char c);
void     PASLIB$Delete(unsigned char size, char * s, unsigned index, unsigned count);
void     PASLIB$Insert(unsigned char ssize, char * source, unsigned char tsize,
                char *t, unsigned index);
void     PASLIB$Lows(unsigned char size, char * s1);
unsigned char PASLIB$Match(unsigned char size, const char * Source, unsigned char Psize,
                   const char * Pattern);
unsigned char PASLIB$NextPos(unsigned char bsize, const char *substr,
                      unsigned char ssize, const char * s, unsigned char place);
unsigned char PASLIB$Pos(unsigned char bsize, const char * substr,
                  unsigned char ssize, char * s);
unsigned PASLIB$RCharPos(unsigned char size, const char * Source, char c);
#ifdef __cplusplus
}
#endif

#pragma restore
#endif

