/* Release 3.00 */
/*-------------------------------------------------------------------------*
*                                                                          *
*  process.h -	header for process control functions.			   *
*                                                                          *
*  COPYRIGHT (C) 1989..1991 Jensen and Partners. All Rights Reserved.      *
*                                                                          *
*--------------------------------------------------------------------------*/

#ifndef _PROCESS_INC_
#define _PROCESS_INC_


/*  Modes available as first argument of the spawn? functions. */

#define P_WAIT	  0	/* child runs separately, parent waits */
#define P_NOWAIT  1	/* both concurrent -- not implemented under MSDOS */
#define P_OVERLAY 2	/* child replaces parent */

#ifdef __cplusplus
extern "C" {
#endif
extern unsigned _psp;

void	   abort(void);
int	   atexit(void (* func)(void));
void   (*  onexit(void (* func)(void)))();
#ifdef _mthread
unsigned   _dosbeginthread(void (* _func)(void), int _stlen, unsigned *_stck);
void       _dosendthread(void);
int far    _beginthread(void (far cdecl *func)(void far *arg),
                        void far *stack, unsigned stack_size, void far *arg);
void       _endthread(void);
void       _dosfreestack(unsigned _stck);
#endif

#ifndef _WINDOWS
int        execl(const char *_path, char *_arg0, ...);
int        execle(const char *_path, char *_arg0, ...);
int        execlp(const char *_path, char *_arg0, ...);
int        execlpe(const char *_path, char *_arg0, ...);
int        execv(const char *_path, char *_argv[]);
int        execve(const char *_path, char *_argv[], char *_env[]);
int        execvp(const char *_path, char *_argv[]);
int        execvpe(const char *_path, char *_argv[], char *_env[]);
#endif

void	   exit(int _status);
void       _exit(int _status);
void       delay(int _msecs);
unsigned   getpid(void);
#ifdef _mthread
unsigned   _getTID(void);
#endif

#ifndef _WINDOWS
int        spawnl(int _mode, const char *_path, char *_arg0, ...);
int        spawnle(int _mode, const char *_path, char *_arg0, ...);
int        spawnlp(int _mode, const char *_path, char *_arg0, ...);
int        spawnlpe(int _mode, const char *_path, char *_arg0, ...);
int        spawnv(int _mode, const char *_path, char *_argv[]);
int        spawnve(int _mode, const char *_path, char *_argv[], char *_env[]);
int        spawnvp(int _mode, const char *_path, char *_argv[]);
int        spawnvpe(int _mode, const char *_path, char *_argv[], char *_env[]);
int        system(const char *_command);
#endif
#ifdef __cplusplus
}
#endif

#ifdef _mthread

/* This structure contains thread globals */

#define MAXTHREAD 32

#ifndef _TM_DEF

struct tm {
    int  tm_sec;
    int  tm_min;
    int  tm_hour;
    int  tm_mday;
    int  tm_mon;
    int  tm_year;
    int  tm_wday;
    int  tm_yday;
    int  tm_isdst;
};
#define _TM_DEF		
#endif

struct thread   {
     int error;
     char _asct_buf[28];	 /* static time conversion buffer */
     struct tm _tm_buf;		 /* static time conversion structure */
     char *token;
};
#endif

#define  MaxProcess  32

#ifndef _PROCTYPE_DEF
#define _PROCTYPE_DEF
typedef void *PROCESS;
#endif

struct _TaskDescriptor;

typedef struct _SigRec {
    int count;
    struct _TaskDescriptor *waiting;
    unsigned long sem;
} SigRec;

typedef SigRec *SIGNAL;

typedef struct _TaskDescriptor {
    struct _TaskDescriptor *next;
    unsigned priority;
    PROCESS cor;
    unsigned due;
    void far *wsp;
    unsigned wspsize;
    struct _TaskDescriptor *nextdue;
} TaskDescriptor;

typedef TaskDescriptor *Task;

#ifdef __cplusplus
extern "C" {
#endif
void   StartProcess(void (far *_P)(void), unsigned _N, unsigned _Pr);
#pragma save
#pragma call(reg_saved=>(ds,di,si,st1,st2))
void   StartScheduler(void);
void   StopScheduler(void);
void   StopProcess(void);
void   SEND(SIGNAL _s);
void   WAIT(SIGNAL _s);
void   Notify(SIGNAL _s);
int    Awaited(SIGNAL _s);
void   Init(SIGNAL *_s);
void   ProcessDelay(unsigned _T);
#define Delay(_t) ProcessDelay(_t)
int MakeReadyProcess(void (far * _P)(void), unsigned _N, unsigned _Pr, Task *_newp);
void StartReadyProcess(Task _p);
extern void (*_core_delay)(unsigned _msecs);
#pragma restore

#pragma save
#pragma call(c_conv=>off, near_call=>off, reg_saved => (ax,bx,cx,dx,ds,si,di,st1,st2))
void Lock(void);
void Unlock(void);
extern void (*_core_lock)(void);
extern void (*_core_unlock)(void);
#pragma restore
#ifdef __cplusplus
}
#endif

#endif
