/* Release 3.00 */
/*-------------------------------------------------------------------------*
*                                                                          *
*  task.hpp -  header file for OOP multi-thread interface.                 *
*                                                                          *
*  COPYRIGHT (C) 1990..1991 Jensen and Partners. All Rights Reserved       *
*                                                                          *
*--------------------------------------------------------------------------*/

#ifndef _TASK_INC_
#define _TASK_INC_
#include <process.h>


typedef struct {
  volatile unsigned char Started;
  volatile unsigned char Destroyed;
} task_control;


class task {

public:

  void start();
  // start thread
  int task_id() { return thread_no; }
  // return thread id
  int active() { return thread_active != 0; }
  // returns non-zero if thread is running.
  void kill();
  // sends signal to thread to stop.
  virtual ~task();

protected:

  static void stop();
  // stop thread - can only be called from thread proc
  static int destroyed();
  // returns non-zero if kill request has been sent - can only be called from thread proc
  static void set_priority(unsigned);
  // set thread priority
  static unsigned get_priority();
  // get thread priority

protected:
  void (*proc_adr())();
  // returns address of thread member function

  void init(void (*func)(), unsigned _stack_size=2000, unsigned _priority=1, int _start=0);

  task();

private:

    task(task& );  // causes RTE

  task& operator=(task& );  // causes RTE

  static void thread_proc();  // thread process
  static task_control tc[MaxProcess];
  Task np;
  int thread_no;
  int thread_active;
  int initialized;
};


class semaphore {

public:

  enum sem_state { _clear=0, _set=1, _failed=-1 };

      semaphore();
    virtual  ~semaphore() {}

  sem_state set();
  // unconditionally sets semaphore
  sem_state clear();
  // unconditionally clears semaphore
  sem_state wait(int _msecs = -1);
  // waits for semaphore to be cleared
  sem_state set_and_wait(int _msecs = -1);
  // unconditionally sets semaphore and waits for it to be cleared
  sem_state request(int _msecs = -1);
  // waits for semaphore to be cleared and then sets

private:

  volatile sem_state sem;
};


class critical_region {

public:

      critical_region();
      critical_region(semaphore*);
  virtual ~critical_region();

  void begin();
  void end();

protected:

  semaphore * get_sem() { return s; }

private:

      critical_region(critical_region&);         // causes RTE
  critical_region& operator=(critical_region& ); // causes RTE

  semaphore *s;
  int n;
  int locked_thread;
  int alloc;
};

class event {

public:

      event();
      event(semaphore*);
  virtual ~event();

  void wait();
  void trigger();
  void notify();
  void reset();
  int  awaited() { return count < 0;}

protected:

  semaphore * get_sem() { return s; }

private:

      event(event&);         // causes RTE
  event& operator=(event& ); // causes RTE

  semaphore *s;
  critical_region cr;
  int count;
  int alloc;
};

#endif

