/* Release 3.00 */
/*-------------------------------------------------------------------------*
*                                                                          *
*  window.h - definitions for window module types and functions.           *
*                                                                          *
*  COPYRIGHT (C) 1989..1991 Jensen and Partners. All Rights Reserved       *
*                                                                          *
*--------------------------------------------------------------------------*/

#ifndef _WINDOWS
#ifndef _TSWIN_INC_
#define _TSWIN_INC_

#pragma project("#set cwindow='jpi'")

#define TRUE  1
#define FALSE 0

#define  SCREENWIDTH 80
#define  SCREENDEPTH 25
#define  PALETTESIZE 10
#define  PALETTEMAX  PALETTESIZE

#ifndef _WINTYPES_DEF
#define _WINTYPES_DEF
typedef struct WDS *wintype;

typedef unsigned  relcoord;
typedef unsigned  abscoord;

typedef enum { Black,     Blue,         Green,      Cyan,
        Red,       Magenta,      Brown,      LightGray,
        DarkGray,  LightBlue,    LightGreen, LightCyan,
        LightRed,  LightMagenta, Yellow,     White } Color ;

typedef  char titlestr[SCREENWIDTH];
typedef  char framestr[9];

typedef  struct {
    abscoord X1, Y1, X2, Y2;        /* outer co-ordinates of opposite corners */
    unsigned char Foreground, Background;   /* not Used if Palette */
    unsigned char CursorOn;                   /* if cursor active    */
    unsigned char WrapOn;                     /* if EOL wrap enabled */
    unsigned char Hidden;                     /* if window on view   */
    unsigned char FrameOn;                    /* if frame            */
    framestr FrameDef;              /* only Used if frame  */
    unsigned char FrameFore,
    FrameBack;                      /* only Used if frame
                                    and not Palette Window */
} windef;

typedef enum { NoTitle,
              LeftUpperTitle,CenterUpperTitle,RightUpperTitle,
              LeftLowerTitle,CenterLowerTitle,RightLowerTitle } TitleMode;

#define SINGLEFRAME  "Ŀ"
#define DOUBLEFRAME  "ͻͼ"

typedef  struct {
    unsigned char Fore, Back;
} PaletteColorDef ;

typedef PaletteColorDef PaletteDef[PALETTEMAX];

#define  NORMALPALETTECOLOR  0
#define  FRAMEPALETTECOLOR   1

typedef  unsigned *bufferptr;

typedef  struct WDS{
    unsigned Guard;
    wintype Next;
    wintype CursorChain;
    windef WDef;
    abscoord XA, YA, XB, YB;
    unsigned OWidth, ODepth;
    unsigned Width, Depth;
    char *Title;
    unsigned char TMode;
    relcoord CurrentX, CurrentY;
    unsigned char IsPalette;
    unsigned char PalAttr[PALETTEMAX];
    unsigned char CurPalColor;
    bufferptr Buffer;
    void *UserRecord;
}  WinDescriptor;

#endif

#ifdef __cplusplus
extern "C" {
#endif
extern windef  fullscreendef;
extern wintype _fullscreen;
extern unsigned CurrentScreenDepth;

wintype        windowopen(windef *WD);
void           settitle(wintype _win, char _NewTitle[], TitleMode _Mode );
void           setframe(wintype _win, framestr _Frame, Color _Fore,
                        Color _Back);
void           use(wintype _win);
void           putontop(wintype _win);
void           putbeneath(wintype _win, wintype _WA);
void           hide(wintype _win);
void           change(wintype _win, abscoord _X1, abscoord _Y1, abscoord _X2,
                      abscoord _Y2 );
void           snapshot(void);
void           windowclose(wintype _win);
wintype        used(void);
wintype        top(void);
int            obscuredat(wintype _win, relcoord _X, relcoord _Y);
wintype        at(abscoord _X, abscoord _Y);
void           gotoxy(relcoord X, relcoord _Y);
relcoord       wherex(void);
relcoord       wherey(void);
void           convertcoords(wintype _win, relcoord _X, relcoord _y,
                             abscoord *_XO, abscoord *_YO);
void           clrscr(void);
void           cursoron(void);
void           cursoroff(void);
void           insline(void);
void           delline(void);
void           clreol(void);
void           textcolor(Color _c);
void           textbackground (Color _c);
void           _directwrite(relcoord _X, relcoord _Y, const void *_A, unsigned _Len);
void           setwrap(int _on);
void           info(wintype _win, windef *_WD);
void           _rdbufferln(wintype _win, relcoord _X, relcoord _Y,
                           void *_Dest, unsigned _Len);
void           _wrbufferln(wintype _win, relcoord _X, relcoord _Y, void *_src,
                           unsigned _Len);

/* Palette Windows                                                     */

wintype        paletteopen(windef *_WD, PaletteDef _Pal);
void           setpalette(wintype _win, PaletteDef _Pal);
int            palettecolor(void);
void           setpalettecolor (int _c);
int            palettecolorused(wintype _win, int _pc);
#ifdef __cplusplus
}
#endif

#endif
#endif
