/* Release 3.00 */

/*-------------------------------------------------------------------------*
*                                                                          *
*  coniostr.cpp - console I/O stream class implementations.                *
*                                                                          *
*  COPYRIGHT (C) 1989..1991 Jensen and Partners. All Rights Reserved       *
*  Written by Simon Knight.                                                *
*                                                                          *
*--------------------------------------------------------------------------*/

#define _USE_CONIO
#include <clib.h>
#include <coniostr.hpp>


#pragma module(init_prio=>22)

extern "C" void _init_con(void) {

  conout.setf(ios::unitbuf);
}

static conbuf inbuf;
static conbuf outbuf;

istream conin(&inbuf);
ostream conout(&outbuf);

int conbuf::underflow() {

  int nr, c;
  char *buf;

#ifdef _mthread
  Lock();
#endif
  nr=ebuf()-eback();
  if(nr == 1) {
    c=getche();
    if(c == EOF) {
#ifdef _mthread
        Unlock();
#endif
      return EOF;
    }
    *eback()=c;
  }
  else {
    buf=new char[nr+3];
    buf[0]=nr;
    cgets(buf);
    nr=buf[1];
    memcpy(eback(), buf, nr);
    delete buf;
  }
  setg(eback(), eback(), eback()+nr);
#ifdef _mthread
  Unlock();
#endif
  return *gptr();
}

int conbuf::pbackfail(int c) {

  return ungetch(c);
}

extern "C" unsigned _flush_init_magic;

int conbuf::overflow(int c) {

  char *pt;
  int nw;

#ifdef _mthread
    Lock();
#endif
  flush_magic=_flush_init_magic;
  if(unbuffered()) {
    if((c == EOF) || (putch(c) == EOF)) {
#ifdef _mthread
      Unlock();
#endif
      return EOF;
    }
  }
  else {
    pt=pbase();
    nw=pptr()-pt;
    while(nw--) {
      if(putch(*pt) == EOF) {
#ifdef _mthread
        Unlock();
#endif
        return EOF;
      }
      ++pt;
    }
  }
  setp(pbase(), epptr());
  if((c != EOF) && (!unbuffered()))
    sputc(c);
#ifdef _mthread
    Unlock();
#endif
  return c;
}

int conbuf::sync() {

#ifdef _mthread
  Lock();
#endif
  overflow();
  setp(base(), ebuf());
  setg(base(), ebuf(), ebuf());
#ifdef _mthread
  Unlock();
#endif
  return 0;
}

