/* Release 3.00 */
/*    Copyright (C) 1989..1991 Jensen and Partners. All Rights Reserved */

#ifdef _OS2
#define _USE_OS2
#include <os2def.h>
#include <bsedos.h>
#endif

#include <dos.h>
#include <io.h>
#include <string.h>
#include <pragma.h>
#include <stdlib.h>

#ifndef _OS2
#pragma _JPI_CALL

#pragma check(stack=>off, index=>off, nil_ptr=>off)
#define FIRST_SEG (sizeof(struct ProcessState)+sizeof(unsigned)+(sizeof(struct seg_header)*seg_total))
#define THIS_SEG_HEADER (sizeof(struct ProcessState)+sizeof(unsigned)+(sizeof(struct seg_header)*seg_num))
#define TOTAL_POS (sizeof(struct ProcessState))
#define MAX_BUCKETS 16
#define BUCKET_SIZE 0x1000U

struct seg_header {
    unsigned frame;
    unsigned seg_size;
    unsigned long file_pos;
    unsigned bucket_sizes[MAX_BUCKETS];
};

struct ProcessState {
    unsigned PSP;        /* PSP segement. 0 if OS2 */
    unsigned TopLim;     /* End of memory block. 0 if OS2  */
    unsigned BPreg;      /* BP contents */
    unsigned DSreg;      /* DS contents */
    unsigned DIreg;      /* DI contents */
    unsigned SIreg;      /* SI contents */
    unsigned CSreg;      /* CS contents */
    unsigned IPreg;      /* IP contents */
    unsigned SSreg;      /* SS contents */
    unsigned SPreg;      /* SP contents */
    unsigned Code;       /* Error code */
    char far *Msg;       /* Error string. Null if not present */
    char FPstate[94];    /* 087 state */
    char ProgName[80];   /* Program Name */
};

extern "C" {

void _PMDbody(struct ProcessState *state);

unsigned lsl(unsigned sel);

#pragma save,call(near_call=>on)
int write_struct(struct ProcessState *state, int fh);
void write_core(int fh, unsigned top, unsigned bottom);
int compress(struct seg_header *h, int handle);
#pragma restore

#pragma save,call(near_call=>off,c_conv=>off,reg_param=>(),reg_saved=>(ds,di,si,st1,st2))
#pragma name(prefix=>"EnvPack$")
int Pack(int size, void far *buffer);
#pragma name(prefix=>"EnvPack@")
extern char far p[];
#pragma restore
}

char err_msg[]="Runtime Error - Creating PM_DUMP.$$$\r\n";

void _PMDbody(struct ProcessState *state)

{
    int fh;

    if(_argc)
        strcpy(state->ProgName, _argv[0]);
    fh=_creat("PM_DUMP.$$$", FA_NORMAL);
    if(fh >= 0)
    {
        _write(2, err_msg, strlen(err_msg));
        if(write_struct(state, fh) == 0)
            write_core(fh, state->TopLim, state->PSP);
        _close(fh);
    }
    return;
}

int write_struct(struct ProcessState *state, int fh)

{
    if(_write(fh, (char *) state, sizeof(struct ProcessState)) != sizeof(struct ProcessState))
        return(-1);
    return(0);
}

void write_core(int fh, unsigned top, unsigned bottom)

{
    struct seg_header h;
    unsigned total_paras;
    unsigned segment=bottom;
    unsigned seg_total=0;
    unsigned seg_num=0;
    unsigned num_to_write;

    total_paras=top-bottom;
    seg_total=(total_paras+0xFFF)/ 0x1000;
        lseek(fh, FIRST_SEG-1, SEEK_SET);
    _write(fh, &segment, 1);
    while(total_paras)
    {
        h.frame=segment;
        num_to_write=min(total_paras, 0x1000);
        h.seg_size=num_to_write<<4;             /* 64K = 0 */
        h.file_pos=lseek(fh, 0, SEEK_END);
        if(compress(&h, fh) == -1)
        {
            seg_total=0;
            break;
        }
        lseek(fh, THIS_SEG_HEADER, SEEK_SET);
        ++seg_num;
        _write(fh, (char *)&h, sizeof(struct seg_header));
        segment+=num_to_write;
        total_paras-=num_to_write;
    }
    lseek(fh, TOTAL_POS, SEEK_SET);
    _write(fh, (char *) &seg_total, sizeof(unsigned));
    return;
}

int compress(struct seg_header *header, int fh)

{
    long seg_size;
    unsigned offset;
    unsigned out_size, bucket_no, bucket_size, num_writ;

    seg_size=header->seg_size;
    if(seg_size == 0)
        seg_size=0x10000L;
    offset=0;
    bucket_no=0;
    while(seg_size > 0)
    {
        if(seg_size > BUCKET_SIZE)
            out_size=BUCKET_SIZE;
        else
            out_size= (unsigned) seg_size;
        bucket_size=Pack(out_size, MK_FP(header->frame, offset));
        header->bucket_sizes[bucket_no]=bucket_size;
        if(_dos_write(fh, (char far *) p, bucket_size, &num_writ))
            return(-1);
        seg_size-=BUCKET_SIZE;
        offset+=BUCKET_SIZE;
        ++bucket_no;
    }
    return(0);
}

#endif

#ifdef _OS2

#pragma _JPI_CALL

#pragma check(stack=>off, index=>off, nil_ptr=>off)

#define FIRST_SEG (sizeof(struct ProcessState)+sizeof(unsigned)+(sizeof(struct seg_header)*seg_total))
#define THIS_SEG_HEADER (sizeof(struct ProcessState)+sizeof(unsigned)+(sizeof(struct seg_header)*segment))
#define TOTAL_POS (sizeof(struct ProcessState))
#define MAX_BUCKETS 16
#define BUCKET_SIZE 0x1000U
#define MAX_SELS 256

struct seg_header {
    unsigned frame;
    unsigned seg_size;
    unsigned long file_pos;
    unsigned bucket_sizes[MAX_BUCKETS];
};

struct ProcessState {
    unsigned PSP;        /* PSP segement. 0 if OS2 */
    unsigned TopLim;     /* End of memory block. 0 if OS2  */
    unsigned BPreg;      /* BP contents */
    unsigned DSreg;      /* DS contents */
    unsigned DIreg;      /* DI contents */
    unsigned SIreg;      /* SI contents */
    unsigned CSreg;      /* CS contents */
    unsigned IPreg;      /* IP contents */
    unsigned SSreg;      /* SS contents */
    unsigned SPreg;      /* SP contents */
    unsigned Code;       /* Error code */
    char far *Msg;       /* Error string. Null if not present */
    char FPstate[94];    /* 087 state */
    char ProgName[80];   /* Program Name */
};

struct sel_info {
    unsigned selector;
    unsigned size;
};

extern "C" {
void _PMDbody(struct ProcessState *state);

unsigned lsl(unsigned sel);

#pragma save
#pragma call(near_call=> on)
int write_struct(struct ProcessState *state, int fh);
void write_core(int fh);
int compress(struct seg_header *h, int handle);
int get_selectors(void);
int os2_write(int handle, char far *buffer, int count);
#pragma restore

#pragma save
#pragma call(near_call=> off, c_conv => off, reg_param => (), reg_saved => (ds, di, si,st1,st2))
#pragma name(prefix => "EnvPack$")
int Pack(int size, void far *buffer);
#pragma name(prefix => "EnvPack@")
extern char far p[];
#pragma restore
}

char err_msg[]="Runtime Error - Creating PM_DUMP.$$$\r\n";
struct sel_info selectors[MAX_SELS];

void _PMDbody(struct ProcessState *state)

{
    int fh;
    int action;

    if(_argc)
        strcpy(state->ProgName, _argv[0]);
    if(DosOpen((PSZ) "PM_DUMP.$$$", (PHFILE) &fh, (PUSHORT) &action, 0L,
                0, 0x12, 0x12, 0L) == 0);
    {
        os2_write(2, (char far *) err_msg, strlen(err_msg));
        if(write_struct(state, fh) == 0)
            write_core(fh);
        _close(fh);
    }
    return;
}

int write_struct(struct ProcessState *state, int fh)

{
    if(os2_write(fh, (char far *) state, sizeof(struct ProcessState)) != sizeof(struct ProcessState))
        return(-1);
    return(0);
}

void write_core(int fh)

{
    struct seg_header h;
    unsigned segment=0;
    unsigned seg_total;

    seg_total=get_selectors();
    lseek(fh, FIRST_SEG-1, SEEK_SET);
    os2_write(fh, (char far *) &segment, 1);
    while(segment < seg_total)
    {
        h.frame=selectors[segment].selector;
        h.seg_size=selectors[segment].size;
        h.file_pos=lseek(fh, 0, SEEK_END);
        if(compress(&h, fh) == -1)
        {
            seg_total=0;
            break;
        }
        lseek(fh, THIS_SEG_HEADER, SEEK_SET);
        os2_write(fh, (char far *)&h, sizeof(struct seg_header));
        segment+=1;
    }
    lseek(fh, TOTAL_POS, SEEK_SET);
    os2_write(fh, (char far *) &seg_total, sizeof(unsigned));
    return;
}

int compress(struct seg_header *header, int fh)

{
    long seg_size;
    unsigned offset;
    unsigned out_size, bucket_no, bucket_size;

    seg_size=header->seg_size;
    if(seg_size == 0)
        seg_size=0x10000L;
    offset=0;
    bucket_no=0;
    while(seg_size > 0)
    {
        if(seg_size > BUCKET_SIZE)
            out_size=BUCKET_SIZE;
        else
            out_size= (unsigned) seg_size;
        bucket_size=Pack(out_size, MK_FP(header->frame, offset));
        header->bucket_sizes[bucket_no]=bucket_size;
        if(os2_write(fh, (char far *) p, bucket_size) == -1)
            return(-1);
        seg_size-=out_size;
        offset+=BUCKET_SIZE;
        ++bucket_no;
    }
    return(0);
}

int get_selectors(void)

{
    int cur_sel;
    unsigned attempt;
    unsigned size;

    cur_sel=0;
    attempt=0xF;
    while(attempt != 0xFFFF)
    {
        if((size=lsl(attempt)) != 0)
        {
            selectors[cur_sel].size=size+1;
            selectors[cur_sel].selector=attempt;
            if(++cur_sel == MAX_SELS)
                return(cur_sel);
        }
        attempt+=0x10;
    }
    attempt=0x7;
    while(attempt != 0xFFF7)
    {
        if((size=lsl(attempt)) != 0)
        {
            selectors[cur_sel].size=size+1;
            selectors[cur_sel].selector=attempt;
            if(++cur_sel == MAX_SELS)
                return(cur_sel);
        }
        attempt+=0x10;
    }
    return(cur_sel);
}

int os2_write(int handle, char far *buffer, int count)

{
    unsigned num_writ;

    if(DosWrite(handle, buffer, count, (PUSHORT) &num_writ) == 0)
        return(num_writ);
    return(-1);
}

#endif
