(* Release 3.00 *)

(* Copyright (C) 1989..1991 Jensen & Partners International *)

(*# call(seg_name => PASPROC) *)
(*# module(implementation=>off) *)
(*# check(stack=>off,
          index=>off,
          range=>off,
          overflow=>off,
          nil_ptr=>off) *)


IMPLEMENTATION UNIT PasProc;

(*%F _OS2 *)
IMPORT paslib, CoreSig, PasInline;
(*%E *)
(*%T _OS2 *)
IMPORT PasDos,Dos,CoreSig, PasInline;
(*%E *)

(*%F _OS2 *)
procedure QInsert(T: CoreProc.Task; var FQ: CoreProc.Task);
(* inserts task after last task in FQ with greater or equal priority *)
var
  q,qb : CoreProc.Task;
begin
  q  := FQ;
  qb := NIL;
  while (q<>NIL) and (T^.priority<=q^.priority) do
    begin
    qb := q;
    q := q^.next;
    end;
  if qb=NIL then
    FQ := T
  else
    qb^.next := T;
  T^.next := q;
end;


procedure AddReadyProcess(T: CoreProc.Task);
(* adds new process to ready list
   NB gets added ahead of current process if at same priority
*)
var
  mp    : PROCESS;
  oldcp : CoreProc.Task;
  ie    : word;
begin
  ie := PasInline.GetFlags; PasInline.DI;
  if CoreProc._cp=NIL then
    QInsert(T,CoreProc._cp)          (* add new process *)
  else
    begin
    oldcp := CoreProc._cp;            (* remove current process *)
    CoreProc._cp := CoreProc._cp^.next;
    QInsert(T,CoreProc._cp);          (* add new process *)
    QInsert(oldcp,CoreProc._cp);      (* add current process *)
    end;
  mp := CoreProc._cp^.cor;
  CoreProc.TRANSFER (mp,mp);
  PasInline.SetFlags(ie);
end ;



procedure StartProcess{(P: PROC; N: word; Pr: word)};

type
  ByteArray = array [0..16] of byte;
  BytePtr = ^ByteArray;
var
  np : CoreProc.Task;
  tno: word;
label
  555;
begin
  tno := 1;
  while true do
    begin
    if tno = MaxProcess then
      CoreSig._FatalError(CoreSig._FatalErrorPos, 040H);
    if CoreProc._ProcIds[tno] = 0 then
      goto 555;
    tno := tno + 1;
    end;
555:
  CoreProc._ProcIds[tno] := tno + 1;
  tno := tno + 1;
  Pr := Pr + 1;
  if N < 32000 then
    N :=  N + 512;                                       (* FP Overhead *)
  paslib._allocate (np, size(CoreProc.TaskDescriptor));
  paslib._allocate (np^.wsp, N);
  (np^.wsp:>BytePtr)^[16] := tno;
  np^.priority := Pr;
  np^.wspsize := N ;
  CoreProc.NEWPROCESS (P, np^.wsp, N , np^.cor);
  AddReadyProcess(np);
end ;

(* new stop me procedure *)

procedure StopProcess;
var
  t0 : CoreProc.Task;
  mp : PROCESS;
  ie : word;
begin
  CoreProc._ProcIds[CoreProc._getTID - 1] := 0;
  ie := PasInline.GetFlags; PasInline.DI;
  t0 := CoreProc._cp ;
  CoreProc._cp := CoreProc._cp^.next;
  paslib._deallocate(t0^.wsp,t0^.wspsize);
  paslib._deallocate(t0,SIZE(CoreProc.TaskDescriptor));
  mp := CoreProc._cp^.cor;
  CoreProc.TRANSFER (mp,mp);
  PasInline.SetFlags(ie);
end ;



procedure SEND{(s: SIGNAL)};
var
  t0 : CoreProc.Task;
  ie : word;
begin
  ie := PasInline.GetFlags; PasInline.DI;
  if s^.count <> maxint16 then
    begin
    s^.count := s^.count + 1;
    if s^.count <= 0 then (* somebody waiting *)
      begin
      t0 := s^.waiting;
      s^.waiting := t0^.next;
      AddReadyProcess(t0);
      end;
    end;
  PasInline.SetFlags(ie);
end ;


procedure Notify{(s: SIGNAL)};
var
  t0 : CoreProc.Task;
  ie : word;
begin
  ie := PasInline.GetFlags; PasInline.DI;
  if s^.count < 0 then (* somebody waiting *)
    begin
    s^.count := s^.count + 1;
    t0 := s^.waiting;
    s^.waiting := t0^.next;
    (* add to waiting q *)
    t0^.nextdue := CoreProc._wq;
    CoreProc._wq := t0;
    end;
  PasInline.SetFlags(ie);
end ;


procedure WAIT{(s: SIGNAL)};
var
  t0 : CoreProc.Task;
  mp : PROCESS;
  ie : word;
begin (* insert CoreProc._cp in queue s *)
  ie := PasInline.GetFlags; PasInline.DI;
  s^.count := s^.count - 1;
  if s^.count < 0 then (* wait *)
    begin
    t0 := CoreProc._cp;
    CoreProc._cp := CoreProc._cp^.next;
    QInsert(t0,s^.waiting);
    mp := CoreProc._cp^.cor;
    CoreProc.TRANSFER (mp,mp);
    end;
  PasInline.SetFlags(ie);
end ;

function Awaited{(s: SIGNAL) : BOOLEAN};
begin
  Awaited := (s^.count < 0);
end ;


procedure Init{(var s: SIGNAL)};
begin
  NEW(s);
  s^.waiting := NIL;
  s^.count := 0;
end ;


procedure CheckTimeQ;
var
  ta,tb,tn : CoreProc.Task;
begin
  ta := CoreProc._dq;
  tb := NIL;
  while ta <> NIL do
    begin
    tn := ta^.nextdue;
    if ta^.due = CoreProc._SchedTime then
      begin
      if tb = NIL then
        CoreProc._dq := tn
      else
        tb^.nextdue := tn;
      ta^.nextdue := CoreProc._wq;
      CoreProc._wq := ta
      end
    else
      tb := ta;
    ta := tn;
    end;
end ;

procedure Slice;
(* Clears waiting queue *)
(* Then schedules next ready process if it is of equal priority *)
var
  nextt,oldt,ta : CoreProc.Task;
begin
  if CoreProc._LockNestMonster = 0 then
    begin
    (* move waiting queue to the ready queue *)
    (* set up by CheckTimeQueue *)
    while CoreProc._wq <> NIL do
      begin
      ta := CoreProc._wq; CoreProc._wq := CoreProc._wq^.nextdue;
      QInsert(ta,CoreProc._cp);
      end;
    (* now do slice *)
    nextt := CoreProc._cp^.next;
    if (nextt <> NIL) AND (nextt^.priority = CoreProc._cp^.priority) then (* slice *)
      begin
      oldt := CoreProc._cp; CoreProc._cp := nextt;
      QInsert(oldt,CoreProc._cp); (* insert old CoreProc._cp at end of processes *)
      end;
    end;
end ;

procedure Scheduler;
var
  nextt,
  oldt  : CoreProc.Task;
  op,np : PROCESS;

label
  555;
begin
  PasInline.DI;
  CoreProc.Stop := false;
  CoreProc._SchedTime := 0;
  while true do
    begin
    np := CoreProc._cp^.cor;
    while true do
      begin
      CoreProc.IOTRANSFER(op,np,8);
      CoreProc.Int8;
      CoreProc._SchedTime := CoreProc._SchedTime + 1;
      if CoreProc.Stop then
        goto 555;
      CheckTimeQ;
      Slice;
      np := CoreProc._cp^.cor;
      end;
555:
    CoreProc.Stop := false;
    CoreProc.TRANSFER(op,np); (* no return until restarted *)
    end;
end;


procedure Idler; (* always on CoreProc._cp chain *)
var
  i : word;
begin
  while true do
    i := i + 1;
end ;

procedure StartScheduler;
var
  ie : word;
begin
  ie := PasInline.GetFlags; PasInline.DI;
  if NOT CoreProc._Started then
    begin
    CoreProc._Started := true;
    if CoreProc._SchedStack = NIL then (* first time *)
      begin
      paslib._allocate( CoreProc._SchedStack, 512 );
      CoreProc.NEWPROCESS( Scheduler, CoreProc._SchedStack, 512, CoreProc._SchedProc );
      end;
    CoreProc.TRANSFER( CoreProc._cp^.cor, CoreProc._SchedProc );
    end;
  PasInline.SetFlags(ie);
end ;

procedure StopScheduler;
var
  ie : word;
begin
  ie := PasInline.GetFlags;
  if CoreProc._Started then
    begin
    PasInline.EI ;
    CoreProc._Started := FALSE;
    CoreProc.Stop := TRUE;
    while CoreProc.Stop do;
    end;
  PasInline.SetFlags(ie);
end ;

procedure Delay{(T: word)};
(* Waits T time slices *)
(* 0 will swap to next process of equal priority, without delaying *)
var
  mp : PROCESS;
  ie : word;
label
  999;
begin
  ie := PasInline.GetFlags; PasInline.DI;
  if T = 0 then
    begin
    mp := CoreProc._cp^.cor;
    Slice;
    if mp = CoreProc._cp^.cor then
      begin
      PasInline.SetFlags(ie);
      goto 999;
      end  (* no other processes ready *)
    end
  else
    begin
    CoreProc._cp^.due := CoreProc._SchedTime + T;
    CoreProc._cp^.nextdue := CoreProc._dq;
    CoreProc._dq := CoreProc._cp;
    CoreProc._cp := CoreProc._cp^.next;
    end;
  mp := CoreProc._cp^.cor;
  CoreProc.TRANSFER(mp,mp);
  PasInline.SetFlags(ie);
999:
end ;


procedure Lock;
(* Critical region lock - prevents timeslicing *)
(* may be nested *)
begin
  CoreProc._LockNestMonster := CoreProc._LockNestMonster + 1;
end ;

procedure Unlock;
(* Unlock procedure, always paired with a call to Lock.
   Will de-schedule current process if there are ready processes
   of equal priority *)

var
  ie : word;
begin
  ie := PasInline.GetFlags; PasInline.DI;
  if CoreProc._LockNestMonster <= 1 then
    begin
    CoreProc._LockNestMonster := 0;
    Delay(0)
    end
  else
    CoreProc._LockNestMonster := CoreProc._LockNestMonster - 1;
  PasInline.SetFlags(ie);
end ;
(*%E *)

(*%T _OS2 *)
procedure ErrorNamed ( IOR: word; Code: word; Msg: maxstring);
var
  s  : ARRAY [0..100] OF CHAR;

begin
  if IOR <> 0 then
    PasDos.RunTimeError(CoreSig._FatalErrorPos, Code, Msg);
end;


procedure StartScheduler;
var
  i: 0..MaxProcess;
begin
  Lock;
  i := 1;
  while i <= CoreProc._LastProc do
    begin
    if CoreProc._ProcIds[i] <> 0 then
      ErrorNamed(Dos.ResumeThread(CoreProc._ProcIds[i]), 0C0H, 'DosResumeThread');
    end;
  CoreProc._Started := true;
  Unlock;
end ;

procedure StartProcess{(P: PROC; N: word; Pr: word)};

var
  ID: word;
begin
  Lock;
  Pr := Pr - 1;
  CoreProc._LastProc := CoreProc._LastProc + 1;
  if CoreProc._LastProc = MaxProcess then
    PasDos.RunTimeError(CoreSig._FatalErrorPos, 0C0H, '');
  if N < 32000 then
    N := N + 512;
  ID := CoreProc._dosbeginthread(P, N, NIL);
  if ID < 0 then
    PasDos.RunTimeError(CoreSig._FatalErrorPos, 0C1H, '');
  CoreProc._ProcIds[ID-1] := ID;
  ErrorNamed(Dos.SetPrty(2,0,Pr,ID), 0C2H, 'DosSetPrty');
  if not CoreProc._Started then
    ErrorNamed(Dos.SuspendThread(ID), 0C3H, 'DosSuspendThread');
  Unlock;
end ;


procedure StopProcess;

var
  t: word;
begin
  t := CoreProc._getTID;
  CoreProc._ProcIds[t] := 0;
  CoreProc._dosendthread;
end ;

procedure StopScheduler;

begin
end ;


procedure Send{(s: SIGNAL)};
begin
  Lock;
  if s^.count < 0 then
     ErrorNamed(Dos.SemClear(FarADR(s^.sem)), 0C4H, 'DosSemClear');
  s^.count := s^.count + 1 ;
  Unlock;
end ;

procedure WAIT{(s: SIGNAL)};
begin
  Lock;
  s^.count := s^.count - 1 ;
  if s^.count < 0 then
    begin
    Unlock;
    ErrorNamed(Dos.SemRequest(FarADR(s^.sem),-1), 0C5H, 'DosSemRequest');
    end
  else
    Unlock;
end ;

procedure Notify{(s: SIGNAL)};
begin
  Lock;
  if s^.count < 0 then
    begin
    ErrorNamed(Dos.SemClear(FarADR(s^.sem)), 0C6H, 'DosSemClear');
    s^.count := s^.count + 1 ;
    end;
  UnLock;
end ;

function Awaited{(s: SIGNAL) : BOOLEAN};
begin
  Awaited := (s^.count < 0);
end ;


procedure Lock;
var
 tid : word ;
begin
 tid := CoreProc._getTID;
 if (CoreProc._LockCount=0)OR(CoreProc._LockedThread<>tid) then
   begin
   Dos.SemRequest(FarADR(CoreProc._LockSem),-1);
   CoreProc._LockedThread := tid ;
   end;
 CoreProc._LockCount := CoreProc._LockCount + 1;
end ;

procedure Unlock;
begin
  CoreProc._LockCount := CoreProc._LockCount - 1;
  if CoreProc._LockCount=0 then
    begin
    Dos.SemClear(FarADR(CoreProc._LockSem));
    Dos.Sleep(0);
    end;
end ;


procedure Init{(var s: SIGNAL)};
begin
  new(s);
  ErrorNamed(Dos.SemSet(FarADR(s^.sem)), 0C7H, 'DosSemSet');
  s^.count := 0;
end ;

procedure Delay{(T: word)};
begin
  ErrorNamed(Dos.Sleep(T*50), 0C8H, 'DosSleep');
end ;

procedure Idler;

begin
end ;
(*%E *)

begin
(*%F _OS2 *)
  if CoreProc._proc_init = false then
    begin
    CoreProc.initprocess;
    CoreProc._dq := NIL;
    CoreProc._wq := NIL;
    paslib._allocate(CoreProc._cp, SIZE(CoreProc.TaskDescriptor));
    if CoreProc._cp = NIL then
      CoreSig._FatalError(CoreSig._FatalErrorPos, 044H);
    CoreProc._cp^.next := NIL;
    CoreProc._cp^.priority := 2;
    CoreProc._cp^.cor  := CoreProc.CurrentProcess;
    StartProcess( Idler, 512, MAX(word));
    CoreProc._SchedStack := NIL;
    CoreProc._Started  := FALSE;
    CoreProc._LockNestMonster := 0;
    CoreProc._NextThread:=2;
    CoreProc._exit_proc:=StopScheduler;
    end;
(*%E *)
(*%T _OS2 *)
  CoreProc._Started := FALSE;
  CoreProc._LastProc := 0;
  CoreProc._LockCount := 0;
  CoreProc._LockSem := 0;
  CoreProc._LockedThread := 0 ;
(*%E *)
  CoreProc._core_lock:=Lock;
  CoreProc._core_unlock:=Unlock;
  CoreProc._core_delay:=Delay;
end.
