/* Release 3.00 */

/*-------------------------------------------------------------------------*
*                                                                          *
*  textwin.cpp - window class implementations.      		           *
*                                                                          *
*  COPYRIGHT (C) 1989..1991 Jensen and Partners. All Rights Reserved       *
*  Written by Simon Knight.                                                *
*                                                                          *
*--------------------------------------------------------------------------*/
#define _USE_WINDOW
#include <clib.h>
#include <textwin.hpp>

#pragma module(init_prio=>22)

int (*WinErrHandler)();

window FullScreen=_fullscreen;

window::window(window& W) {

  windef wd;

  ::info(W.handle(), &wd);
  h=windowopen(&wd);
  check_win();
}

window::window(wintype WT) {

  if(WT) {
    h=WT;
    check_win();
  }
}

window::window(windef& WD) {

  h=windowopen(&WD);
  check_win();
}

window::window(abscoord X1, abscoord Y1, abscoord X2, abscoord Y2,
           Color fore, Color back) {

  windef WD;

  WD.X1=X1;
  WD.Y1=Y1;
  WD.X2=X2;
  WD.Y2=Y2;
  WD.Foreground=fore;
  WD.Background=back;
  WD.WrapOn=TRUE;
  WD.FrameOn=FALSE;
  WD.Hidden=FALSE;
  h=windowopen(&WD);
  check_win();
}

window::window(abscoord X1, abscoord Y1, abscoord X2, abscoord Y2,
           framestr  frame,
           Color framefore, Color frameback,
           Color fore, Color back) {

  windef WD;

  WD.X1=X1;
  WD.Y1=Y1;
  WD.X2=X2;
  WD.Y2=Y2;
  WD.Foreground=fore;
  WD.Background=back;
  strncpy(WD.FrameDef, frame, 9);
  WD.FrameFore=framefore;
  WD.FrameBack=frameback;
  WD.WrapOn=TRUE;
  WD.FrameOn=TRUE;
  WD.Hidden=FALSE;
  h=windowopen(&WD);
  check_win();
}

window::~window() {

  windowclose(h);
  h=NULL;
}

void  window::operator=(window& W) {

  windef wd;

  if(h) {
    windowclose(h);
    h=NULL;
  }
  ::info(W.handle(), &wd);
  h=windowopen(&wd);
  check_win();
}

void  window::settitle(char NewTitle[], TitleMode Mode ) {

  check_win();
  ::settitle(h, NewTitle, Mode);
}

void  window::setframe(framestr Frame, Color Fore, Color Back) {

  check_win();
  ::setframe(h, Frame, Fore, Back);
}

void  window::use() {

  check_win();
  ::use(h);
}

void  window::putontop() {

  check_win();
  ::putontop(h);
}

void  window::putbeneath(window& w) {

  check_win();
  ::putbeneath(h, w.handle());
}

void  window::hide() {

  check_win();
  ::hide(h);
}

void  window::change(abscoord X1, abscoord Y1, abscoord X2, abscoord Y2 ) {

  check_win();
  ::change(h, X1, Y1, X2, Y2);
}

void  window::close() {

  check_win();
  windowclose(h);
  h=NULL;
}

int window::isused() {

  check_win();
  return ( used() == h );
}

int window::istop() {

  check_win();
  return ( top() == h );
}

int  window::obscuredat(relcoord X, relcoord Y) {

  check_win();
  return  ::obscuredat(h, X, Y);

}

void  window::convertcoords(relcoord X, relcoord Y, abscoord& XO, abscoord& YO) {

  check_win();
  ::convertcoords(h, X, Y, &XO, &YO);
}

void  window::info(windef& NWD) {

  check_win();
  ::info(h, &NWD);
}



palettewindow::palettewindow(PaletteDef Pal, windef& WD) : window((wintype)NULL) {

  h=paletteopen(&WD, Pal);
  check_win();
}

palettewindow::palettewindow(PaletteDef Pal,
                  abscoord X1, abscoord Y1, abscoord X2, abscoord Y2,
                  Color fore, Color back) : window((wintype)NULL) {

  windef WD;

  WD.X1=X1;
  WD.Y1=Y1;
  WD.X2=X2;
  WD.Y2=Y2;
  WD.Foreground=fore;
  WD.Background=back;
  WD.WrapOn=TRUE;
  WD.FrameOn=FALSE;
  WD.Hidden=FALSE;
  h=paletteopen(&WD, Pal);
  check_win();
}

palettewindow::palettewindow(PaletteDef Pal,
                  abscoord X1, abscoord Y1, abscoord X2, abscoord Y2,
                  framestr  frame,
                  Color framefore, Color frameback,
                  Color fore, Color back) : window((wintype)NULL) {

  windef WD;

  WD.X1=X1;
  WD.Y1=Y1;
  WD.X2=X2;
  WD.Y2=Y2;
  WD.Foreground=fore;
  WD.Background=back;
  strncpy(WD.FrameDef, frame, 9);
  WD.FrameFore=framefore;
  WD.FrameBack=frameback;
  WD.WrapOn=TRUE;
  WD.FrameOn=TRUE;
  WD.Hidden=FALSE;
  h=paletteopen(&WD, Pal);
  check_win();
}

void  palettewindow::setpalette(PaletteDef Pal) {

  check_win();
  ::setpalette(h, Pal);
}

int   palettewindow::palettecolorused(int pc) {

  check_win();
  return ::palettecolorused(h, pc);
}

void window::check_win() {

  if(!h) {
    if((WinErrHandler) && (WinErrHandler()))
      return;
    else
      InternalError(0x30);
  }
  _checkwindow(h);
}


