(* Release 3.00 *)
(* Copyright (C) 1989..1991 Jensen & Partners International *)

(*%F _fdata *)
(*# call(seg_name => null) *)
(*%E *)
(*# module(implementation=>off) *)
(*# data(seg_name => null) *)
(*# check(stack=>off,
          index=>off,
          range=>off,
          overflow=>off,
          nil_ptr=>off) *)

IMPLEMENTATION UNIT TurboSys;

(*%F _OS2 *)
IMPORT PasLib,TurboDos,PasDos,CoreSig,CoreMath,CoreFile,CoreMain,CoreIO,PasCore(*%T _mthread *),PasProc(*%E *);
(*%E *)
(*%T _OS2 *)
IMPORT PasLib,Dos,TurboDos,PasDos,CoreSig,CoreMath,CoreFile,CoreMain,CoreIO,PasCore(*%T _mthread *),PasProc(*%E *);
(*%E *)

procedure Mark{(var p: pointer)}; in PasLow;
procedure Release{(var p:address)}; in PasLow;
function  Swap{(x: word): word}; in PasLow;

type
  CharPtr = ^char;
  FileInf = ^ CoreFile.CStream;

(************************* Miscellaneous functions ****************************)

procedure Abort{(code: word)};

begin
  TurboDos.SetExitCode(code);
  HALT;
end;

function  Hi{(v: integer): byte};

begin
  Hi := (v >> 8) and 0FFH;
end;


function  Int{(x: longreal): longreal};

begin
  Int := CoreMath.floor(x);
end;


function  Lo{(v: integer): byte};

begin
  Lo := v and 8;
end;




function  Pi{: longreal};

begin
  Pi := 3.1415926535897932385;
end;



procedure RunError{(code: word)};

begin
  PasDos.RunTimeError(CoreSig._FatalErrorPos, code, 'User Error');
end;


function _r_rand{: longreal};

var
  x: record low,high: word end;
BEGIN
  x.low := random(0);
  x.high := random(0);
  _r_rand := ((x:>integer)/(maxint)+1.0);
end;

function _w_rand{(range: word): word};

begin
  if range = 0 then
    _w_rand := 0
  else
    _w_rand := random(range);
end;

function  UpCase{(ch: char): char};

begin
  if (ch >= 'a') and (ch <= 'z') then
    ch := chr(ord(ch) - 20H);
  UpCase := ch;
end;


(************************** Conversion functions *****************************)

procedure _val_int{(s: string[high]; var i: integer; var code: word)};
type
  stat_type = (signed, hex);
  stat_set = set of stat_type;
  ConvIntType = array ['0'..'F'] of byte;
const
  HexConv = ConvIntType( 0,1,2,3,4,5,6,7,8,9,255,255,255,255,255,255,255,10,11,12,13,14,15);
  DecConv = ConvIntType( 0,1,2,3,4,5,6,7,8,9,255,255,255,255,255,255,255,255,255,255,255,255,255);
var
  n, l, ml: word;
  status: stat_set;
  val, t: integer;
  c: char;
label
  999;
begin
  n := 1;
  l := length(s);
  status := [];
  val := 0;
  ml := 10;
  if s[n] = '-' then
    begin
    status := status + [signed];
    n := n + 1;
    end
  else if s[n] = '$' then
    begin
    status := status + [hex];
    n := n + 1;
    ml := 16;
    end;
  while n <= l do
    begin
    c := UpCase(s[n]);
    if (c < '0') or (c > 'F') then
      begin
      code := n;
      goto 999;
      end;
    if hex in status then
      t := HexConv[c]
    else
      t := DecConv[c];
    if t = high then
      begin
      code := n;
      goto 999;
      end;
    val := val * ml;
    val := val + t;
    n := n + 1;
    end;
  code := 0;
  if signed in status then
    i := -val
  else
    i := val;
999:
end;

procedure _val_real{(s: string[high]; var i: longreal; var code: word)};

type
  scan_state = (Start, ManSign, LeftDigits, DecPoint, RightDigits, Exponent, ExpSign, ExpDigits);
var
  ps: string[64];
  n, l: word;
  state: scan_state;
  c: char;
  OK: boolean;
label
  555;
begin
  n := 1;
  l := length(s);
  while n <= l do
    begin
    c := UpCase(s[n]);
    if (c = '-') or (c = '+') then
      begin
      if state = Start then
        state := ManSign
      else if state = Exponent then
        state := ExpSign
      else
        goto 555;
      end
    else if c = 'E' then
      begin
      if (state <> LeftDigits) and (state <> RightDigits) then
        goto 555;
      state := Exponent;
      end
    else if c = '.' then
      begin
      if (state <> LeftDigits) then
        goto 555;
      state := DecPoint;
      end
    else if (c >= '0') and (c <= '9') then
      begin
      case state of
        Start, ManSign :
          state := LeftDigits;
        DecPoint:
          state := RightDigits;
        Exponent, ExpSign:
          state := ExpDigits;
        ExpDigits, LeftDigits, RightDigits:
        end;
      end
    else
      goto 555;
    ps[n] := s[n];
    n := n + 1;
    end;
555:
  if n <> l + 1 then
    code := n
  else
    begin
    ps.length := n - 1;
    i := StrToReal(ps, OK);
    end;
end;


procedure format(var s: string[high]; sc: convstring; w: word);

var
  Left: boolean;
  pad: int16;
  len, n, p: word;
begin
  Left := false;
  len := length(sc);
  pad := w - len;
  n := 1;
  if pad < 0 then
    pad := 0;
  if w < 0 then
    begin
    w := -w;
    Left := true;
    end;
  if (not Left) and (pad <> 0) then
    begin
    while (pad <> 0) and (n < high) do
      begin
      s[n] := ' ';
      n := n + 1;
      pad := pad - 1;
      end;
    end;
  if len <> 0 then
    begin
    p := 1;
    while (p <= len) and (n < high) do
      begin
      s[n] := sc[p];
      n := n + 1;
      p := p + 1;
      end;
    end;
  while (pad <> 0) and (n < high) do
    begin
    s[n] := ' ';
    n := n + 1;
    pad := pad - 1;
    end;
  s[0] := chr(n-1);
end;

procedure _str_int{(i:integer;w:int16; var s: string[high])};

var
  sc: convstring;
  OK: boolean;
begin
  IntToStr(i, sc, 10, OK);
  if OK then
    format(s, sc, w)
  else
    s[0] := chr(0);
end;

procedure _str_real_fix{(r:longreal;w:int16; var s: string[high])};

begin
  _str_real(r, w, -29, s);
end;

procedure _str_real{(r:longreal;w:int16;d:int16; var s: string[high])};

var
  Left, OK: boolean;
  fs: string[64];
  sc: convstring;
  l: word;

  procedure convert;

  var
    n: word;
  begin
    n := 1;
    while n <= l do
      begin
      sc[n] := fs[n];
      n := n + 1;
      end;
    sc[0] := chr(n-1);
  end;

begin
  if w < 0 then
    begin
    Left := true;
    w := -w;
    end
  else
    Left := false;
  RealToStr(r,d,w,fs,OK);
  l := fs.length;
  if OK then
    begin
    if w < l then
      w := l;
    if Left then
      w := -w;
    convert;
    format(s, sc, w)
    end
  else
    s[0] := chr(0);
end;

(***************************** I/O functions ********************************)

type
  ByteArray = array [0..maxword-1] of byte;

procedure _r_blockread{(var f:file; var buf; count: word; var result: word)};

var
  St: PasCore.FileInf;
  stat: int16;
  numtoread, n, pos: word;
label
  999;
begin
(*%T _mthread *)
  PasProc.Lock;
(*%E *)
  SetIOR(0);
  result := 0;
  if not (Initialized in f.state) then
    begin
    SetIOR(103);
    goto 999;
    end;
  if Seeked in f.state then
    begin
    f.mode := rm;
    f.state := f.state - [Seeked];
    end;
  if f.mode <> rm then
    begin
    SetIOR(104);
    goto 999;
    end;
  St := (f.stream:>PasCore.FileInf);
  if (CoreFile._F_EOF in St^.flag) then
    begin
    SetIOR(100);
    goto 999;
    end;
  f.bufferptr := St^.Ptr;
  count := count * f.objectsize;
  pos := 0;
  while count > 0 do
    begin
    if St^.Cnt+f.objectsize > count then
      numtoread := count
    else
      numtoread := St^.Cnt+f.objectsize;
    count := count - numtoread;
    move(f.bufferptr^, (buf:>ByteArray)[pos], numtoread);
    St^.Cnt := St^.Cnt - numtoread;
    pos := pos + numtoread;
    if St^.Cnt < 0 then
      begin
      stat := PasCore.FilBuf(St, (Acrt in f.state));
      if stat < 0 then
        goto 999;
      f.bufferptr := St^.Base;
      St^.Cnt := St^.Cnt - f.objectsize;
      end
    else
      begin
      (St^.Ptr:>word) := (St^.Ptr):>word + numtoread;
      f.bufferptr := St^.Ptr;
      end;
    result := result + numtoread div f.objectsize;
    end;
999:
(*%T _mthread *)
  PasProc.Unlock;
(*%E *)
end;

procedure _blockread{(var f:file; var buf; count: word)};

var
  r: word;
begin
  _r_blockread(f, buf, count, r);
  if (r <> count) and IOCheck then
    PasDos.RunTimeError(CoreSig._FatalErrorPos, 0E1H, 'blockread: Read error');
end;

procedure _r_blockwrite{(var f:file; var buf; count: word; var result: word)};

var
  St: PasCore.FileInf;
  Stat: int16;
  Pos: Integer;
  numtowrite: integer;
label
   999;
begin
(*%T _mthread *)
  PasProc.Lock;
(*%E *)
  SetIOR(0);
  if not (Initialized in f.state) then
    SetIOR(103);
  St := (f.stream:>PasCore.FileInf);
  if Seeked in f.state then
    begin
    if f.mode = rm then
      begin
      Pos := CoreIO.tell(f.handle);                   (* input stream *)
      Pos := Pos-St^.Cnt-f.objectsize;
      Pos := CoreIO.lseek(f.handle, Pos, CoreIO.SEEK_SET);
      St^.Cnt := 0;
      St^.Flag := St^.Flag + [CoreFile._F_RST];
      St^.Flag := St^.Flag - [CoreFile._F_OUT, CoreFile._F_IN];
      St^.Ptr := St^.Base;
      f.mode := wm;
      end;
    f.state := f.state - [Seeked];
    end;
  if f.mode <> wm then
      SetIOR(105);

  f.bufferptr := St^.Ptr;
  pos := 0;
  result := 0;
  count := count * f.objectsize;
  while true do
    begin
    if St^.Cnt = 0 then
      begin
      if (not (ACrt in f.state)) and (App in f.state) then
        PasCore.FindEOF(f);
      Stat := PasCore.FlsBuf(St, (Acrt in f.state));
      if (Stat = -1) or (Stat = 0) then
        begin
        SetIOR(101);
        goto 999;
        end;
      f.bufferptr := St^.Base;
      end;
    if count = 0 then
      goto 999;
    if St^.Cnt > count then
      numtowrite := count
    else
      numtowrite := St^.Cnt;
    count := count - numtowrite;
    move((buf:>ByteArray)[pos], f.bufferptr^, numtowrite);
    (St^.Ptr:>word) := (St^.Ptr):>word + numtowrite;
    f.bufferptr := St^.Ptr;
    St^.Cnt := St^.Cnt - numtowrite;
    pos := pos + numtowrite;
    result := result + numtowrite div f.objectsize;
    end;
999:
(*%T _mthread *)
  PasProc.Unlock;
(*%E *)
end;

procedure _blockwrite{(var f:file; var buf; count: word)};

var
  r: word;
begin
  _r_blockwrite(f, buf, count, r);
  if (r <> count) and IOCheck then
    PasDos.RunTimeError(CoreSig._FatalErrorPos, 0E1H, 'blockwrite: Write error');
end;


function SeekEof{(var f: filetype): boolean};

var
  c: char;
label
  888, 999;
begin
  SetIOR(-1);
  if TextFile in f.state then
    SetIOR(0);
    begin
    while not PasLib._f_eof(f) do
      begin
      c := (f.bufferptr:>CharPtr)^;
      if (c <> ' ') and (not (EOLinBuffer in f.state)) and (c <> chr(9)) then
        goto 888;
      get(f);
      end;
    SeekEof := true;
    goto 999;
888:
    SeekEof := false;
    end;
999:
end;

function SeekEoln{(var f: filetype): boolean};

var
  c: char;
label
  777, 888, 999;
begin
  SetIOR(-1);
  if TextFile in f.state then
   SetIOR(0);
    begin
    while not Paslib._f_eof(f) do
      begin
      if EOLinBuffer in f.state then
        goto 777;
      c := (f.bufferptr:>CharPtr)^;
      if (c <> ' ') and (c <> chr(9)) then
        goto 888;
      get(f);
      end;
777:
    SeekEoln := true;
    goto 999;
888:
    SeekEoln := false;
    end;
999:
end;



procedure SetTextBuf{(var f: text; var: buf; size: word)};

var
  BI: FileInf;
begin
  if (size <> 0) and (TextFile in f.state) then
    begin
    BI := f.stream:>FileInf;
    if not (CoreFile._F_RST in BI^.Flag) then
      SetIOR(-1)
    else
      begin
      SetIOR(0);
      if not (CoreFile._F_UBUF in BI^.Flag) then
        paslib._deallocate(BI^.Base, 0);
      BI^.Ptr := adr(buf);
      BI^.Base := BI^.ptr;
      f.bufferptr := BI^.ptr;
      BI^.Size := size;
      BI^.Flag := BI^.Flag + [CoreFile._F_UBUF];
      end;
    end
  else
    SetIOR(-1);
end;

(*%F _OS2 *)
(*# save *)
(*# call(near_call=>off) *)
procedure HardErrorHandler(deverror: word; errcode: word; devhdr: faraddress);

begin
  if IOCheck then
    PasDos.hardretn(errcode + 150)
  else
    PasDos.hardresume(2);
end;
(*# restore *)
(*%E *)
begin
  StackLimit := CoreMain._stk_base;
(*%F _OS2 *)
  PasDos.harderr(HardErrorHandler);
(*%E *)
end .
